/**
    $Header: RODhardware Object interface for the ROD-Controller Application
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $
    
    11/May/1999:
    Each ROD Module Maps as SLAVE VME csr_space, Input_Buffer, Output_Buffer
   
    19/May/1999:
    Added o_idx, i_idx indexes to perform DMAs over ROD Modules Buffers
    Added lives which counts the number of Not ACKs from each Module    
    
    30/September/1999: 
    Front-End definition starts. Many2One scheme is used
    Types created: Front_End, Mod_link
**/

#ifndef __RODhwINC__
#define __RODhwINC__

#include <stdio.h>
#include <stdlib.h>

#include "Commands.h"
#include "RODvme.h"

/* For 16Kbytes CSR, 8 Modules - 8 FrontEnds */
#define CSR_ROD_SIZE 64 /* 1/4 Kybte of Controller CSR for Each ROD */
#define CSR_FE_SIZE  64 /* 1/4 Kbyte of Controller CSR for Each FE  */
#define VME_Obj   1
#define BOOT_TIME_OUT 10 /* Local Time Out in seconds */
#define ROD_DEAD        0xDeadBeef
#define BUSY            0xabcdefab
#define GET_EVENT          0xcaca1
#define RELEASE            0xcaca2
#define CONTROLLER_DEAD    0xcaca3
#define ROD_ALIVEs         0xcaca4
#define CONTROLLER_ALIVEs  0xcaca5
#define COMMAND_TIME_OUT  100
#define UNDEF    -2
#define DOWN     -1
#define UP        1
#define LIVE_CHANCE 5
#define __feBOTH__  0
#define __feVME__   1
#define __feSLINK__ 2

/*+ Data types to interface data-base raw values and RODhardware data types +*/
/*+ Front_End Info +*/
typedef struct{
  char    ip_addr[ 20 ];
  char    io_type[ 20 ]; 
}Link;
typedef struct{
  char    id[ 20 ];
  char    ip_addr[ 20 ];
  word32  no_itm;
  Link    link[ 10 ]; /*+ Maximum 10 modules per Front_End +*/
}FE_itm;
typedef struct{
  FE_itm set[ 10 ];/*+ Maximum 10 fe per controller +*/
  word32  no_itm;
}FE_entry;
/*+ Module Info +*/
typedef struct{
  char    id[ 20 ];
  char    ip_addr[ 20 ];
  char    fe_addr[ 20 ];
}M_itm;
typedef struct{
  word32  no_itm;
  M_itm   set[ 10 ];/*+ Maximum 10 Modules per controller +*/
}M_entry;
/*+ Controller Info +*/
typedef struct{
  char id[ 20 ];
  char ip_addr[ 20 ];
}C_entry;
/* 07/Oct/99 RODmap.cfg is inserted on below pattern */
typedef struct{
  FE_entry fe;
  C_entry  ctrl;
  M_entry  mod;
}ROD_info;


/*+ ROD hardware Data Types ++++++++++++++++++++++++++++++++++++++++++++++++ */
typedef struct{
  char   ip_addr[20];
  word32 io_type;
}Mod_Link;

typedef struct{
  char      id[20];
  char      ip_addr[20];
  word32    no_modules;
  int       lives;  
  int       status;
  word32   *l_csr;
  word32   *r_csr;
  word32    evt_size;
  Mod_Link *modules; /*+ Modules vector */
}Front_End;

typedef struct{
  char    id[20];
  char    ip_addr[20];
  word32  o_index; /*+ Index of Output Buffer Object, shortcut for DMAs +*/
  word32  i_index; /*+ Index of Input  Buffer Object, shortcut for DMAs +*/
  int     lives; 
  int     status;
  word32 *l_csr;
  word32 *r_csr;
  word32 *r_ibuf;
  word32 *r_obuf;
  word32  evt_size;
  word32  fe_ptr;  /*+ Pointer to fe_set[i]. Beware, you need casting   +*/
}ROD_Module;

/*+ 02/June/1999: Virtual pointers to local Slave Buffers +*/
/*+ 07/October/1999: FrontEnd Vector +*/
typedef struct{
  VME_Devs   *vme;         /*+ RODVMECrate Configuration         +*/
  ROD_Module *module_set;  /*+ RODModule Vector                  +*/
  word32      mod_size;    /*+ Module size                       +*/
  Front_End  *fe_set;      /*+ FrontEnd  Vector                  +*/
  word32      fe_size;     /*+ FrontEnd size                     +*/
  word32     *l_ibuf;      /*+ VirtualAddresses of Local Buffers +*/
  word32     *l_obuf;
  word32     *l_csr;       /*+ VirtualAddresses of Local CSR[0]  +*/
}ROD_Hardware;

int Open_Hardware     ( ROD_Hardware** , word32 info_ptr );
int Close_Hardware    ( ROD_Hardware* );
int Detect_Modules    ( ROD_Hardware* , word32 info_ptr );
int Detect_FrontEnd   ( ROD_Hardware* , word32 info_ptr );
int SetUp_Modules     ( ROD_Hardware* , word32 info_ptr );
int SetUp_FrontEnd    ( ROD_Hardware* , word32 info_ptr );
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

/*+ 07/December/99: Data Types to issue Commands updated  +*/
typedef struct{
  word32 *l_csr;      /*+ Local CSR  +*/
  word32 *r_csr;      /*+ Remote CSR +*/
  word32 *lives;      /*+ Device Lives Flag +*/
  word32  cmd;        /*+ Command Word to be written +*/
  word32 *status;     /*+ Device Status Flag +*/
  char   *stat;       /*+ String Info +*/ 
}Command_pack;
int  Send_Command( Command_pack* );
int Device_Check     ( ROD_Hardware*, word32 item, word32 device ,char *stat );

/*+ 24/May/1999 Interfaces to Config ROD Module +*/
typedef struct{
  ROD_Hardware  *rod_hard;
  word32         module;
  word32        *data;
  char           stat[ 80 ];
}CFG_Command;
int Module_Set_Config ( CFG_Command* );
int FEnd_Set_Config   ( CFG_Command* );

/*+ Interfaces to Sample Module CSRs-Output_Buffers +*/
typedef struct{
  ROD_Hardware *rod_hard;
  word32        module;
  char          stat[ 40 ];
}Event_Packet;
int Module_Get_Event( Event_Packet* );

typedef struct{
  ROD_Hardware *rod_hard;
  word32       *buffer;
  word32        module;
  char          stat[ 40 ];
}CSR_Packet;
int Module_Get_CSR  ( CSR_Packet*   );
/* ------------------------------------------------------------------------- */
#endif
