/** 
   $Header: RODSoftware interface for the ROD-Controller Application.
   Created & Maintaned by JB. Juanba.Romance@ific.uv.es $
   
   03/June/99:
   Data Types Definition 

   09/December/99: 
   Threads Syncronisms redesigned. Purpose: Avoid any race condition between 
   Productor and IU Engines. 
**/

#ifndef __RODSoftwareINC__
#define __RODSoftwareINC__

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include "RODhardware.h"

#define __IOData__    10
#define __Up__         1
#define __Down__       2
#define __Pause__      3
#define __Continue__   4
#define __Run__        5

#define __EventSize__ 300

#ifndef SUCCESS
#define SUCCESS 1
#endif
#ifndef FAIL
#define FAIL   -1
#endif

/* ----------------- Data Types To Manage VME Data Injection --------------- */
typedef struct{
  char       io_id[ 20 ];   /*+ Thread Identifier                        +*/
  word32    *buffer;        /*+ Driver Mode - VME data Objects           +*/ 
  word32     dma_obj;       /*+ Target Mode                              +*/
  word32    *l_csr;         /*+ Local  CSR vp.                           +*/
  word32    *r_csr;         /*+ Remote CSR vp.                           +*/
  word32     cfg;           /*+ Module Configuration Word                +*/
  word32    *stat;          /*+ Common Flag {STOP, PAUSE, CONTINUE}      +*/ 
  word32    *evt_buf;       /*+ Event Buffer Ptr                         +*/
  /* Synchronization tools */ 
  int              io_ready; 
  int             *ready;   /*+ ptr to either Productor or Monitor flag  +*/
  pthread_mutex_t *io_mtx;  /*+ To Lock ready flags and cv               +*/
  pthread_cond_t  *io_cv;   /*+ Syncronism                               +*/
  pthread_t        phy_th;  /*+ Physical Thread                          +*/
}IO_Type;
int Input_Engine      ( void *par );
int Output_Engine     ( void *par );
int In_Module_Handshake( IO_Type *in, word32 *count );


typedef struct{
  word32     cfg;           /*+ Module Configuration Word                 +*/
  word32    *stat;          /*+ Control Common Flag {STOP,PAUSE,CONTINUE} +*/ 
  word32    *evt_buf;       /*+ Event Buffer Ptr                          +*/
  /* Synchronization tools */ 
  int            *io_ready; /*+ ptr to IO flag                            +*/
  int             ready;    /*+ Local Flag                                +*/
  pthread_mutex_t io_mtx;   /*+ To Lock ready flags and cv                +*/
  pthread_cond_t  io_cv;    /*+ IO condition variable                     +*/
  pthread_t       phy_th;   /*+ Physical Thread                           +*/
}Productor;
int Production_Engine ( void *par );


typedef struct{
  word32     cfg;           /*+ Module Configuration Word                 +*/
  word32    *stat;          /*+ Control Common Flag {STOP,PAUSE,CONTINUE} +*/ 
  word32    *evt_buf;       /*+ Event Buffer Ptr                          +*/  
  /* Synchronization tools */
  int            *io_ready; /*+ ptr to IO flag                            +*/
  int             ready;    /*+ Local Flag                                +*/
  pthread_mutex_t io_mtx;   /*+ To Lock ready flags and cv                +*/
  pthread_cond_t  io_cv;    /*+ Syncronism                                +*/
  pthread_t       phy_th;   /*+ Physical Thread                           +*/
}Monitor;
int Monitor_Engine    ( void* par );


/*+ ROD-Software Object +*/
typedef struct{
  IO_Type    in;
  IO_Type    out;
  Monitor    mon;
  Productor  pro;
  word32    *stat;     /*+ Software State: RUNNING, STOPED +*/
}ROD_Software;
int Open_Software     ( ROD_Software**, word32 stat_addr );
int Close_Software    ( ROD_Software* );
int Run_Software      ( ROD_Software*, word32 *io_data );
int Pause_Software    ( ROD_Software* );
int Continue_Software ( ROD_Software* );

#endif
