/*
    $Header: RODhardware Object interface for the ROD Front-End Application
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $
*/

#ifndef __RODhardwareINC__
#define __RODhardwareINC__

#include <stdio.h>
#include <stdlib.h>

#include "Commands.h"
#include "RODvme.h"
#if defined( LinuxPCBit3 )
#include "RODslink.h"
#endif

/* For 16Kbytes CSR, 8 Modules - 8 FrontEnds */
#define CSR_ROD_SIZE 64            /* 1/4 Kybte for Modules */
#define CSR_CTRL_SIZE CSR_ROD_SIZE /* idem for CTRL */
#define CSR_FE_SIZE  64            /* idem for  FE  */
#define VME_Obj   1
#define BOOT_TIME_OUT 30 /* Local Time Out in seconds */
#define ROD_DEAD           0xDeadBeef
#define BUSY               0xabcdefab
#define GET_EVENT          0xcaca1
#define RELEASE            0xcaca2
#define CONTROLLER_DEAD    0xcaca3
#define FE_ALIVEs          0xcaca4
#define CONTROLLER_ALIVEs  0xcaca5

#define COMMAND_TIME_OUT   100000
#define UNDEF    -2
#define DOWN     -1
#define UP        1
#define LIVE_CHANCE 5
#define __feBOTH__  0
#define __feVME__   1
#define __feSLINK__ 2
#define __Undef__   -1

/*+ Data types to interface data-base raw values and RODhardware data types +*/
/*+ Front_End Info +*/
typedef struct{
  char    link_id[ 20 ];
  char    ip_addr[ 20 ];
  char    io_type[ 20 ]; 
}Link;
typedef struct{
  char    id[ 20 ];
  char    ip_addr[ 20 ];
  word32  no_itm;
  word32  crate_mod_size;
  word32  idx;
  Link    link[ 10 ]; /*+ Maximum 10 modules per Front_End +*/
}FE_entry;

/*+ Controller Info +*/
typedef struct{
  char id[ 20 ];
  char ip_addr[ 20 ];
}C_entry;

typedef struct{
  FE_entry   fe;
  C_entry  ctrl;
}ROD_info;

/*+ ROD hardware Data Types ++++++++++++++++++++++++++++++++++++++++++++++++ */
/*+ ROD Controller +*/
typedef struct{
  char    id[20];
  char    ip_addr[20];
  word32  o_index;  /*+ Index of Output Buffer Object, shortcut for DMAs +*/
  word32  i_index;  /*+ Index of Input  Buffer Object, shortcut for DMAs +*/
  word32 *l_csr;    /*+ VAddr(ctrl_base) of Local  CSR, VME Slave  +*/
  word32 *r_csr;    /*+ VAddr(ctrl_base) of Remote CSR, VME Master +*/
  int     lives; 
  int     status;
}ROD_Ctrl;

/*+ ROD Module +*/
typedef struct{
  char    id[20];
  char    ip_addr[20];
  word32  o_index;  /*+ Index of Output Buffer Object, shortcut for DMAs +*/
  word32  i_index;  /*+ Index of Input  Buffer Object, shortcut for DMAs +*/
  word32 *l_csr;    /*+ VAddr(m_base) of Local  CSR, VME Slave  Segment_i  +*/
  word32 *r_csr;    /*+ VAddr(m_base) of Remote CSR, VME Master Segment_i  +*/
  word32 *r_ibuf;   /*+ VAddr(m_base) of Ibuffer, VME DMA      +*/
  word32 *r_obuf;   /*+ VAddr(m_base) of Obuffer, VME DMA      +*/
  word32  evt_size;
  int     lives; 
  int     status;
  int     io_type;
}ROD_Module;

typedef struct{
#if defined( LinuxPCBit3 )
  ROD_slink   *slink;
#endif
  ROD_vme     *vme;         /*+ RODVMECrate Configuration                +*/  
  word32       mod_size;    /*+ Module size                              +*/
  ROD_Module  *module_set;  /*+ RODModule Vector                         +*/
  ROD_Ctrl     ctrl;        /*+ ROD Controller Resources                 +*/
  word32       o_index;     /*+ Index of Local Slave Objects             +*/
  word32       i_index;
  word32      *l_ibuf;      /*+ VAddr(base) of Local Buffers, VME Slaves +*/
  word32      *l_obuf;
  word32      *l_csr;       /*+ VAddr(base) of local CSR, VME Slave      +*/
}ROD_Hardware;
int Open_Hardware        ( ROD_Hardware**, word32  info_ptr );
int Close_Hardware       ( ROD_Hardware* );
int Detect_Controller    ( ROD_Hardware* , word32  info_ptr );
int Retrieve_Virtual_Map ( ROD_Hardware* , word32* info, word32* cfg_word );

int SetUp_Hardware       ( ROD_Hardware* , word32 info_ptr );
int SetUp_Controller     ( ROD_Hardware* , word32 info_ptr );
int SetUp_Module         ( ROD_Hardware* , word32 info_ptr );
int SetUp_Local          ( ROD_Hardware* );
int Handshake_Controller ( ROD_Hardware* );



/* ------------------------------------------------------------------------- */
#endif
