/** 
   $Header: RODSoftware interface for the Front-End Application.
   Created & Maintaned by JB. Juanba.Romance@ific.uv.es $
   
   09/December/99:
   Front-End Data injection STARTs. Data Types And Interfaces quoted from 
   RODSoftware on the ROD-Controller application
**/

#ifndef __RODSoftwareINC__
#define __RODSoftwareINC__

#ifndef SUCCESS
#define SUCCESS 1
#endif
#ifndef FAIL
#define FAIL   -1
#endif

#include <pthread.h>
#include "RODhardware.h"

#define __IOData__    10
#define __Up__         1
#define __Down__       2
#define __Pause__      3
#define __Continue__   4
#define __Run__        5

typedef union{
  struct{
    word8  pro   :8;
    word8  in    :8;
    word16 dummy :16;
  }bit;
  word32 all;
  word8  vect[ 2 ];
}Soft_State;


/*+ Productor Pattern +*/
typedef struct{
  char       io_id[ 20 ];   /*+ Thread Identifier +*/
  word32     cfg;           /*+ Module Configuration Word +*/
  word32    *stat;          /*+ Control Common Flag { STOP,PAUSE,CONTINUE } +*/
  word32    *evt_buf;       /*+ Event Buffer Ptr +*/
  /* Synchronization tools */ 
  int            *io_ready; /*+ ptr to IO flag +*/
  int             ready;    /*+ Local Flag +*/
  pthread_mutex_t io_mtx;   /*+ To Lock ready flags and cv +*/
  pthread_cond_t  io_cv;    /*+ IO condition variable +*/
  pthread_t       phy_th;   /*+ Physical Thread  +*/
}Productor;
int Production_Engine ( void *par );

/*+ IO Thread +*/
typedef struct{
  char       io_id[ 20 ];   /*+ Thread Identifier +*/
  word32    *buffer;        /*+ Driver Mode - VME data Objects +*/ 
  word32     dma_obj;       /*+ Target Mode +*/
  word32    *l_csr;         /*+ Local  CSR vp. +*/
  word32    *r_csr;         /*+ Remote CSR vp. +*/
  word32     cfg;           /*+ Module Configuration Word +*/
  word32    *stat;          /*+ Control Common Flag {STOP, PAUSE, CONTINUE} +*/
  word32    *evt_buf;       /*+ Event Buffer Ptr +*/
  word32     io_type;       /*+ Slink/VME flag mode +*/
  word32     slink_addr;    /*+ Slink Device Virtual Address +*/
  /* Synchronization tools */ 
  int              io_ready; 
  int             *ready;   /*+ ptr to either Productor flag  +*/
  pthread_mutex_t *io_mtx;  /*+ To Lock ready flags and cv +*/
  pthread_cond_t  *io_cv;   /*+ condition variable +*/
  pthread_t        phy_th;  /*+ Physical Thread +*/
}IO_Type;
int Input_Engine       ( void *par );
int In_Handshake_VME   ( IO_Type*, word32 *count );
int In_Handshake_Slink ( IO_Type*, word32 *count );

/*+ ROD Software Interface +*/
typedef struct{
  IO_Type   in;
  Productor pro;
  word32    *stat;     /* Data Injection State */
}ROD_Software;
int Open_Software     ( ROD_Software**, word32 stat_addr );
int Close_Software    ( ROD_Software* );
int Run_Software      ( ROD_Software* );
int Pause_Software    ( ROD_Software* );
int Continue_Software ( ROD_Software* );
int Config_Software   ( ROD_Software*, word32* io_data, word32* cfg_word );
#endif
