#ifndef __CommonINC__
#define __CommonINC__

#if defined( LinuxPCBit3 )
#include <sys/time.h>
#else
#include <time.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include "Commands.h"


#ifndef __TypesOn__
#define __TypesOn__
typedef unsigned int       word32;
typedef unsigned short int word16;
typedef unsigned char      word8;
#endif

#ifndef CLEAR
#define CLEAR 0
#endif
#ifndef SET
#define SET 1
#endif
#ifndef SUCCESS
#define SUCCESS 1
#endif

#ifndef FAIL
#define FAIL -1
#endif

#define __BarrierSize__    3
#define __NoProcessors__   1

#define UNIT_LOST      -1
#define UNIT_READY     CLEAR
#define UNIT_BUSY      SET
#define UNIT_RUNNING  (SET+1)
#define UNIT_OPEN     (SET+2)


#define ATLAS_HEADER  9
#define ATLAS_TRAILER 3
#define ATLAS_FORMAT  ( ATLAS_HEADER + ATLAS_TRAILER )
#define PU_IB_SIZE ( __NoChannels__*__NoSamples__ ) /* Local PUInput Buffer  */
#define PU_OB_SIZE ( __NoChannels__*__NoSamples__ ) /* Local PUOutPut Buffer */


#define VME_IO       SET
#define SLINK_IO     CLEAR

typedef unsigned long int w32;
typedef unsigned char     w8;

typedef struct{
  int  iu;
  int  du;
  int  pu[ __NoProcessors__ ]; /* Dynamically allocated */
  int  ebu;
  int  cu;
  int  ou;
  int  ttc;
  w32  quit;
}Arch_State;

/*+ Proccessing Unit +*/
typedef struct{
  char              id[ 20 ];   /*+ Unit ID                         +*/
  int               hit;        /*+ Triggers Received               +*/  
  
  pthread_t         phy_th;     /*+ Physical Thread                 +*/
  pthread_mutex_t  *ib_mtx;     /*+ Ptr Barrier Mutex               +*/
  pthread_cond_t   *ib_cv;      /*+ Ptr to PU-Barrier cv            +*/
  pthread_cond_t   *du_cv;      /*+ Ptr to DU cv                    +*/
  int              *i_barrier;  /*+ Ptr to PU-Barrier index         +*/

  pthread_mutex_t  *ob_mtx;     /*+ Ptr Barrier Mutex               +*/
  pthread_cond_t   *ob_cv;      /*+ Ptr to PU-Barrier cv            +*/
  pthread_cond_t   *ebu_cv;     /*+ Ptr to EBU cv                   +*/
  int              *o_barrier;  /*+ Ptr to PU-Barrier index         +*/

  w32               i_size;     /*+ Local Buffer Size               +*/
  w32              *ib;         /*+ IBuffer vector                  +*/ 
  w32               o_size;     /*+ Local OBuffer Size              +*/
  w32              *ob;         /*+ OBuffer vector                  +*/
  w32              *pat;        /*+ Signal Shape Buffer             +*/
  w32               cfg[ 2 ];   /*+ Input & Output Conf. words      +*/

  w32              *quit;       /*+ Ptr to Quit Flag                +*/
  int              *stat;       /*+ Ptr to PUi State on ROD Machine +*/
}PUnit;


typedef struct{
  w32 event_no;
  w32 bcid_no;
  w32 trg_type;
}TTC_Info;


/*+ 26/May/99: Object to Interface IUnit, OUnit with RODHardware. 

  This Packet is filled On User Request incoming from RODController,
  therefore the SetUp is made "ONFLY"
+*/
typedef struct{
  w32   io_type;    /*+ See Commands.h                +*/
  w32   io_mode;    /*+ Idem than above               +*/
  /* Local Parametters */
  w32   buffer;     /*+ VAddr to data buffer          +*/
  w32   paddr;      /*+ Physical Address of above     +*/
  w32   size;       /*+ Buffer Size                   +*/
  w32   evt_size;   /*+ Event Size                    +*/
  w32   lcsr_vaddr; /*+ VAddr to local CSR            +*/
  /* Remote Parametters */
  w32   dma_dev;    /*+ Ptr to DMA Object             +*/
  w32   rcsr_vaddr; /*+ VAddr to Remote CSR           +*/
  w32   slink_addr; /*+ VAddr to RODSlink Object      +*/
}IOParams;

typedef struct{
  IOParams i_params;
  IOParams o_params;
}IO_Params;
#endif
