/****
     Object Interface Barrier-Machine version for the Input Unit 
*****/

#ifndef __IUnitINC__
#define __IUnitINC__

#include "RODhardware.h"
#include "Common.h"

/*+ Input Unit Object +*/
typedef struct{
  int               hit;
  char              id[ 20 ];

  pthread_t         phy_th;     /*+ Physical Thread                    +*/
  pthread_mutex_t   iu_mtx;     /*+ Input Buffer Mutex                 +*/
  pthread_cond_t    iu_cv;      /*+ IU condition variable              +*/

  pthread_mutex_t   b_mtx;      /*+ Barrier Mutex                      +*/
  pthread_cond_t    b_cv;       /*+ Barrier condition variable         +*/
  int               barrier;    /*+ Barrier index                      +*/

  int              *du_ready;
  int               iu_ready;

  w32              *quit;       /*+ Pointer to Quit Flag               +*/
  int              *stat;       /*+ Pointer to IU State on ROD Machine +*/
  TTC_Info         *ttc_info;   /*+ Pointer to TTC Pool                +*/

  /* I/O Fields to Gather Input Data  */
  /* 26/May/99 VME/Slink I/O Integration Begins.... */
  /* To minimize Latencies Below pointers will be ready after SetUp IUnit, 
     By this way we don't lose time calling Hardware Layer to perform 
     VME/Slink DMAs */

#if defined( PowerPC_Lynx3_0  )
  w32              rod_slink;   /*+ RODSlink core Addr                    +*/ 
#endif

  w32              *l_csr;      /*+ Virtual pointer to Local CSR          +*/
  w32              *r_csr;      /*+ Virtual pointer to Remote CSR         +*/
  w32               dma_dev;    /*+ Address of VME Object                 +*/
  int               io_mode;    /*+ I/O Mode: Target/Driver               +*/
  int               io_type;    /*+ I/O Type: Slink/VME/AutoMode          +*/
  w32               evt_size;   /*+ Event_size                            +*/
  w32               size;       /*+ Buffer Size                           +*/
  w32              *buffer;     /*+ Virtual pointer to Local Input Buffer +*/

}Input_Unit;
int  IU_Open          ( Input_Unit* );
int  IU_Close         ( Input_Unit* );
int  IU_Run           ( Input_Unit* );
int  IU_Engine        ( void *par   );
int  IU_Config        ( Input_Unit*, IOParams* );
int  IHandshake_Slink ( Input_Unit* );
int  IHandshake_VME   ( Input_Unit* );
int  IHandshake_Dummy ( );
void Update_Err_Count ( w32* l_csr, w32* r_csr, w32 hit );

#endif
