/** 
   $Header: Object Interface Barrier-Machine version for the OUnit
   Created & Maintaned by JB. Juanba.Romance@ific.uv.es $

   08/November/1999: Integration Starts for the Barrier Version 
**/

#ifndef __OUnitINC__
#define __OUnitINC__

#include "Common.h"
#include "RODhardware.h"

/*+ OutPut Engine Unit Object +*/
typedef struct{
  int               hit;
  char              id[ 20 ];
  int               h_type;      /*+ Hardware Output Type                 +*/
  
  pthread_t         phy_th;      /*+ Physical Thread                      +*/

  pthread_mutex_t  *b_mtx;       /*+ Ptr to Barrier mtx                   +*/
  pthread_cond_t   *b_cv;        /*+ Ptr to Barrier cv                    +*/
  int              *barrier;     /*+ Ptr to Barrier index                 +*/

  pthread_mutex_t  *ebu_mtx;     /*+ Ptr to EBU mtx                       +*/
  pthread_cond_t   *ebu_cv;      /*+ Ptr to EBU cv                        +*/
  
  TTC_Info         *ttc_info;    /*+ TTCInfo                              +*/
  pthread_mutex_t  *ttc_mtx;     /*+ Ptr to TTC mtx                       +*/
  pthread_cond_t   *ttc_cv;      /*+ Ptr to TTC cv                        +*/
  w32               ou_ready;    /*+ OU Ready flag                        +*/
  w32              *ebu_ready;   /*+ Ptr to EBU Ready flag                +*/

  /* I/O Fields to Gather Input Data  */
  /* 26/May/99 VME/Slink I/O Integration Begins.... */
  /* To minimize Latencies Below pointers will be ready after SetUp IUnit, 
     By this way we don't lose time calling Hardware Layer to perform 
     VME/Slink DMAs */
  w32              *l_csr;      /*+ Virtual pointer to Local CSR          +*/
  w32              *r_csr;      /*+ Virtual pointer to Remote CSR         +*/
  w32               dma_dev;    /*+ Pointer to DMA Object                 +*/
  int               io_mode;    /*+ I/O Mode: Target/Driver               +*/
  int               io_type;    /*+ I/O Type: Slink/VME/AutoMode          +*/
  w32               size;       /*+ Buffer size */
  w32              *buffer;     /*+ Virtual pointer to Local Input Buffer +*/

  w32              *quit;       /*+ Pointer to Quit Flag                  +*/
  int              *stat;       /*+ Pointer to OU State on ROD Machine    +*/
}OutPut_Unit;

int OU_Open           ( OutPut_Unit* );
int OU_Close          ( OutPut_Unit* );
int OU_Run            ( OutPut_Unit* );
int OU_Engine         ( void *par    );
int OU_Config         ( OutPut_Unit*, IOParams* );
int OHandshake_VME    ( OutPut_Unit* );
int OHandshake_Slink  ( OutPut_Unit* );
int OHandshake_Dummy  ( );

#endif
