/*
    $Header: RODhardware Object interface for the ROD-Module Application
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $

    26 March/99      ROD-Hardware Layer Integration starts

    29 March/99      ROD-VME interface done

    03 May/99        Cross Develop. starts. LinuxPCBit3/MVME167/MVME2604PowerPC

    08 May/99        ROD-Slink squeletor ready

    13 May/99        ROD-Controller Interface starts

    02 November/99   Front-End Integration Starts

    03 November/99   Hardware-SetUp functions built

    04 November/99   Interface with ROD.c built
*/

#ifndef __RODhardwareINC__
#define __RODhardwareINC__

#include <stdio.h>
#include <stdlib.h>

#include "Commands.h"
#include "RODvme.h"
#include "RODslink.h"

/* For 16Kbytes CSR, 8 Modules - 8 FrontEnds */
#define CSR_ROD_SIZE  64 /* 1/4 Kybte of CSR for the RODModule */
#define CSR_FE_SIZE   64 /* idem for the FrontEnd  */
#define CSR_CTRL_SIZE CSR_ROD_SIZE /* idem for the ROD-Controller */ 
#define VME_Obj            1
#define SLINK_Obj          2
#define BOOT_TIME_OUT      20 /* Local Time Out in seconds */
#define BUSY               0xabcdefab
#define CONTROLLER_DEAD    0xcaca3
#define ROD_ALIVEs         0xcaca4
#define LIVE_CHANCE 5

/*+ ROD hardware Data Types ++++++++++++++++++++++++++++++++++++++++++++++++ */
/*+ ROD Controller  +*/
typedef struct{
  char     id[20];
  char     ip_addr[20];
  word32   o_index;    /*+ Index of Output Buffer Object, shortcut for DMAs +*/
  word32   i_index;    /*+ Index of Input  Buffer Object, shortcut for DMAs +*/
  word32  *l_csr;      /*+ VAddr(ctrl_base) of Local  CSR, VME Slave        +*/
  word32  *r_csr;      /*+ VAddr(ctrl_base) of Remote CSR, VME Master       +*/
  int      lives; 
  int      status;
}ROD_Ctrl;

/*+ FrontEnd  +*/
typedef struct{
  char     id[20];
  char     ip_addr[20];
  word32   o_index;    /*+ Index of Output Buffer Object, shortcut for DMAs +*/
  word32   i_index;    /*+ Index of Input  Buffer Object, shortcut for DMAs +*/
  word32  *l_csr;      /*+ VAddr(ctrl_base) of Local  CSR, VME Slave        +*/
  word32  *r_csr;      /*+ VAddr(ctrl_base) of Remote CSR, VME Master       +*/
  int      lives; 
  int      status;
}Front_End;

typedef struct{
#if defined( PowerPC_Lynx3_0 )
  RODSlink   *slink;   /*+ Slink Port only Available @ PowerPC +*/
#endif
  VME_Devs   *vme;     /*+ Global VME Configuration Map             +*/
  ROD_Ctrl    ctrl;    /*+ ROD Controller Resources                 +*/
  Front_End   fe;      /*+ FrontEnd Resources                       +*/
  /*+ Slave Parameters +*/
  word32     *l_ibuf;  /*+ VAddr(base) of Local Buffers, VME Slaves +*/
  word32     *l_obuf;
  word32     *l_csr;   /*+ VAddr(base) of local CSR, VME Slave      +*/
  word32      i_index; /*+ Index of Input  Buffer Object +*/
  word32      o_index; /*+ Index of Output Buffer Object +*/
}ROD_Hardware;

int Open_Hardware        ( ROD_Hardware** , word32 info_ptr );
int Close_Hardware       ( ROD_Hardware* );

int Handshake_Controller ( ROD_Hardware* );
int Retrieve_VirtualMap  ( ROD_Hardware*, word32 *info, word32 src_type );


int SetUp_Hardware       ( ROD_Hardware*, word32 info_ptr );
int SetUp_Controller     ( ROD_Hardware*, word32 info_ptr );
int SetUp_FrontEnd       ( ROD_Hardware*, word32 info_ptr );
int SetUp_Local          ( ROD_Hardware* );
/* ------------------------------------------------------------------------- */

#endif
