/**
    $Header: ROD object for the ROD-Controlller Application
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $

    19/May/99  
    BroadCast, Event_Request, Module_Command implementation
   
    24/May/99  
    ROD_Config implementation
   
    27/May/99  
    VME Input Data Flow Defined ( Module )
   
    01/June/99 
    VME Data Injection Starts ( Controller )

    20/November/99
    FrontEnd Management Done. 
    New RODMap.cfg crate configuration ReadOut

    04/December/99
    Common Send_Command primitives implemented to manage Module-s And 
    FrontEnd-s

    09/December/99
    ROD-Config Updated to work with the Front-End Application And new 
    RODhardware.o interfaces

**/

#include "ROD.h"
ROD* sh_rod;
Handler_Filter h_filter;



/*++++++++++++++++++++++++++++++++++++++
  Rod Handler process OS signals from BusError, Ctrl-C And Segmentation Faults

  int sig_num Processed Signal
 
  ++++++++++++++++++++++++++++++++++++++*/
void
ROD_Handler( int sig_num ){

  printf("\n\n[ Received Signal %d ]", sig_num ); 
  switch( sig_num ){
    case SIGBUS  :printf("[ Bus Error ]" );break;
    case SIGINT  :printf("[ Ctrl-C    ]" );break;
    case SIGSEGV :printf("[ Segmentation Fault ]");break;
  }
  sleep( 2 );
  if( h_filter.bit.f_1 == CLEAR ){
    h_filter.bit.f_1 = SET;
    if( sh_rod != NULL ){
      if( sh_rod->hardware != NULL ){
	Module_BroadCast( sh_rod, SHUTDOWN );
	FEnd_BroadCast( sh_rod, SHUTDOWN );
      }
    }
  }
  if( h_filter.bit.f_3 == CLEAR ){
    h_filter.bit.f_3 = SET;
    if( sh_rod->software != NULL )
      Close_Software( sh_rod->software );
  }
  if( h_filter.bit.f_2 == CLEAR ){
    h_filter.bit.f_2 = SET;
    if( sh_rod->hardware != NULL )
      Close_Hardware( sh_rod->hardware );
  }
  if( h_filter.bit.f_4 == CLEAR ){
    h_filter.bit.f_4 = SET;
    if( sh_rod != NULL )
      free( sh_rod );
  }
  printf("\n[ %-20s ]\n","Signal Handler OutPut");
  exit( FAIL );

}

/*++++++++++++++++++++++++++++++++++++++
  Handler Install

  Warning: If Install Fails Application Ends
 
  ++++++++++++++++++++++++++++++++++++++*/

void
Install_Handler( ){
struct sigaction rod_sa;
sigset_t rod_set;

  sigemptyset( &rod_set );
  sigaddset( &rod_set, SIGINT );
  sigprocmask( SIG_BLOCK, &rod_set, NULL ); /* Block CTRLc only on SetUp */ 
  sigfillset( &rod_sa.sa_mask );
  rod_sa.sa_flags = CLEAR; 
  rod_sa.sa_handler = ROD_Handler;
  if( sigaction( SIGINT, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGSEGV, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGBUS, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  sigprocmask( SIG_UNBLOCK, &rod_set, NULL ); /* Now Ready to Proccess */

}

void
Dump_RODinfo( ROD *rod ){
word32 ii, jj;
M_entry    *m_entry; /*+ Shorcuts */
FE_entry   *fe_entry;
C_entry    *c_entry;
Front_End  *fe;
ROD_Module *mod;

 /* Data Base Info */ 
 printf("\n\n#--------------= DataBase Info =--------------#");
 fe_entry = &rod->info.fe; /* Shortcut */
 for( ii = 0; ii < fe_entry->no_itm; ii++ ){
   printf("\n[%s]", fe_entry->set[ ii ].id );
   for( jj =0 ; jj < fe_entry->set[ ii ].no_itm; jj++ ){
     printf("[%s][%s]", fe_entry->set[ ii ].link[ jj ].ip_addr,
	    fe_entry->set[ ii ].link[ jj ].io_type );
   }
   printf("\n");
 }
 m_entry = &rod->info.mod;
 for( ii = 0; ii < m_entry->no_itm; ii++ ){
   printf("\n[Module_id   ][%s]", m_entry->set[ ii ].id );
   printf("\n[Module_addr ][%s]", m_entry->set[ ii ].ip_addr );
   printf("\n[Front End   ][%s]", m_entry->set[ ii ].fe_addr );
 }
 printf("\n");
 c_entry = &rod->info.ctrl;
 printf("\n[ROD Controller:%s][%s]", c_entry->id, c_entry->ip_addr );
 printf("\n#--------------= DataBase Info =--------------#");

 /* RODhardware info */
 /* Front End  */
 printf("\n\n#--------------= FrontEnd Data =--------------#");
 fe = rod->hardware->fe_set; /* Shortcut */
 for( ii = 0; ii < rod->hardware->fe_size; ii++ ){
   if( fe[ ii ].status == UNDEF )
     continue;
   printf("\n[ fe_id   ][ %s ]", fe[ ii ].id );
   printf("\n[ ip_addr ][ %s ]", fe[ ii ].ip_addr );
   printf("\n[ Status  ][ %d ]", fe[ ii ].status );
   printf("\n[ Lives   ][ %d ]", fe[ ii ].lives );
   printf("\n[ l_csr   ][ %08x ]", ( word32 )fe[ ii ].l_csr );
   printf("\n[ r_csr   ][ %08x ]", ( word32 )fe[ ii ].r_csr );
   printf("\n[ NoLinks ][ %d ]", fe[ ii ].no_modules );
   for( jj = 0; jj < fe[ ii ].no_modules; jj++ ){
     printf("\n[ io_type ][ %d ]", fe[ ii ].modules[ jj ].io_type );
     printf("\n[ ip_addr ][ %s ]", fe[ ii ].modules[ jj ].ip_addr );
   }
 }
 printf("\n#-------------= FrontEnd Data =---------------#");

 /* RODModules */
 printf("\n\n#--------------= Module Data =----------------#");
 mod = rod->hardware->module_set; /* Shortcut */
 for( ii = 0; ii < rod->hardware->mod_size; ii++ ){
   printf("\n[ fe_id     ][ %s ]", mod[ ii ].id );
   printf("\n[ ip_addr   ][ %s ]", mod[ ii ].ip_addr );
   printf("\n[ Status    ][ %d ]", mod[ ii ].status );
   printf("\n[ Lives     ][ %d ]", mod[ ii ].lives );
   printf("\n[ EventSize ][ %08x ]", ( word32 )mod[ ii ].evt_size );
   printf("\n[ l_csr     ][ %08x ]", ( word32 )mod[ ii ].l_csr );
   printf("\n[ r_csr     ][ %08x ]", ( word32 )mod[ ii ].r_csr );
   printf("\n[ Remote IBuffer  ][ %08x ]", ( word32 )mod[ ii ].r_ibuf );
   printf("\n[ Remote OBuffer  ][ %08x ]", ( word32 )mod[ ii ].r_obuf );
   printf("\n[ ptr    IBuffer  ][ %08x ]", mod[ ii ].o_index );
   printf("\n[ ptr    OBuffer  ][ %08x ]", mod[ ii ].i_index );
   printf("\n[ ptr    FrontEnd ][ %d ]",   mod[ ii ].fe_ptr );
 }
 printf("\n#--------------= Module Data =----------------#\n");

}

/*++++++++++++++++++++++++++++++++++++++
  Create memory for the MacroStructure ROD
  Initialise Software And Hardware objects on deeper Layers
  
  ROD **rod Address of ROD pointer
 
  ++++++++++++++++++++++++++++++++++++++*/
int
ROD_Open( ROD **rod ){
word32 stat_addr, ii;

 Install_Handler( );
 h_filter.all = CLEAR;
 *rod = ( ROD* )malloc( sizeof( ROD ) );
 if( *rod == NULL ){
   printf("\n[ Not Enough Memory to Alloc ROD ]");
   fflush( stdout );
   return( FAIL );
 }
 sh_rod = *rod;

 /* RODMap.cfg processing -------------------------------------------------- */
 if( Make_FEset( &( ( *rod )->info.fe  ) ) == FAIL )
   return( FAIL );
 if( Make_Mset ( &( ( *rod )->info.mod  ) ) == FAIL )
   return( FAIL );
 if( Make_Ctrl ( &( ( *rod )->info.ctrl ) ) == FAIL )
   return( FAIL );

 /* Still Alive? Open Hardware And Software Resources ---------------------- */
 ( *rod )->hardware = NULL;
   if( Open_Hardware( &( ( *rod )->hardware ), 
		    ( word32 )( &( ( *rod )->info ) )) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Hardware: Open: FAIL" );
   fflush( stdout );
   return( FAIL );
 }

 ( *rod )->software = NULL;
 stat_addr = ( word32 )( &( ( *rod )->stat ) );
 if( Open_Software( &( ( *rod )->software ), stat_addr ) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Software: Open: FAIL" );
   fflush( stdout );
   return( FAIL );
 } 

 ( *rod )->list = ( Crate_Config* )
   malloc( sizeof( Crate_Config )*( *rod )->info.mod.no_itm );
 if( ( *rod )->list == NULL ){
   printf("\n[ %-30s ]", "ROD: Alloc Config List: FAIL" );
   fflush( stdout );
   return( FAIL );
 }
 for( ii = 0 ; ii < ( *rod )->info.mod.no_itm ; ii++ )
   ( ( *rod )->list )[ ii ].stat = __Ready__ ;

 return( SUCCESS );

}

int
Make_Ctrl( C_entry *c_entry ){
FILE* file;
char entry[ 80 ];
word32 dummy;
int done = CLEAR;

 file = fopen( RODMap, "r" );
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Controller_Section:%d", &dummy ) == 1 ){
     if( fgets( entry, 80, file ) == NULL )
       break;
     else{
       strcpy( c_entry->id, strtok( entry, "\t" ) );
       strcpy( c_entry->ip_addr, strtok( NULL, "\n" ) );
       done = SET;
     }
   }
 }
 fclose( file );
 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );      
  
}

int
Make_Mset( M_entry *m_entry ){
word32 ii, done = CLEAR;
FILE* file;
char entry[ 80 ];
M_itm  *set; /*+ shorcut +*/
 
 set = m_entry->set;
 file = fopen( RODMap, "r" );
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Module_Section:%d", &m_entry->no_itm ) == 1 ){
#ifdef VERBOSE
     printf("\nModule Processing starts for %d items", m_entry->no_itm );
     fflush( stdout );
#endif
     for( ii = 0; ii < m_entry->no_itm; ii++ ){
       if( fgets( entry, 80, file ) == NULL )
	   break;
       strcpy( set[ ii ].id, strtok( entry, "\t" ) );
       strcpy( set[ ii ].ip_addr, strtok( NULL, "\t" ) );
       strcpy( set[ ii ].fe_addr, strtok( NULL, "\n" ) );
     }
     done = SET;
   }
 }
 fclose( file );
 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );
 
}


int
Make_FEset(FE_entry *fe_entry ){
word32 ii, jj, done = CLEAR;
FILE* file;
char entry[ 80 ];
FE_itm  *set; /*+ shorcut +*/

 set = fe_entry->set;
 file = fopen( RODMap, "r" );
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Front_End_Section:%d", &fe_entry->no_itm ) == 1 ){
#ifdef VERBOSE
     printf("\nFrontEnd Processing starts for %d items", fe_entry->no_itm );
     fflush( stdout );
#endif
     for( ii = 0; ii < fe_entry->no_itm ; ii++ ){
       if( fgets( entry, 80, file ) == NULL )
	   break;
       strcpy( set[ ii ].id, strtok( entry, "\t" ) );
       strcpy( set[ ii ].ip_addr, strtok( NULL, "\t" ));
       set[ ii ].no_itm = atoi( strtok( NULL, "\t" ) );     
       for( jj = 0 ; jj < set[ ii ].no_itm - 1 ; jj++ ){
	 strcpy( set[ ii ].link[ jj ].io_type, strtok( NULL, "\t" ) );
	 strcpy( set[ ii ].link[ jj ].ip_addr, strtok( NULL, "\t" ) );
       }
       strcpy( set[ ii ].link[ jj ].io_type, strtok( NULL, "\t" ) );
       strcpy( set[ ii ].link[ jj ].ip_addr, strtok( NULL, "\n" ) );
       done = SET;
       break;
     }
   }
 }
 fclose( file );

 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );

}


int
ROD_Close( ROD *rod ){
int status = SUCCESS;
 
 if( rod != NULL ){ 

   if( rod->software != NULL ){
     if( Close_Software( rod->software ) != SUCCESS ){
       printf("\n[ %-30s ]", "ROD: Software: Close: FAIL" );
       fflush( stdout );
       status = FAIL;
     }
   }

   if( rod->hardware != NULL ){
     if( Close_Hardware( rod->hardware ) != SUCCESS ){
       printf("\n[ %-30s ]", "ROD: Hardware: Close: FAIL" );
       fflush( stdout );
       status = FAIL;
     }
   }
   fflush( stdout );
   free( rod->list );
   free( rod );
 }
 return( status );
 
}

int
ROD_Config( ROD *rod, word32 module, word32 config_data ){
int status = FAIL, ii;
Config_Word *cfg;
CFG_Command cfg_cmd;

 if( rod == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops Null Pointer" );
   return( status );
 }
 cfg = ( Config_Word* )config_data;

 /* Trigger Mode */
 cfg[ 4 ].all     = module;/* Will be Recoded on RODhardware.c */
 cfg_cmd.module   = module;
 cfg_cmd.rod_hard = rod->hardware;
 cfg_cmd.data     = ( word32* )( cfg ); 
 sprintf( cfg_cmd.stat, "%s", "Module-Config");

 if( Module_Set_Config( &cfg_cmd ) != SUCCESS )
   printf("\n[ Warning: %-20s", cfg_cmd.stat );
 else{
   cfg[ 0 ].all = __ConfigPackSize__ +1;
   if( FEnd_Set_Config( &cfg_cmd ) != SUCCESS ){
     printf("\n[ Warning: %-20s ]", cfg_cmd.stat );
   }
   else{
     rod->list[ module ].cfg[ 0 ] = cfg[ 1 ].all;
     rod->list[ module ].cfg[ 1 ] = cfg[ 2 ].all;
     rod->list[ module ].cfg[ 2 ] = cfg[ 3 ].all;
     rod->list[ module ].cfg[ 3 ] = cfg[ 4 ].all;
     status = SUCCESS;
   }
 }

 printf("\n\n[ %-20s ]", cfg_cmd.stat );
 printf("\n[ %-20s ]", "Configuration Packet" );
 for( ii = 0 ; ii < __ConfigPackSize__ ; ii++ ){
   printf("\n[ $%08x ]", ( unsigned )( cfg[ ii+1 ].all ) );
 }
 fflush( stdout );
 return( status );
 
}

int
ROD_Sampling( ROD *rod, word32 module, word32 *bs_addr ){
int status = FAIL;
Event_Packet e_pack;
char stat[ 40 ];

 if( Module_Command( rod, module, SAMPLE, stat ) != SUCCESS ){
   printf("\n[ %s ]", stat );
   fflush( stdout ); 
   return( status );
 }
 e_pack.rod_hard = rod->hardware;
 e_pack.module = module;
 if( Module_Get_Event( &e_pack )!= SUCCESS ){
   printf("\n[ Warning Module %u: %s ]", 
	  ( unsigned )e_pack.module, e_pack.stat );
   fflush( stdout ); 
 }
 else{
   *bs_addr = ( word32 )rod->hardware->module_set[ module ].r_obuf;
   printf("\n[ Sampling Module-%u: %s ]", 
	  ( unsigned )e_pack.module, e_pack.stat );
   fflush( stdout ); 
   status = SUCCESS;
 }
 return( status );

}

/* ------------------------------ Module I/O Primitives -------------------- */
int
Module_BroadCast( ROD *rod, word32 command ){
int status = SUCCESS;
char stat[ 40 ];
ROD_Module *module;
word32 ii;
Command_pack cmd_pack;

 if( rod == NULL )
   return( FAIL );
 module = rod->hardware->module_set; /* shortcut */
 cmd_pack.cmd    = command;
 cmd_pack.stat   = stat;
 for( ii = 0; ii < rod->hardware->mod_size ; ii++ ){
   if( Device_Check( rod->hardware, ii, __Module__, stat ) != SUCCESS ){
     printf("\n[ Module-%d: state: %s ]", ii, stat );
     continue;
   }   
   sprintf( stat, "Module-%d", ii );   
   cmd_pack.l_csr  = module[ ii ].l_csr;
   cmd_pack.r_csr  = module[ ii ].r_csr;
   cmd_pack.lives  = &( module[ ii ].lives  );
   cmd_pack.status = &( module[ ii ].status );
   if( Send_Command( &cmd_pack ) != SUCCESS ){
     status = FAIL;
   }
   printf("\n[ %-20s ]", cmd_pack.stat );
 }
 printf("\n[ %-20s ]", "Module-BroadCast Done" );
 return( status );


}


int
Module_Command( ROD *rod, word32 item, word32 cmd, char *stat ){
int status = FAIL;
word32 index, io_data[ __IOData__  ], stat_addr;
ROD_Module *module;
ROD_Hardware *rod_hw;

Config_Word w1,w2;
Command_pack cmd_pack;

 if( rod == NULL )
   return( status );

 module = rod->hardware->module_set;
 if( ( cmd == PAUSE )||( cmd == RUN )||( cmd == CONTINUE )||( cmd == STOP ) ){
   w1.all = rod->list[ item ].cfg[ 0 ];
   w2.all = rod->list[ item ].cfg[ 1 ];

   if( ( rod->stat != __Down__ )&&( rod->list[ item ].stat == __Ready__ )){
     printf("\n[ %-30s ]"," Warning: Local Data Injection Busy");
     return( status );
   }
   

   /* -----------------------= Local Data Injection =----------------------- */
   if(((w1.bit.src_type == __Controller__ )&&( w1.bit.io_type == __VME__ ))||
      ((w2.bit.src_type == __Controller__ )&&( w2.bit.io_type == __VME__ ))){

     switch( cmd ){

     case PAUSE:
       if( rod->stat == __Pause__ ){
	 sprintf( stat,"%s", "Data Injection is Already Paused" );
	 printf("\n [ %-30s ]", stat );
	 return( status );
       }
       if( Pause_Software( rod->software ) != SUCCESS ){
	 return( status );
       }
       break;

     case RUN:	
       if( rod->software != NULL ){
	 if( rod->stat == __Run__ ){
	   sprintf( stat,"%s", "Data Injection is Already Running" );
	   printf("\n [ %-30s ]", stat );
	   return( status );
	 }
       }
       rod_hw = rod->hardware;
       io_data[ 0 ] = ( word32 )( &module[ item ].l_csr[ 0 ] );
       io_data[ 1 ] = ( word32 )( &module[ item ].r_csr[ 0 ] );
       io_data[ 2 ] = ( word32 )( rod_hw->l_ibuf );
       io_data[ 3 ] = ( word32 )( rod_hw->l_obuf );

       index = module[ item ].i_index;
       io_data[ 4 ] = ( word32 )( rod_hw->vme->rod_dev[ index ].vme_obj );
       index = module[ item ].o_index;
       io_data[ 5 ] = ( word32 )( rod_hw->vme->rod_dev[ index ].vme_obj );

       io_data[ 6 ] = ( word32 )module[ item ].r_ibuf;
       io_data[ 7 ] = ( word32 )module[ item ].r_obuf;	 
       io_data[ 8 ] = rod->list[ item ].cfg[ 0 ];
       io_data[ 9 ] = rod->list[ item ].cfg[ 1 ];

       if( rod->software == NULL ){
	 stat_addr = ( word32 )( &( rod->stat ) );
	 if( Open_Software( &( rod->software ), stat_addr ) != SUCCESS ){
	   printf("\n[ %-30s ]", "Software: Open: FAIL" );
	   fflush( stdout );
	   return( status );
	 } 
       }
       if( Run_Software( rod->software, io_data ) != SUCCESS ){
	 return( status );
       }
       rod->list[ item ].stat = __Busy__ ;
       break;
	 
     case CONTINUE:
       switch( rod->stat ){
       case __Run__ :
       case __Continue__ :
	 sprintf( stat, "%s", "Data Injection is Already Running" );
	 printf("\n [ %-30s ]", stat );
	 return( status );
       case __Down__ :
	 sprintf( stat, "%s", "Data Injection is Down" );
	 printf("\n [ %-30s ]", stat );
	 return( status );
       }
       if( Continue_Software( rod->software ) != SUCCESS ){
	 return( status );
       }
       break;

     case STOP:
       if( rod->software == NULL ){
	 printf("\n[ %-30s ]", "Data Injection is Down" );
	 status = FAIL;
	 break;
       }
       if( Close_Software( rod->software ) != SUCCESS ){
	 printf("\n[ %-30s ]", "ROD: Software: Close: FAIL" );
	 status = FAIL;
       }
       printf("\n[ %-30s ]", "Local Data Injection: STOPPED" );
       break;
     }
   }
   else
     if( w1.bit.src_type == __FrontEnd__ ){
       if( rod->list[ item ].stat == __Busy__ ){
	 printf("\n[ %-30s ]"," Warning: Local Data Injection Busy");
	 return( status );
       }
       /* Commands Forwared also to the FronEnd Application */
       switch( cmd ){
       case PAUSE:
       case RUN:
       case CONTINUE: 
       case STOP: 
	FEnd_Command( rod, item , cmd, stat  );
	printf("\n[ %-40s ]", stat );
	break;
       }
     }
   /* ---------------------------------------------------------------------- */
 }

 if( Device_Check( rod->hardware, item,  __Module__, stat ) != SUCCESS ){
   printf("\n[ FrontEnd-%d: state: %s ]", item, stat );
   return( status );
 }
 sprintf( stat, "Module-%d", item );


 cmd_pack.l_csr  = module[ item ].l_csr;
 cmd_pack.r_csr  = module[ item ].r_csr;
 cmd_pack.lives  = &( module[ item ].lives  );
 cmd_pack.status = &( module[ item ].status );
 cmd_pack.cmd    = cmd;
 cmd_pack.stat   = stat;
 if( Send_Command( &cmd_pack ) != SUCCESS ){
   return( status );
 }
 printf("\n[ %-40s ]", cmd_pack.stat );
 
 fflush( stdout );
 status = SUCCESS;
 return( status );

}


int
Get_Module_Stat( ROD *rod, word32 module, word32 *state  ){
int status = FAIL;
CSR_Packet  csr_pack;

 if( rod == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops Null Pointer" );
   return( status );
 }
 if( state == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops User Null Pointer" );
   return( status );
 }
 state[ 0 ] = CSR_ROD_SIZE;
 csr_pack.module   = module; 
 csr_pack.rod_hard = rod->hardware;
 csr_pack.buffer   = &state[ 1 ];
 if( Module_Get_CSR( &csr_pack ) != SUCCESS )
   printf("\n[ Warning Module %u: %s ]", 
	  ( unsigned )csr_pack.module, csr_pack.stat );
 else
   status = SUCCESS;
 return( status );

}

void
Decode_Module_Stat( word32 *status ){
word32 ii;
struct Stat_Lut{
  char   *str;
  word32 idx;
}stat_lut[]={
  { "Number of PUs", 0 },
  { "IUnit hits",    1 },
  { "TTCUnit hits",  2 },
  { "DUnit hits",    3 },
  { "EBUnit hits",   4 },
  { "AutoTrigger count", 5 },
  { "Machine State", 6 },
  { "State Word",    7 },
  { "IOMode", 8 },
  { "IOType", 9 },
  { "BufferSize", 10 },
  { "IOMode", 11 },
  { "IOType", 12 },
  { "BufferSize", 13},
  { "CFGWord1", 14 },
  { "CFGWord2", 15 },
  { "IUFails", 16 }
};

 printf("\n[ STATUS Info ]");
 for( ii = 0; ii < 17 ; ii++ ){
   printf("\n[ %-25s ][ $%08x ]", 
	  stat_lut[ ii ].str, ( unsigned )status[ stat_lut[ ii ].idx + 2 ] );
 }
 printf("\n[ Input Flow ]" );
 Decode_CFGword( status[ stat_lut[ 14 ].idx + 2 ] );
 printf("\n[ OutPut Flow ]" );
 Decode_CFGword( status[ stat_lut[ 15 ].idx + 2 ] );

}

void
Decode_CFGword( word32 cfg_word ){
Config_Word cfg;
char str[ 20 ];

 cfg.all = cfg_word;
 printf("\n[ %-25s ][ %-12s ]", "I/O Mode", 
	cfg.bit.io_mode ? "Driver" : "Target" );
 printf("\n[ %-25s ][ %-12s ]", "Data Source", 
	cfg.bit.src_type ? "FrontEnd" : "Controller" );
 switch( cfg.bit.io_type ){
 case __Slink__ : sprintf( str, "%s", "Slink" );break;
 case __VME__   : sprintf( str, "%s", "VME" )  ;break;
 case __Auto__  : sprintf( str, "%s", "Auto" ) ;break;
 default: sprintf( str, "Unknown Mode" ) ;break;
 }
 printf("\n[ %-25s ][ %-12s ]", "I/O Type", str );
 
 switch( cfg.bit.format_id ){
 case __Calibration__ : sprintf( str, "%s", "Calibration" );break;
 case __ATLAS__       : sprintf( str, "%s", "ATLAS" )  ;break;
 case __Monitor__     : sprintf( str, "%s", "Monitor" ) ;break;
 case __ByPass__      : sprintf( str, "%s", "ByPass" ) ;break;
 default: sprintf( str, "Unknown Mode" ) ;break; 
 }
 printf("\n[ %-25s ][ %-12s ]", "I/O Format",  str );
 sprintf( str, "%04u",  ( unsigned )cfg.bit.no_channels ) ;
 printf("\n[ %-25s ][ %-12s ]", "NoSamples",  str );
 sprintf( str, "%04u",  ( unsigned )cfg.bit.no_samples ) ;
 printf("\n[ %-25s ][ %-12s ]", "NoChannels",  str );

}


/* --------------------- FrontEnd I/O Primitives --------------------------- */
int
FEnd_BroadCast( ROD* rod, word32 cmd ){
int status = SUCCESS;
char stat[ 40 ];
Front_End  *fe;
word32 ii;
Command_pack cmd_pack;

 if( rod == NULL )
   return( FAIL );
 fe = rod->hardware->fe_set;
 cmd_pack.cmd    = cmd;
 cmd_pack.stat   = stat;
 for( ii = 0; ii < rod->hardware->fe_size ; ii++ ){
   if( Device_Check( rod->hardware, ii, __FrontEnd__, stat ) != SUCCESS ){
     printf("\n[ FrontEnd-%d: state: %s ]", ii, stat );
     continue;
   }   
   sprintf( stat, "FrontEnd-%d", ii );   
   cmd_pack.l_csr  = fe[ ii ].l_csr;
   cmd_pack.r_csr  = fe[ ii ].r_csr;
   cmd_pack.lives  = &( fe[ ii ].lives  );
   cmd_pack.status = &( fe[ ii ].status );
   if( Send_Command( &cmd_pack ) != SUCCESS ){
     status = FAIL;
   }
   printf("\n[ %-20s ]", cmd_pack.stat );
 }
 printf("\n[ %-20s ]", "FrontEnd-BroadCast Done" );
 return( status );

}


int
FEnd_Command( ROD* rod, word32 module, word32 cmd, char *stat  ){
int status = FAIL;
Front_End  *fe;
ROD_Module *mod;
Command_pack cmd_pack;
word32 jj, fe_idx;

 if( rod == NULL )
   return( status );
 fe = rod->hardware->fe_set; /* shorcut */
 mod = rod->hardware->module_set; /* shorcut */
 
 fe_idx = mod[ module ].fe_ptr;
 for( jj = 0; jj < fe[ fe_idx ].no_modules; jj++ ){
   if( strcmp( fe[ fe_idx ].modules[ jj ].ip_addr, 
	       mod[ module ].ip_addr ) == 0 ){
     break;
   }
 }

 if( Device_Check( rod->hardware, jj ,  __FrontEnd__, stat ) != SUCCESS ){
   printf("\n[ FrontEnd-%d: state: %s ]", jj, stat );
   return( status );
 }
 sprintf( stat, "FrontEnd-%d", jj );
 
 cmd_pack.l_csr  = fe[ jj ].l_csr;
 cmd_pack.r_csr  = fe[ jj ].r_csr;
 cmd_pack.lives  = &( fe[ jj ].lives  );
 cmd_pack.status = &( fe[ jj ].status );
 cmd_pack.cmd    = cmd;
 cmd_pack.stat   = stat;
 Send_Command( &cmd_pack );
 return( status );

}





