/**
    $Header: RODhardware Object for the RODController Application
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $

    30/September/1999: 
    FrontEnd integration starts
    
    04/December/99
    Common Send_Command primitives implemented
**/

#include <RODhardware.h>


int 
Open_Hardware( ROD_Hardware **rod_hard, word32 info_ptr ){
int status = SUCCESS;

 *rod_hard = ( ROD_Hardware* )malloc( sizeof( ROD_Hardware ) );
 if( *rod_hard == NULL ){
   printf( "\n[ %-30s ]", "Hardware: Malloc: Fail");
   fflush( stdout );
   return( FAIL );
 }
 if( Open_RODVME( &( ( *rod_hard )->vme ) ) != SUCCESS ){
   printf("\n[ %-30s ]","VMEHardware: Open: FAIL" );
   fflush( stdout  );
   status = FAIL;
 }
 if( status == SUCCESS ){
   ( *rod_hard )->module_set = NULL;
   if( Detect_Modules( *rod_hard, info_ptr ) != SUCCESS ){
     printf("\n[ %-30s ]","VME Hardware: ROD Module Detection: FAILs" );
     status = FAIL;
   }
   ( *rod_hard )->fe_set = NULL ;
   if( Detect_FrontEnd( *rod_hard, info_ptr ) != SUCCESS ){
     printf("\n[ %-30s ]","VME Hardware: ROD Module Detection: FAILs" );
     status = FAIL;
   }
 }
 return( status );

}


int
Close_Hardware( ROD_Hardware *rod_hard ){
int status = SUCCESS;
word32 ii;

 if( rod_hard != NULL ){
   if( rod_hard->vme != NULL ){
     if( Close_RODVME( rod_hard->vme ) != SUCCESS ){
       printf("\n[ %-30s ]","ROD VMEHardware: Close: FAIL" );
       fflush( stdout  );
       status = FAIL;
     }
#ifdef VERBOSE
     else{
       printf("\n[ %-30s ]","ROD vme: Close: SUCESS" );
       fflush( stdout  );
     }
#endif
   }
   
   /* Module Release */
   if( rod_hard->module_set != NULL )
     free( rod_hard->module_set );
   /* Front-End Release */
   if( rod_hard->fe_set != NULL ){
     for( ii = 0; ii < rod_hard->fe_size; ii++ ){
       if( rod_hard->fe_set[ ii ].status == UNDEF )
	 continue;
       if( rod_hard->fe_set[ ii ].modules != NULL )
	 free( rod_hard->fe_set[ ii ].modules );
     }
     free( rod_hard->fe_set );
   }
   free( rod_hard );
 }
#ifdef VERBOSE
 printf("\n[ %-30s ]","ROD Hardware: Close: SUCESS" );
 fflush( stdout  );
#endif
 return( status );

}


int
SetUp_FrontEnd( ROD_Hardware* rod_hard, word32 info_ptr ){
Search_Key key;
word32 ii, jj, kk, *l_csr, obj_index, offset, addr;
FE_entry *fe_info;
char io_type[ 10 ];
Front_End* fe_set;
ROD_Module* mod_set;
int status = FAIL; 

 fe_info = &( ( ( ROD_info* )info_ptr )->fe ); 
 rod_hard->fe_size = fe_info->no_itm;

 /* Front-End Buffer Alloc ------------------------------------------------ */
 if( rod_hard->fe_set != NULL ){
   printf("\n[ %-30s ]", "Front End set: Already built" );
   return( FAIL );
 }
 rod_hard->fe_set = ( Front_End* )
   malloc( sizeof( Front_End )*fe_info->no_itm );
 if( rod_hard->fe_set == NULL ){
   printf("\n[ %-30s ]", "Not Enough Memory to build Front End set" );
   return( FAIL );
 }
 
 mod_set = rod_hard->module_set; /* Shrotcuts */
 fe_set  = rod_hard->fe_set;

 l_csr = mod_set[ 0 ].l_csr; /* Base Address of local CSR */
 /* Front-End Vector SetUp ------------------------------------------------ */
 offset = ( CSR_ROD_SIZE * rod_hard->mod_size );
 for( ii = 0 ; ii < rod_hard->fe_size ; ii++ ){
   l_csr[ offset + (CSR_FE_SIZE*ii) ] = ROD_DEAD;
   sprintf( fe_set[ ii ].ip_addr, "%s", fe_info->set[ ii ].ip_addr );
   sprintf( fe_set[ ii ].id, "%s", fe_info->set[ ii ].id );
   sprintf( key.machine, "%s", fe_info->set[ ii ].ip_addr );
   sprintf( key.dev_id, "%s", "CSR_Space" );
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     fe_set[ ii ].status = UNDEF;
     continue;
   }
   if( fe_set[ ii ].status == UNDEF )
     continue;

   fe_set[ ii ].r_csr = ( word32* )addr;
   fe_set[ ii ].l_csr = &( l_csr[ offset + (CSR_FE_SIZE * ii) ] );

   fe_set[ ii ].modules = 
     ( Mod_Link* )malloc( sizeof( Mod_Link )*fe_info->set[ ii ].no_itm );

   fe_set[ ii ].no_modules = fe_info->set[ ii ].no_itm;
   sprintf( fe_set[ ii ].id, "%s", fe_info->set[ ii ].id );
   /* Module List Processing */
   for( jj = 0 ; jj < fe_info->set[ ii ].no_itm ; jj++ ){
     sprintf( fe_set[ ii ].modules[ jj ].ip_addr, "%s", 
	      fe_info->set[ ii ].link[ jj ].ip_addr );
     strcpy( io_type, fe_info->set[ ii ].link[ jj ].io_type );
     if( strcmp( io_type, "Both" ) == 0 )
       fe_set[ ii ].modules[ jj ].io_type = __feBOTH__ ;
     else
       if( strcmp( io_type, "Slink" ) == 0 )
	 fe_set[ ii ].modules[ jj ].io_type = __feSLINK__ ;
       else
	 if( strcmp( io_type, "VME" ) == 0 )
	   fe_set[ ii ].modules[ jj ].io_type = __feVME__ ;
   }
 }
#ifdef VERBOSE
 printf("\n[ RODhardware: FESetUp: Trying To Make Fe-Mod Cross References ]");
#endif
 for( ii = 0; ii < rod_hard->mod_size ; ii++ ){
   status = FAIL;
   mod_set[ ii ].fe_ptr = -1;
   for( jj = 0; jj < rod_hard->fe_size; jj++ ){
     for( kk = 0; kk < fe_set[ jj ].no_modules ; kk++ ){
       if( strcmp( mod_set[ ii ].ip_addr, 
		   fe_set[ jj ].modules[ kk ].ip_addr ) == 0 ){
	 mod_set[ ii ].fe_ptr = jj;
	 status = SUCCESS;
#ifdef VERBOSE
	 printf("\n[%s --> %s: Linked]", mod_set[ ii ].id, fe_set[ jj ].id);
#endif
	 break;
       }
#ifdef VERBOSE
       printf("\n[ Mod-%s ]", fe_set[ jj ].modules[ kk ].ip_addr );
       printf("\n[ Fe-%s ]", mod_set[ jj ].ip_addr );
#endif
     }
     if( status == SUCCESS )
       break;
   }
   if( status == FAIL )
     printf("\n\n[ WARNING %s has not associated FE ]", mod_set[ ii ].id );
 }
 fflush( stdout );
 return( SUCCESS );

}




/*+ FrontEnd detection  
  07/October/1999 Reviewed to use info_ptr pointer +*/
int
Detect_FrontEnd( ROD_Hardware *rod_hard, word32 info_ptr ){
int status = SUCCESS;
word32 *l_csr, *r_csr, detected, ii, time_out;
Front_End* fe_set;

 if( SetUp_FrontEnd( rod_hard, info_ptr ) != SUCCESS )
   return( FAIL );

 /* --------------------------- Detection Protocol ----------------------- */
 fe_set  = rod_hard->fe_set; /* shorcut */

 for( ii = 0; ii < rod_hard->fe_size; ii++ ){
   if( fe_set[ ii ].status == UNDEF )
     continue;

   r_csr = fe_set[ ii ].r_csr;/* shorcut */
   l_csr = fe_set[ ii ].l_csr;/* shorcut */
   fe_set[ ii ].lives = LIVE_CHANCE;
#ifdef VERBOSE
   printf("\n\nTrying to Detect FrontEnd %s", fe_set[ ii ].id );
   fflush( stdout );
#endif
   detected = SET;
   time_out = BOOT_TIME_OUT;
   while( l_csr[ 0 ] == ROD_DEAD ){
     --time_out;
     printf("\n[ %-15s ]", "Waiting FrontEnd connection" );
     if( !time_out ){
       printf("\n[ FrontEnd %u is Dead ..... ]\n", ( unsigned )ii );
       detected = CLEAR;
       break;
     }
     sleep( 1 );
   }
   if( detected == SET ){
     /* Front End is Alive */
     printf("\n[ FrontEnd%u %s is UP ]", ( unsigned )ii, fe_set[ ii ].id );
     l_csr[ 0 ] = BUSY; 
     usleep( 100000 );
     r_csr[ 0 ] = CONTROLLER_ALIVEs;/* Awake FE ii --> ACK code */
     fe_set[ ii ].status = UP;
   }
   else{
     status = FAIL;
     fe_set[ ii ].status = DOWN;
   }
 }
 /* ----------------------------------------------------------------------- */
 return( SUCCESS );

}

int
SetUp_Modules( ROD_Hardware *rod_hard, word32 info_ptr ){
Search_Key key;
word32 ii, obj_index, addr;
M_entry *m_info;
ROD_Module* mod_set;

 m_info = &( ( ( ROD_info* )info_ptr )->mod );
 /* Retrieve virtual Pointers to VME layer 
    CSR-InputBuffer-OutPutBuffer ---------------------------- */
 sprintf( key.machine, "%s", "localhost" );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_csr = ( word32* )addr;
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_ibuf = ( word32* )addr;
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_obuf = ( word32* )addr;

 /* Build Module Vector ------------------------------------------------ */
 rod_hard->mod_size = m_info->no_itm;
 if( rod_hard->module_set != NULL ){
   printf("\n[ %-12s ]", "Module Vector already created" );
   return( FAIL );
 }
 rod_hard->module_set = ( ROD_Module* )
   malloc( sizeof( ROD_Module )*rod_hard->mod_size ); 
 if( rod_hard->module_set == NULL ){
   printf("\n[ %-30s ]", "Not Enough Memory to build Module set" );
   return( FAIL );
 }
 /* --------------------------------------------------------------------- */

 /* Module Vector SetUp ------------------------------------------------- */
 mod_set = rod_hard->module_set; /* Shortcut */
 for( ii = 0 ; ii < rod_hard->mod_size ; ii++ ){
   rod_hard->l_csr[ CSR_ROD_SIZE * ii  ] = ROD_DEAD;   
   /* CSR Space */
   sprintf( mod_set[ ii ].id, "%s", m_info->set[ ii ].id );
   sprintf( mod_set[ ii ].ip_addr, "%s", m_info->set[ ii ].ip_addr );

   sprintf( key.machine, "%s", m_info->set[ ii ].ip_addr );
   sprintf( key.machine, "%s", m_info->set[ ii ].ip_addr );
   sprintf( key.dev_id, "%s", "CSR_Space" );
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     mod_set[ ii ].status = UNDEF;
     continue;
   }
  /* Controller-RODmodule Map. The problem Arrives:
    Q: Which Offset is sampled by ROD_Module to interface with RODModule?
    A: FE_ROD_SIZE*/
   mod_set[ ii ].r_csr = &(( word32* )addr )[ CSR_FE_SIZE ];
   mod_set[ ii ].l_csr = &(rod_hard->l_csr[ CSR_ROD_SIZE * ii ]);

   /* OutPut Buffer */
   sprintf( key.dev_id, "%s", "Output_Buffer" );
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
       != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     mod_set[ ii ].status = UNDEF;
     continue;
   }
   mod_set[ ii ].r_obuf = ( word32* )addr;
   mod_set[ ii ].o_index = obj_index;
   /* Input Buffer */
   sprintf( key.dev_id, "%s", "Input_Buffer" );
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
       != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     mod_set[ ii ].status = UNDEF;
     continue;
   }
   mod_set[ ii ].r_ibuf = ( word32* )addr;
   mod_set[ ii ].i_index = obj_index;
   mod_set[ ii ].evt_size = CLEAR;
 }
 return( SUCCESS );

}


/*+ Modules Detection 
  07/October/1999 Reviewed to use info_ptr pointer +*/
int
Detect_Modules( ROD_Hardware *rod_hard, word32 info_ptr ){
word32 *l_csr, *r_csr, detected, time_out, ii;
ROD_Module* mod_set;

 if( SetUp_Modules( rod_hard, info_ptr ) != SUCCESS )
   return( FAIL );

 /* --------------------------- Detection Protocol ----------------------- */
 mod_set = rod_hard->module_set; /* Shotcut */

 for( ii = 0 ; ii < rod_hard->mod_size ; ii++ ){
   r_csr  = mod_set[ ii ].r_csr; /* Shorcut */
   l_csr  = mod_set[ ii ].l_csr; /* Shorcut */
   mod_set[ ii ].lives = LIVE_CHANCE;
#ifdef VERBOSE
   printf("\n\nTrying to Detect ROD %s", mod_set[ ii ].id );
   fflush( stdout );
#endif
   detected = SET;
   time_out = BOOT_TIME_OUT;
   while( l_csr[ 0 ] == ROD_DEAD ){
     --time_out;
     printf("\n[ %-15s ]", "Waiting ROD connection" );
     if( !time_out ){
       printf("\n[ Rod %u is Dead ..... ]\n", ( unsigned )ii );
       detected = CLEAR;
       break;
     }
     sleep( 1 );
   }
   if( detected == SET ){
     printf("\n[ ROD%u %s is UP ]", 
	    ( unsigned )ii, mod_set[ ii ].id );
     l_csr[ 0 ] = BUSY; /* Still Alive ---> ROD ii is Alive */
     usleep( 100000 );
     r_csr[ 0 ] = CONTROLLER_ALIVEs;/* Awake ROD ii Code Ack. last loop */
     mod_set[ ii ].status = UP;
   }
   else{
     mod_set[ ii ].status = DOWN;
     printf("\n[ ROD%u %s is marked as DOWN ]", 
	    ( unsigned )ii, mod_set[ ii ].id );
   }
 }
 /* ---------------------------------------------------------------------- */
 return( SUCCESS );

}



/*+ To Get Module OutPut Event +*/
int
Module_Get_Event( Event_Packet *e_pack  ){
word32 evt_size, idx, obj_index;
int status = FAIL;
DMA_Dev *dma_dev;
ROD_Hardware *rod_hard;

 rod_hard = e_pack->rod_hard;
 idx = e_pack->module;
 evt_size = rod_hard->module_set[ e_pack->module ].evt_size;
 obj_index = rod_hard->module_set[ e_pack->module ].o_index;
 dma_dev = ( DMA_Dev* )( rod_hard->vme->rod_dev[ obj_index ].vme_obj );
 if( DMA_Go( dma_dev, VMEtoLOCAL, evt_size * sizeof( word32 ) ) != SUCCESS ){
   printf("\n[ %-30s ]", "DMAGo FAILs" );
   fflush( stdout );
   sprintf( e_pack->stat, "%s", "DMA: FAILs" );
 }
 else{
   sprintf( e_pack->stat, "%s", "DMA: SUCESS" );
   status = SUCCESS;
 }
 return( status );

}


/*+ To Get Module State CSR +*/
int
Module_Get_CSR( CSR_Packet *csr_pack ){
int status = FAIL;
ROD_Module *module;
word32 idx ;

 module = csr_pack->rod_hard->module_set; /* shorcut */
 idx   = csr_pack->module;
 if( Device_Check( csr_pack->rod_hard, idx, 
		   __Module__, csr_pack->stat ) != SUCCESS )
   return( status );
 memcpy( csr_pack->buffer, module[ idx ].r_csr, CSR_ROD_SIZE*sizeof( word32 ));
 status = SUCCESS;
 return( status );

}


/*+ To Download the ConfigurationPacket to the ROD-Module +*/
int
Module_Set_Config( CFG_Command *cfg_cmd ){
int status = SUCCESS;
ROD_Module *module;
word32 jj;
volatile word32 *r_csr;
Config_Word cfg; 
Command_pack cmd_pack;

 module = cfg_cmd->rod_hard->module_set; /* shorcut */
 cmd_pack.l_csr  = module[ cfg_cmd->module ].l_csr;
 cmd_pack.r_csr  = module[ cfg_cmd->module ].r_csr;
 cmd_pack.lives  = &( module[ cfg_cmd->module ].lives  );
 cmd_pack.status = &( module[ cfg_cmd->module ].status );
 cmd_pack.cmd    = CONFIG;
 cmd_pack.stat   = cfg_cmd->stat;
 r_csr = module[ cfg_cmd->module ].r_csr;
 if( Device_Check( cfg_cmd->rod_hard, cfg_cmd->module, __Module__, 
		   cfg_cmd->stat ) != SUCCESS ){
   printf("\n[ %-20s ]", cfg_cmd->stat );
   return( FAIL ); 
 } 
 for( jj = 0 ; jj < cfg_cmd->data[ 0 ] ; jj ++ ){
   r_csr[ __cfgOffset__ +jj ] = cfg_cmd->data[ jj+1 ];
 }
 
 if( Send_Command( &cmd_pack ) != SUCCESS ){
   status = FAIL;
 }
 cfg.all = cfg_cmd->data[ 1 ];
 module[ cfg_cmd->module ].evt_size = cfg.bit.no_channels*cfg.bit.no_samples;
 return( status );

}

/*+ To Download the ConfigurationPacket to the ROD-Module +*/
int
FEnd_Set_Config( CFG_Command *cfg_cmd ){
int status = SUCCESS;
Front_End  *fe;
ROD_Module *mod;
word32 jj, fe_idx;
volatile word32 *r_csr;
Config_Word cfg;
Command_pack cmd_pack;

 fe =  cfg_cmd->rod_hard->fe_set; /* shorcut */ 
 mod = cfg_cmd->rod_hard->module_set; /* shorcut */
 fe_idx = mod[ cfg_cmd->module ].fe_ptr;

 for( jj = 0; jj < fe[ fe_idx ].no_modules; jj++ ){
   if( strcmp( fe[ fe_idx ].modules[ jj ].ip_addr, 
	       mod[ cfg_cmd->module ].ip_addr ) == 0 ){
     cfg_cmd->data[ 4 ] = jj;
     break;
   }
 }

 cmd_pack.l_csr  = fe[ fe_idx ].l_csr;
 cmd_pack.r_csr  = fe[ fe_idx ].r_csr;
 cmd_pack.lives  = &( fe[ fe_idx ].lives  );
 cmd_pack.status = &( fe[ fe_idx ].status );
 cmd_pack.cmd    = CONFIG;
 cmd_pack.stat   = cfg_cmd->stat;
 if( Device_Check( cfg_cmd->rod_hard, fe_idx, __Module__, 
		   cfg_cmd->stat ) != SUCCESS ){
   printf("\n[ %-20s ]", cfg_cmd->stat );
   return( FAIL ); 
 }
 r_csr = fe[ fe_idx ].r_csr;
 for( jj = 0 ; jj < cfg_cmd->data[ 0 ] ; jj ++ ){
   r_csr[ __cfgOffset__ +jj ] = cfg_cmd->data[ jj+1 ];
 }
 if( Send_Command( &cmd_pack ) != SUCCESS ){
   status = FAIL;
 }
 
 cfg.all = cfg_cmd->data[ 1 ];
 fe[ fe_idx ].evt_size = cfg.bit.no_channels*cfg.bit.no_samples;
 return( status );

}


/*+ Generic Primitives to Modules And FrontEnds +*/
int
Send_Command( Command_pack *cmd_pack ){
char  *str[ ]={"RUN" ,"PAUSE" ,"CONFIG" ,"CONTINUE" ,"SHUTDOWN" , 
	       "SAMPLE", "STOP" };
word32 time_out;
int status = FAIL;

  cmd_pack->l_csr[ 0 ] = MASK_COMMAND;
  cmd_pack->r_csr[ 0 ] = cmd_pack->cmd;
  time_out = COMMAND_TIME_OUT; 
  while( cmd_pack->l_csr[ 0 ] == MASK_COMMAND ){
    --time_out;
    cmd_pack->r_csr[ 0 ] = cmd_pack->cmd;
    if( time_out == 0 ){
      sprintf( cmd_pack->stat, "%s: Command: %s: Not ACK", 
	       cmd_pack->stat, str[ cmd_pack->cmd - 4 ] );
     --( *( cmd_pack->lives ) );
     if( ( *( cmd_pack->lives ) ) == 0 ){
       sprintf( cmd_pack->stat, "%s-%s", cmd_pack->stat, "Marked as: DOWN" );
       *( cmd_pack->status )  = DOWN;
     }
     break;
   }
   usleep( 10000 );
 }

 if( time_out != 0 ){
   *( cmd_pack->lives ) = LIVE_CHANCE;
   if( cmd_pack->l_csr[ 0 ] == CMD_RECEIVED ){
     status = SUCCESS;
     sprintf( cmd_pack->stat, "%s: Command: %s: Done", 
	      cmd_pack->stat, str[ cmd_pack->cmd - 4 ] );
   }
   else
     sprintf( cmd_pack->stat, "%s: Command: %s: Ilegal", 
	      cmd_pack->stat, str[ cmd_pack->cmd - 4 ] );

   if( cmd_pack->l_csr[ 1 ] != cmd_pack->cmd ){
     status = FAIL;
     switch( cmd_pack->l_csr[ 1 ] ){
     case ACK_1:
       printf("\n[WARNING][Data-Injection: NOT Configured]");
       break;
     case ACK_2:
       printf("\n[WARNING][Data-Injection: Already Running]");
       break;
     case ACK_3:
       printf("\n[WARNING][Data-Injection: Down]");
       break;
     case ACK_4:
       printf("\n[WARNING][Data-Injection: NOT Running]");
       break;
     case ACK_5:
       printf("\n[WARNING][ROD-Machine: Internal FAIL]");
       break;
     case ACK_6:
       printf("\n[WARNING][ROD-Machine: Invalid Configuration Value ]");
       break;
     case ACK_7:
       printf("\n[WARNING][ROD-Machine: Running: STOP/PAUSE Before]");
       break;
       
     }
   }
 }
 if( cmd_pack->cmd == SHUTDOWN )
   *( cmd_pack->status ) = DOWN;

 return( status );

}


int
Device_Check( ROD_Hardware *rod_hw, word32 item, word32 device ,char *stat  ){
int status = FAIL, dev_stat ;
Front_End  *fe;
ROD_Module *module;
word32 size;

 if( device == __Module__ ){
   module = rod_hw->module_set; /* shorcut */
   if( module == NULL ){
     sprintf( stat, "%s", "Module Set Not Initialised" );
     return( status );
   }
   size = rod_hw->mod_size;
   dev_stat =  module[ item ].status;
 }
 else{
   fe = rod_hw->fe_set; /* shorcut */
   if( fe == NULL ){
     sprintf( stat, "%s", "Set Not Initialised" );
     return( status );
   }
   size = rod_hw->fe_size;
   dev_stat = fe[ item ].status;
 }

 if( ( item + 1 ) > size ){
   sprintf( stat, "%s", "Does not Exist" );
   return( status ); 
 }
 if( dev_stat != UP ){
   sprintf( stat, "%s", "Down" );
   return( status );
 }
 status = SUCCESS;
 return( status );

  
}
