/**
   $Header: RODSoftware object for the ROD-Controlller Application
   Created & Maintaned by JB. Juanba.Romance@ific.uv.es $   
 
   01/June/99:
   Synchronous VME Data Injection/Monitoring Implementation Starts 
   
   09/December/99:
   Data Injection Syncronization reviewed using Module-Machine implementation
   experience. Objective: To improve Fault Tolerance for race conditions
   Still Only implemented Push Flow on the Module.
**/

#include "RODsoftware.h"

#define __DEBUGSoftware__
#define __ContinueDelay__ 2000000
#define __InyectionDelay__ 300



int
Open_Software( ROD_Software **rod_soft, word32 stat_addr ){
int status = SUCCESS;

 *rod_soft = ( ROD_Software* )malloc( sizeof( ROD_Software ) );
 if( *rod_soft == NULL ){
   printf( "\n[ %-30s ]", "Software: Malloc: Fail");
   fflush( stdout );
   return( FAIL );
 }
 ( *rod_soft )->stat = ( word32* )stat_addr;
 *( ( *rod_soft )->stat ) = __Down__ ;
 return( status );

}


int
Close_Software( ROD_Software *rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   if( *( rod_soft->stat ) != __Down__ ){
     pthread_cancel( rod_soft->pro.phy_th );
     pthread_cancel( rod_soft->in.phy_th );
     pthread_cancel( rod_soft->out.phy_th );
     pthread_cancel( rod_soft->mon.phy_th );
     
     pthread_mutex_destroy ( &rod_soft->mon.io_mtx );
     pthread_mutex_destroy ( &rod_soft->pro.io_mtx );
     pthread_cond_destroy  ( &rod_soft->mon.io_cv  );
     pthread_cond_destroy  ( &rod_soft->pro.io_cv  );
     *( rod_soft->stat ) = __Down__ ;
   
     free( rod_soft->out.evt_buf );
     free( rod_soft->in.evt_buf );   
   }
   free( rod_soft );
 }
 return( status );

}

/* 02/June/99
   Pointers routing to I/O Data Threads, io_data Pointer format :
   [ 0 ][ 1 ]Virtual Addresses to local_csr, remote_csr
   [ 2 ][ 3 ]Idem for Local Buffers: Input, Output. ( Driver Mode )
   [ 4 ][ 5 ]Idem for DMA Objects: Input, Output ( Target Mode )
   [ 6 ][ 7 ]Virtual Pointers of Above Buffers.
   [ 8 ][ 9 ]Module Configuration Words: Input, Output
   Last items will route the I/O threads with the selected VME objects.
   */
int
Run_Software( ROD_Software *rod_soft, word32 *io_data ){
int status = SUCCESS;
Config_Word cfg_w;

pthread_attr_t  pth_attr;
pthread_t *pth_list[ 4 ];
void      *fun_list[ 4 ];
void      *par_list[ 4 ];
word32 ii;

 if( rod_soft == NULL ){
   printf("[ %-20s ]", "Oooooooops: ROD: Software: NULL ptr" );
   return( FAIL );
 }

 /* Input Analysis */
 cfg_w.all = io_data[ 8 ];
 if( cfg_w.bit.io_mode == __Driver__ ){
   printf("\n[ Input Stage: Driver Mode ]" );
   rod_soft->in.dma_obj = 0xDeadBeef;
   rod_soft->in.buffer  = ( word32* )io_data[ 2 ];
 }
 else{
   rod_soft->in.dma_obj = io_data[ 4 ];
   rod_soft->in.buffer  = ( word32* )io_data[ 6 ];
 }   
 rod_soft->in.l_csr = ( word32* )io_data[ 0 ];
 rod_soft->in.r_csr = ( word32* )io_data[ 1 ];
 rod_soft->in.cfg   = io_data[ 8 ];

 /* Counters Locally Driven by Controller */
 rod_soft->in.l_csr[ __IUHandshake__ +6 ] = CLEAR;
 rod_soft->in.l_csr[ __IUHandshake__ +5 ] = CLEAR;
 rod_soft->in.l_csr[ __IUHandshake__ +4 ] = CLEAR; 
 rod_soft->in.l_csr[ __IUHandshake__ +7 ] = CLEAR; 
 rod_soft->in.evt_buf  = ( word32* )malloc( sizeof( word32 )*4096 );

 /* Output Analysis */
 cfg_w.all = io_data[ 9 ];
 if( cfg_w.bit.io_mode == __Driver__ ){
   printf("\n[ OutPut Stage: Driver Mode ]" );
   rod_soft->in.dma_obj = 0xDeadBeef;
   rod_soft->out.buffer = ( word32* )io_data[ 3 ];
 }
 else{
   rod_soft->out.dma_obj = io_data[ 5 ];
   rod_soft->out.buffer  = ( word32* )io_data[ 7 ];
 }
 rod_soft->out.l_csr   = ( word32* )io_data[ 0 ];
 rod_soft->out.r_csr   = ( word32* )io_data[ 1 ];
 rod_soft->out.cfg     = io_data[ 9 ];
 rod_soft->mon.evt_buf = ( word32* )malloc( sizeof( word32 )*4096 );
 

 /* Threads SetUp */
 sprintf( rod_soft->in.io_id, "%s", "Input Unit" );
 sprintf( rod_soft->out.io_id, "%s", "OutPut Unit" );
 
 rod_soft->in.stat  = rod_soft->stat;
 rod_soft->out.stat = rod_soft->stat;
 rod_soft->pro.stat = rod_soft->stat;
 rod_soft->mon.stat = rod_soft->stat;
 
#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "IU Thread Attr" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "IU Thread Attr" );
   return( status );
 }
#endif
 /* Syncronization Tools Routing */
 rod_soft->pro.io_ready = &( rod_soft->in.io_ready );
 rod_soft->mon.io_ready = &( rod_soft->out.io_ready );
 rod_soft->in.ready     = &( rod_soft->pro.ready );
 rod_soft->out.ready    = &( rod_soft->mon.ready );
 rod_soft->in.io_ready  = CLEAR;
 rod_soft->out.io_ready = CLEAR;
 rod_soft->pro.ready    = CLEAR;
 rod_soft->mon.ready    = CLEAR;

 pthread_mutex_init ( &rod_soft->mon.io_mtx, NULL );
 pthread_mutex_init ( &rod_soft->pro.io_mtx, NULL );
 pthread_cond_init  ( &rod_soft->mon.io_cv, NULL  );
 pthread_cond_init  ( &rod_soft->pro.io_cv, NULL  );
 rod_soft->in.io_mtx  = &rod_soft->pro.io_mtx;
 rod_soft->in.io_cv   = &rod_soft->pro.io_cv;
 rod_soft->out.io_mtx = &rod_soft->mon.io_mtx;
 rod_soft->out.io_cv  = &rod_soft->mon.io_cv;
 

 rod_soft->pro.evt_buf = rod_soft->in.evt_buf; 
 rod_soft->out.evt_buf = rod_soft->mon.evt_buf;
 for( ii = 0; ii < 1024 ; ii++ ){
   rod_soft->in.evt_buf[ ii ] = ( 1 << ( ii % 32 ) );
 }
 
 pth_list[ 0 ] = &rod_soft->pro.phy_th;
 pth_list[ 1 ] = &rod_soft->mon.phy_th;
 pth_list[ 2 ] = &rod_soft->out.phy_th;
 pth_list[ 3 ] = &rod_soft->in.phy_th;
 
 fun_list[ 0 ] = ( void *)Production_Engine;
 fun_list[ 1 ] = ( void *)Monitor_Engine;
 fun_list[ 2 ] = ( void *)Output_Engine;
 fun_list[ 3 ] = ( void *)Input_Engine;
 
 par_list[ 0 ] = ( void* )( &rod_soft->pro );
 par_list[ 1 ] = ( void* )( &rod_soft->mon );
 par_list[ 2 ] = ( void* )( &rod_soft->out );
 par_list[ 3 ] = ( void* )( &rod_soft->in );
 
 for( ii = 0; ii < 4; ii++ ){
#if defined( LinuxPCBit3 )
   if( pthread_create( pth_list[ ii ],
		       &pth_attr, 
		       fun_list[ ii ], 
		       par_list[ ii ] ) != 0 ){
     perror( "Thread Create" );
     *( rod_soft->stat ) = __Down__ ;
     break;
   }
#else
   if( pthread_create( pth_list[ ii ],
		       pth_attr, 
		       fun_list[ ii ], 
		       par_list[ ii ] ) != 0 ){
     perror( "Thread Create" );
     *( rod_soft->stat ) = __Down__ ;
     break;
   }
#endif
   else
     *( rod_soft->stat ) = __Up__ ;
 } 
 if( *rod_soft->stat == __Down__ )
   status = FAIL;
 return( status );
 

}

int 
Pause_Software( ROD_Software *rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   *( rod_soft->stat ) = __Pause__ ;
 } 
 return( status );

}

int
Continue_Software( ROD_Software *rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   *( rod_soft->stat ) = __Continue__ ;
 } 
 return( status );

}


/* -------------------------------- VME Data Injection --------------------- */
int
Input_Engine( void *par ){
IO_Type   *in;
word32 count = CLEAR, evt_size;
Config_Word cfg_w;

 printf("\n[ %-20s ]", "Input_Engine Starts" );
 fflush( stdout );
 in = ( IO_Type* )par; 
 *( in->stat ) = __Run__ ;

 cfg_w.all = in->cfg;
 printf("\n[ %-20s ][%s]", 
	"Input Mode", cfg_w.bit.io_mode ? "Driver" : "Target" );
 printf("\n[ %-20s ][%d][%d]",
	"Channels/Samples", 
	( unsigned )cfg_w.bit.no_channels, 
	( unsigned )cfg_w.bit.no_samples );
 fflush( stdout );
 evt_size = cfg_w.bit.no_channels * cfg_w.bit.no_samples;
 
 while( *( in->stat ) != __Down__  ){
   if( *( in->stat ) == __Pause__ ){
     printf("\n[ %-20s ][ %-8s ]", "Input_Engine","PAUSED" );
     fflush( stdout );
     usleep( __ContinueDelay__ );
   }
   if( (*( in->stat ) == __Run__ )||( *( in->stat ) == __Continue__ )){
     /* Production Thread Synchronism --------------------------------- */
     pthread_mutex_lock( in->io_mtx );
     if( *( in->ready ) == CLEAR ){
       in->io_ready = SET;
       pthread_cond_wait( in->io_cv, in->io_mtx );
       in->io_ready = CLEAR;
     }
     memcpy( in->buffer, in->evt_buf, evt_size*sizeof( word32 ) );
     pthread_cond_signal( in->io_cv );
     pthread_mutex_unlock( in->io_mtx );
     /* --------------------------------------------------------------- */
     if( In_Module_Handshake( in, &count ) != SUCCESS ){
       if( in->l_csr[ __IUHandshake__ + 1 ] > __DataSRCFail__ ){
	 *( in->stat ) = __Pause__;
       }
     }
     ++count;
   }
 }
 printf("\nROD: Software: Input Engine: ENDs");
 return( SUCCESS );

}


int
Production_Engine( void *par ){
Productor *pro;
word32 ii, count = CLEAR;

 printf("\n[ %-20s ]", "Production_Engine Starts" );
 fflush( stdout );
 pro = ( Productor* )par;
 *( pro->stat )= __Run__ ;

 while( *( pro->stat ) != __Down__ ){
   if( *( pro->stat ) == __Pause__ ){
     printf("\n[ %-20s ][ %-8s ]", "Data_Pro_Engine", "PAUSED" );
     fflush( stdout );
     usleep( __ContinueDelay__ );
   }
   if(( *( pro->stat ) == __Run__ )||( *( pro->stat ) == __Continue__ )){
     /* IO Thread Synchronism ---------------------------------------- */
     pthread_mutex_lock( &pro->io_mtx );
     if( *( pro->io_ready ) == CLEAR ){
       pro->ready = SET;
       pthread_cond_wait( &pro->io_cv, &pro->io_mtx );
       pro->ready = CLEAR;
     }
     /* Insert your own code here ------------------------------------ */


     for( ii = 0 ; ii < 1024 ; ii++ )
       pro->evt_buf[ ii ] = ~( 1 << ( ii % 32 )) ;


     /* -------------------------------------------------------------- */

     pthread_cond_signal( &pro->io_cv );
     pthread_mutex_unlock( &pro->io_mtx );
     /* --------------------------------------------------------------- */
     ++count;
   }
 }
 printf("\nROD: Software: Data-Productor Engine: ENDs");
 return( SUCCESS );
}


int
In_Module_Handshake( IO_Type *in, word32 *count ){
#define __InCTRLTimeOut__ 1000000
int status = SUCCESS;
volatile word32 *l_csr, *r_csr;
word32 time_out, size ;
Config_Word cfg_w;
DMA_Dev   *dma_dev;


 l_csr = in->l_csr;
 r_csr = in->r_csr;
 cfg_w.all = in->cfg;

 if( !( ( *count )%10000 ) ){
   printf("\n[ Trigger State ]\n[ Hits %04d ]", ( unsigned )( *count ) );
   printf("\n[ Local  Trials   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +6 ]);
   printf("\n[ Local  Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +5 ]);
   printf("\n[ Local  Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +4 ]);
   printf("\n[ Module Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +2 ]);
   printf("\n[ Module Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +1 ]);
   printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]", 
	            ( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
   printf("\n[ %-20s ][%s]", 
	  "VME Input DMA Mode",  cfg_w.bit.io_mode ? "Driver" : "Target" );
   fflush( stdout );
 }

 /* ----------------------------------- Trigger-On ------------------------- */
 l_csr[ __IUHandshake__ +6 ] = *count;
 time_out = __InCTRLTimeOut__ ;
 while( l_csr[ __IUHandshake__ +3 ] != *count ){
   --time_out;
   if( time_out == 0 ){
     ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 10 ) ){
	printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]", 
	( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
	fflush( stdout );
     }
      *count = l_csr[ __IUHandshake__ +3 ];
     return( FAIL );
   }
 }

 if( cfg_w.bit.io_mode == __Target__ ){
   size = cfg_w.bit.no_samples * cfg_w.bit.no_channels ;
   dma_dev = ( DMA_Dev* )in->dma_obj;
   if( DMA_Go( dma_dev, LOCALtoVME, size * sizeof( word32 ) ) != SUCCESS ){
     printf("\n[ ROD: Software: DMA to Data Source: Fails ]" );
     fflush( stdout );
     return( FAIL );
   }
 }

 /* ------------------------------- Trigger-Off ---------------------------- */
 time_out = __InCTRLTimeOut__ ;
 while( l_csr[ __IUHandshake__ ] != __TriggerOff__ ){
   --time_out;
   if( time_out == 0 ){
     ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 100 ) ){
       printf("\n[ ROD: Software: TriggerOff Lost ][%d]", (unsigned)(*count));
       fflush( stdout );
     }
     l_csr[ __IUHandshake__ ] = MASK_COMMAND;
     return( FAIL );
   }
 }
 l_csr[ __IUHandshake__ ] = MASK_COMMAND;

 r_csr[ __IUHandshake__  ] = __TriggerOff__ ;
 ++l_csr[ __IUHandshake__ + 5 ];
 return( status );
 
}




/* ------------------------- Data Module Gathering ------------------------- */
int
Output_Engine( void *par ){
IO_Type   *out;

 printf("\n[ %-20s ]", "OutPut_Engine Starts" );
 fflush( stdout );
 out = ( IO_Type* )par; 
 *( out->stat ) = __Run__ ;

 while( *( out->stat ) != __Down__ ){  
   if( *( out->stat ) == __Pause__ ){
     printf("\n[ %-20s ][ %-8s ]", "Output_Engine","PAUSED" );
     fflush( stdout );
     usleep( __ContinueDelay__ );
   }
   if(( *( out->stat ) == __Run__ )||( *( out->stat ) == __Continue__ )){ 
     usleep( __InyectionDelay__ );
   }
 }
 printf("\nROD: Software: OutPut Engine: ENDs"); 
 return( SUCCESS );

}


int
Monitor_Engine( void* par ){
Monitor   *mon;

 printf("\n[ %-20s ]", "Monitor_Engine Starts" );
 fflush( stdout );
 mon = ( Monitor* )par;
 *( mon->stat ) = __Run__ ;

 while( *( mon->stat ) != __Down__ ){
   if( *( mon->stat ) == __Pause__ ){
     printf("\n[ %-20s ][ %-8s ]", "Monitor_Engine", "PAUSED" );
     fflush( stdout );
     usleep( __ContinueDelay__ );
   }
   if(( *( mon->stat ) == __Run__ )||( *( mon->stat ) == __Continue__ )){
     usleep( __InyectionDelay__ );
   }
 }
 printf("\nROD: Software: Monitor Engine: ENDs");
 return( SUCCESS );

}
/* ------------------------------------------------------------------------- */
