/**
    $Header: TileCal ROD Front End Application --> ROD.c
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $

    11/October/99:
    Front-End Implementation Starts
    
    10/December/99:
    Front-End Configuration Decoding Starts
    
    11/December/99
    Data Injection starts.
**/

#include "ROD.h"
ROD* sh_rod;
Handler_Filter h_filter;


/*++++++++++++++++++++++++++++++++++++++
  Rod Handler process OS signals from BusError, Ctrl-C And Segmentation Faults

  int sig_num Processed Signal
  ++++++++++++++++++++++++++++++++++++++*/
void
ROD_Handler( int sig_num ){

  printf("\n\n[ FrontEnd-SHandler : Received Signal %d ]", sig_num ); 
  switch( sig_num ){
    case SIGBUS  :printf("[ Bus Error ]" );break;
    case SIGINT  :printf("[ Ctrl-C    ]" );break;
    case SIGSEGV :printf("[ Segmentation Fault ]");break;
  }
  sleep( 1 );
  if( sh_rod != NULL ){
    if( h_filter.bit.f_3 == CLEAR ){
      h_filter.bit.f_3 = SET;
      if( sh_rod->software != NULL ){
	Close_Software( sh_rod->software );
	sh_rod->software = NULL;
      }
    }
    if( h_filter.bit.f_2 == CLEAR ){
      h_filter.bit.f_2 = SET;
      if( sh_rod->hardware != NULL ){
	Close_Hardware( sh_rod->hardware );
	sh_rod->hardware = NULL;
      }
    }
  
    if( h_filter.bit.f_4 == CLEAR ){
      h_filter.bit.f_4 = SET;
      free( sh_rod );
      sh_rod = NULL;
    }
  }
  printf("\nFront-End: Signal Handler Output\n");
  exit( FAIL );

}


/*++++++++++++++++++++++++++++++++++++++
  Handler Install

  Warning: If Install Fails Application Ends
  ++++++++++++++++++++++++++++++++++++++*/
void
Install_Handler( ){
struct sigaction rod_sa;
sigset_t rod_set;

  sigemptyset( &rod_set );
  sigaddset( &rod_set, SIGINT );
  sigprocmask( SIG_BLOCK, &rod_set, NULL ); /* Block CTRLc only on SetUp */ 
  sigfillset( &rod_sa.sa_mask );
  rod_sa.sa_flags = CLEAR; 
  rod_sa.sa_handler = ROD_Handler;
  if( sigaction( SIGINT, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGSEGV, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGBUS, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  sigprocmask( SIG_UNBLOCK, &rod_set, NULL ); /* Now Ready to Proccess */

}


/*++++++++++++++++++++++++++++++++++++++
  Create memory for the MacroStructure ROD 

  **rod: Address of ROD pointer
  ++++++++++++++++++++++++++++++++++++++*/
int
ROD_Open( ROD **rod ){ 
word32 stat_addr;

 Install_Handler( );
 h_filter.all = CLEAR;
 *rod = ( ROD* )malloc( sizeof( ROD ) );
 if( *rod == NULL ){
   printf("\n[ Not Enough Memory to Alloc ROD ]");
   fflush( stdout );
   return( FAIL );
 }
 sh_rod = *rod;

 /* RODMap.cfg processing -------------------------------------------------- */
 if( Make_FE( &( ( *rod )->info.fe  ) ) == FAIL )
   return( FAIL );
 if( Make_Ctrl ( &( ( *rod )->info.ctrl ) ) == FAIL )
   return( FAIL );
 if( Make_Module ( &( ( *rod )->info.fe )  ) == FAIL )
   return( FAIL );

 /* Still Alive? Open Hardware And Software Resources ---------------------- */
 ( *rod )->software = NULL;
 ( *rod )->hardware = NULL;
 stat_addr = ( word32 )( & ( ( *rod )->mi_stat ) );
 if( Open_Hardware( &( ( *rod )->hardware ), 
		    ( word32 )( &( ( *rod )->info ) )) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Hardware: Open: FAIL" );
   fflush( stdout );
   return( FAIL );
 }
 if( Open_Software( &( ( *rod )->software ), stat_addr ) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Software: Open: FAIL" );
   fflush( stdout );
   return( FAIL );
 }
#ifdef VERBOSE
 else{
   printf("\n[ %-30s ]", "ROD: Software: Open: SUCESS" );
   fflush( stdout );
 }
#endif

 ( *rod )->list = ( MOD_Config* )
   malloc( sizeof( MOD_Config )*( *rod )->info.fe.no_itm );
 if( ( *rod )->list == NULL ){
   printf("\n[ %-30s ]", "ROD: Alloc Config List: FAIL" );
   fflush( stdout );
   return( FAIL );
 }

 return( SUCCESS );

}


/*+ ROD-Module Entry Processing +*/
int
Make_Module( FE_entry *fe ){
word32 done = CLEAR, no_itm, ii, jj;
FILE* file;
char entry[ 80 ], id[ 20 ], ip_target[ 20 ], ip_src[ 20 ];

 file = fopen( RODMap, "r" );
#ifdef VERBOSE
 printf("\n[ Searching Module_Section.. ]");
 fflush( stdout );
#endif
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Module_Section:%d", &no_itm ) == 1 ){
     fe->crate_mod_size = no_itm;

     for( jj = 0; jj < no_itm ; jj++ ){
       if( fgets( entry, 80, file ) == NULL )
	 break;
       if( sscanf( entry, "%s\t%s\t%s", id, ip_src, ip_target ) != 3 ){
	 fclose( file );
	 return( FAIL );
       }
       for( ii = 0; ii < fe->no_itm ; ii++ ){
	 if( !strcmp( ip_src, fe->link[ ii ].ip_addr ) )
	   sprintf( fe->link[ ii ].link_id, id );
       }
     }
#ifdef VERBOSE
     printf("\n[%d RODs linked to %s]", no_itm, fe->id );
     fflush( stdout );
#endif
     done = SET;
     break;
   }
 }
 fclose( file );
 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );

}


/*+ ROD-Controller Entry Processing +*/
int
Make_Ctrl( C_entry *c_entry ){
FILE* file;
char entry[ 80 ];
word32 dummy;
int done = CLEAR;

 file = fopen( RODMap, "r" );
#ifdef VERBOSE
 printf("\n[ Searching Controller_Section... ]");
 fflush( stdout );
#endif
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Controller_Section:%d", &dummy ) == 1 ){
     if( fgets( entry, 80, file ) == NULL )
       break;
     else{
       strcpy( c_entry->id, strtok( entry, "\t" ) );
       strcpy( c_entry->ip_addr, strtok( NULL, "\n" ) );
       done = SET;
       break;
     }
   }
 }
 fclose( file );

 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );      
  
}


/*+ Front-End Entry Processing +*/
int
Make_FE(FE_entry *fe ){
word32 ii, jj, done = CLEAR, no_itm;
FILE* file;
char entry[ 80 ];

 file = fopen( RODMap, "r" );
#ifdef VERBOSE
 printf("\n[ Searching Front_End_Section... ]");
 fflush( stdout );
#endif
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Front_End_Section:%d", &no_itm ) == 1 ){
     for( ii = 0; ii < no_itm ; ii++ ){
       if( fgets( entry, 80, file ) == NULL )
	   break;
       strcpy( fe->id, strtok( entry, "\t" ) );
       strcpy( fe->ip_addr, strtok( NULL, "\t" ));
       if( !strcmp( fe->ip_addr, __MySelf__ ) ){
#ifdef VERBOSE
	printf("\n[FrontEnd %s: Detected]",  __MySelf__ );
	fflush( stdout );
#endif
	fe->idx = ii;   
	fe->no_itm = atoi( strtok( NULL, "\t" ) );     
	for( jj = 0 ; jj < fe->no_itm - 1 ; jj++ ){
	  strcpy( fe->link[ jj ].io_type, strtok( NULL, "\t" ) );
	  strcpy( fe->link[ jj ].ip_addr, strtok( NULL, "\t" ) );
	}
	strcpy( fe->link[ jj ].io_type, strtok( NULL, "\t" ) );
	strcpy( fe->link[ jj ].ip_addr, strtok( NULL, "\n" ) );
	done = SET;
	break;
       }
     }
   }
 }
 fclose( file );
 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );

}


/*+ Destroys ROD object +*/
int
ROD_Close( ROD* rod ){
int status = SUCCESS;
 
 if( rod != NULL ){ 

   if( rod->software != NULL ){
     if( Close_Software( rod->software ) != SUCCESS ){
       printf("\n[ %-30s ]", "ROD: Software: Close: FAIL" );
       status = FAIL;
     }
#ifdef VERBOSE
     else{
       printf("\n[ %-30s ]", "ROD: Software: Close: SUCESS" ); 
       fflush( stdout );
     }
#endif
     rod->software = NULL;
   }
 
   if( rod->hardware != NULL ){
     if( Close_Hardware( rod->hardware ) != SUCCESS ){
       printf("\n[ %-30s ]", "ROD: Hardware: Close: FAIL" );
       status = FAIL;
     }
#ifdef VERBOSE
     else{
       printf("\n[ %-30s ]", "ROD: Hardware: Close: SUCESS" );
       fflush( stdout );
     }
#endif
     rod->hardware = NULL;
   }
   fflush( stdout );

   if( rod->list !=  NULL ){
     free( rod->list );
     rod->list = NULL;
   }
   free( rod );
   rod = NULL;
 }
 return( status );
 
}

int
ROD_Run( ROD* rod ){
int status = FAIL;

 if( rod == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops Null Pointer" );
   fflush( stdout );
   return( status );
 }
 if( Run_Software( rod->software ) != SUCCESS ){
   printf("\n[ %-30s ]","ROD: Software: DataInjection FireUp: FAILs" );
   fflush( stdout );
   return( status );
 }
 status= SUCCESS;
 return( status );

}

int
ROD_Stop( ROD* rod ){
int status = FAIL;

 if( rod == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops Null Pointer" );
   fflush( stdout );
   return( status );
 }
 if( rod->software == NULL ){
   printf("\n[ %-30s ]","ROD: Software: DataInjection is DOWN" );
   return( FAIL );
 }

 if( Close_Software( rod->software ) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Software: DataInjection Close: FAIL" );
   status = FAIL;
 }
 rod->software = NULL;

 status= SUCCESS;
 return( status );

}

int
ROD_Config( ROD* rod, word32 *cfg_word ){
int status = FAIL;
word32 info[ 30 ], stat_addr;

 if( rod == NULL ){
   printf("\n[ %-30s ]","ROD: Oooooops Null Pointer" );
   fflush( stdout );
   return( status );
 }

 if( Retrieve_Virtual_Map( rod->hardware, info, cfg_word ) != SUCCESS){
   printf("\n[ %-30s ]", "ROD: Hardware: Map Retrieve: FAIL" );
   fflush( stdout );
   return( status );
 }

 if( rod->software == NULL ){
   stat_addr = ( word32 )( & ( rod->mi_stat ) );
   if( Open_Software( &( rod->software ), stat_addr ) != SUCCESS ){
     printf("\n[ %-30s ]", "ROD: Software: Open: FAIL" );
     fflush( stdout );
     return( status );
   }
#ifdef VERBOSE
   else{
    printf("\n[ %-30s ]", "ROD: Software: Config: Open: SUCCESS" ); 
    fflush( stdout );
   }
#endif
 }

 if( Config_Software( rod->software, info, cfg_word ) != SUCCESS ){
   printf("\n[ %-30s ]", "ROD: Software: Config: FAIL" );
   fflush( stdout );
   return( status );
 }

 status = SUCCESS;
 return( status );

}


void
RODinfo_Dump( ROD *rod ){
word32 jj;
ROD_Ctrl   *ctrl;
ROD_Module *module;

 /* Rodhardware Info */
 ctrl = &( rod->hardware->ctrl ); /* shorcut */
 printf("\n\n#--------------= RODhardware Info =--------------#");
 /* Controller Info */
 printf("\n# RODController #");
 printf("\n[%s]", ctrl->id );
 printf("\n[%s]", ctrl->ip_addr );
 printf("\n[Output Buffer][idx][%d]", ctrl->o_index );
 printf("\n[Input  Buffer][idx][%d]", ctrl->i_index );
 printf("\n[ACK base address][ local ][%08x]", ( unsigned )ctrl->l_csr );
 printf("\n[ACK base address][ VME   ][%08x]", ( unsigned )ctrl->r_csr );
 
 printf("\n# RODModule Info #");
 for( jj = 0; jj < rod->hardware->mod_size; jj++ ){
   module = &( rod->hardware->module_set[ jj ] );/* ShorCut */
   printf("\n[%s]", module->id );
   printf("\n[%s]", module->ip_addr );
   printf("\n[Output Buffer][idx][%d]", module->o_index );
   printf("\n[Input  Buffer][idx][%d]", module->i_index );
   printf("\n[ACK base address][ local ][%08x]", ( unsigned )( module->l_csr));
   printf("\n[ACK base address][ VME   ][%08x]", ( unsigned )( module->r_csr));
   printf("\n[Event Size][%d]", ( unsigned )( module->evt_size ));
   switch( module->io_type ){
   case __feBOTH__  :printf("\n[%s]", "Both" );break;
   case __feSLINK__ :printf("\n[%s]", "Slink" );break;
   case __feVME__   :printf("\n[%s]", "VME" );break;
   case __Undef__   :printf("\n[%s]", "Undefined" );break;
   }
   fflush( stdout );
 }
 printf("\n#--------------= RODhardware Info =--------------#\n");
 
}


int
Ptr_Checking( ROD *rod ){
word32 ii, jj, kk, obj_index;
word32 *r_ptr, *l_ptr;
DMA_Dev *dma_dev;

 printf("\n\n# --------------= Checking Pointers Integrity =-------------- #");
 /* MASTER Pointers. PCI->VME */
 r_ptr = rod->hardware->ctrl.r_csr;
 r_ptr[ 0 ] = 0xDeadBeef;
 if( r_ptr[ 0 ] != 0xDeadBeef )
   printf("\nController CSR: Master VME: STINKS");
 else
   printf("\nController CSR: Master VME: PASSED");
 
 for( ii = 0; ii < rod->hardware->mod_size; ii++ ){
   r_ptr = rod->hardware->module_set[ ii ].r_csr;
   r_ptr[ 0 ] = 0xDeadBeef;
   if( r_ptr[ 0 ] != 0xDeadBeef )
     printf("\nModule [ %s ] CSR: Master VME: STINKS", 
	    rod->hardware->module_set[ ii ].id );
   else
     printf("\nModule [ %s ] CSR: Master VME: PASSED", 
	    rod->hardware->module_set[ ii ].id );
   sleep( 1 );
   printf("\nChecking DMAs");
   obj_index = rod->hardware->module_set[ ii ].o_index;
   dma_dev = ( DMA_Dev* )( rod->hardware->vme->rod_dev[ obj_index ].vme_obj );
   
   for( jj = 0; jj < 10000; jj++ ){
     if( DMA_Go( dma_dev, VMEtoLOCAL, 1024 * sizeof( word32 ) ) != SUCCESS ){
       printf("\n[ DMAGo FAILs" );
       fflush( stdout );
     }
   }
   printf("\nModule [ %s ]: OutPut Buffer: PASSED", 
	  rod->hardware->module_set[ ii ].id );
   l_ptr = rod->hardware->module_set[ ii ].r_obuf;
   printf("\n[ Buffer Sampling ]");
   for( kk = 0; kk < 1024; kk++ ){
     if( !( kk%100 ) )
       printf("\n[%04d][$%08x]", kk, l_ptr[ kk ] );
   }
   
   obj_index = rod->hardware->module_set[ ii ].i_index;
   dma_dev = ( DMA_Dev* )( rod->hardware->vme->rod_dev[ obj_index ].vme_obj );
   for( jj = 0; jj < 10000; jj++ ){
     if( DMA_Go( dma_dev, VMEtoLOCAL, 1024 * sizeof( word32 ) ) != SUCCESS ){
       printf("\n[ DMAGo FAILs" );
       fflush( stdout );
     }
   }
   printf("\nModule [ %s ]: Input Buffer: PASSED", 
	  rod->hardware->module_set[ ii ].id );
   l_ptr = rod->hardware->module_set[ ii ].r_ibuf;
   printf("\n[ Buffer Sampling ]");
   for( kk = 0; kk < 1024; kk++ ){
     if( !( kk%100 ) )
       printf("\n[%04d][$%08x]",kk, l_ptr[ kk ] );
   }
   
 }
 printf("\n# --------------= Checking Pointers Integrity =-------------- #\n");
 return( SUCCESS );
 
}
