/**
    $Header:  RODhardware Object for the Front End Application
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $

    22 October  1999: 
    Front End integration
    
    02 November 1999: 
    FE-Controller Booting already going ON
    
    03 Movember 1999: 
    RODvme gathering is broken in ROD-pieces

    10/December/99:
    Front-End Configuration Decoding Starts
**/

#include "RODhardware.h"

int 
Open_Hardware( ROD_Hardware **rod_hard, word32 info_ptr ){
int status = SUCCESS;

  *rod_hard = ( ROD_Hardware* )malloc( sizeof( ROD_Hardware ) );
 if( *rod_hard == NULL ){
   printf( "\n[ %-30s ]", "ROD Hardware: Malloc: Fail");
   fflush( stdout );
   return( FAIL );
 }
 if( Open_RODVME( &( ( *rod_hard )->vme ) ) != SUCCESS ){
   printf("\n[ %-30s ]","ROD VME: Open: FAIL" );
   fflush( stdout  );
   return( FAIL );
 }

 else{
#ifdef VERBOSE
   printf("\n[ %-30s ]", "ROD VME: Open: SUCCESS" );
   fflush( stdout );
#endif
   if( SetUp_Hardware( *rod_hard, info_ptr ) != SUCCESS ){
     printf("\n[ %-30s ]","ROD Module Detection: FAILs" );
     return( FAIL );
   }
   if( Handshake_Controller( *rod_hard ) != SUCCESS ){
     printf("\n[ %-40s ]","ROD-Hardware: Controller: Detection: FAILs" );
     fflush( stdout );
     status = FAIL;
   }
#ifdef VERBOSE
   else{
     printf("\n[ %-30s ]", "ROD-Hardware: Controller: Detection: SUCCESS" );
     fflush( stdout );
   }
#endif

#ifdef LinuxPCBit3
   if( Open_RODSlink( &( ( *rod_hard )->slink ) ) != SUCCESS ){
     printf("\n[ %-30s ]", "ROD Slink: Open: FAIL" );
     fflush( stdout );
     return( FAIL );
   }
#endif
 }
 return( status );

}


int
Close_Hardware( ROD_Hardware *rod_hard ){
int status = SUCCESS;

 if( rod_hard != NULL ){ 
   if( rod_hard->vme != NULL ){
     if( Close_RODVME( rod_hard->vme ) != SUCCESS ){
       printf("\n[ %-30s ]","ROD VME: Close: FAIL" );
       fflush( stdout  );
       status = FAIL;
     }
#ifdef VERBOSE
     else{
       printf("\n[ %-30s ]","ROD VME: Close: SUCCESS" );
       fflush( stdout  );
     }
#endif
   }
   /* Module Release */
   if( rod_hard->module_set != NULL ){
     free( rod_hard->module_set );
#ifdef VERBOSE
     printf("\n[ %-30s ]","ROD Hardware: Module-Set FREE: SUCESS" );
     fflush( stdout  );
#endif
   }
   else{
     printf("\n[ %-30s ]","ROD Hardware: Module-Set: Already FREE" );
     fflush( stdout  );
   }
#ifdef LinuxPCBit3
   if( rod_hard->slink != NULL ){
     if( Close_RODSlink( rod_hard->slink ) != SUCCESS ){
       printf("\n[ %-30s ]", "ROD Slink: Close: FAIL" );
       fflush( stdout );
       status = FAIL;
     }
#ifdef VERBOSE
     else{
       printf("\n[ %-30s ]", "ROD Slink: Close: SUCESS" );
       fflush( stdout );
     }
#endif
   }
#endif
   free( rod_hard );
 }
 return( status );

}


int
SetUp_Hardware( ROD_Hardware* rod_hard, word32 info_ptr ){
int status =SUCCESS;

 if( SetUp_Local( rod_hard ) != SUCCESS ){
   printf("\n[ %-40s ]","ROD-Hardware: ROD-Controller SetUp: FAILs" );
   return( FAIL );
 }
 if( SetUp_Controller( rod_hard, info_ptr ) != SUCCESS ){
   printf("\n[ %-40s ]","ROD-Hardware: ROD-Controller SetUp: FAILs" );
   status = FAIL;
 }
 if( SetUp_Module( rod_hard, info_ptr ) != SUCCESS ){
   printf("\n[ %-40s ]","ROD-Hardware: FE-SetUp: FAILs" );
   status = FAIL;
 }
 return( status );

}


/*+ 

  03/November/99: Retrieves from RODvme Local hardware resources.
  CSR pointers SetUp and DMA objects

+*/
int
SetUp_Local( ROD_Hardware *rod_hard ){
Search_Key  key;
word32      obj_index, addr;

 /* CSR */ 
 sprintf( key.machine, "%s", "localhost" );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_csr = ( word32* )addr; 
 /* IBuffer */
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_ibuf = ( word32* )addr;
 rod_hard->i_index = obj_index;
 /* OBuffer */
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_obuf = ( word32* )addr;
 rod_hard->o_index = obj_index;
 fflush( stdout ); 
 return( SUCCESS );

}


/*+ 

  03/November/99: Retrieves from RODvme ROD-Controller hardware resources.
  CSR pointers SetUp and DMA objects
  !!!!! WARNING !!!!! SetUp_Local MUST be called first.

+*/
int
SetUp_Controller( ROD_Hardware *rod_hard, word32 info_ptr ){
Search_Key  key;
word32      obj_index, addr;
ROD_info   *info;

 info = ( ROD_info* )info_ptr;
 sprintf( rod_hard->ctrl.id, "%s", info->ctrl.id );
 sprintf( rod_hard->ctrl.ip_addr, "%s", info->ctrl.ip_addr );
  
 /* CSR */ 
 sprintf( key.machine, "%s", rod_hard->ctrl.ip_addr );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.r_csr = ( word32* )addr;
 /* The problem Arrives:
    Q: Which Offset is sampled by ROD_Controller to boot the FE_i?
    A: CSR_mod*no_modules + CSR_fe*FE_i */
 rod_hard->ctrl.r_csr = 
   &rod_hard->ctrl.r_csr[ ( info->fe.crate_mod_size * CSR_ROD_SIZE ) 
			+ ( info->fe.idx * CSR_FE_SIZE ) ];

 rod_hard->ctrl.l_csr = rod_hard->l_csr;

 /* IBuffer */
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.i_index = obj_index;
 /* OBuffer */
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.o_index = obj_index;
 rod_hard->ctrl.lives   = LIVE_CHANCE;
 return( SUCCESS );

}


/*+ 

  03/November/99: Retrieves from RODvme hardware resources for each Module.
  CSR pointers SetUp and DMA objects

  !!!!! WARNING !!!!! SetUp_Local MUST be called first.

+*/
int
SetUp_Module( ROD_Hardware *rod_hard, word32 info_ptr ){
Search_Key   key;
word32       obj_index, addr, ii;
Link        *link;/* shorcuts */
ROD_info    *info;
ROD_Module  *mod;

 info = ( ROD_info* )info_ptr;
 rod_hard->mod_size = info->fe.no_itm; 
 link = &( info->fe.link[ 0 ] );

 /* Module Alloc */
 if( rod_hard->module_set == NULL ){
   rod_hard->module_set = ( ROD_Module* )
     malloc( sizeof( ROD_Module )*rod_hard->mod_size);
   if( rod_hard->module_set == NULL ){
     printf("\n[ %-30s ]", "Not Enough Memory to build Module set" );
     return( FAIL );
   }
 }
 else{
    printf("\n[ %-30s ]", "ROD Hardware: Module-Set: Already Done" );
    fflush( stdout );
 }

 mod  = ( ROD_Module* )rod_hard->module_set;
 /* Module Vector SetUp */
 for( ii = 0 ; ii < rod_hard->mod_size ; ii++ ){
   
   sprintf( mod[ ii ].id, "%s", link[ ii ].link_id );
   sprintf( mod[ ii ].ip_addr, "%s", link[ ii ].ip_addr );
   if( strcmp( link[ ii ].io_type, "Both" ) == 0 )
     mod[ ii ].io_type = __feBOTH__ ;
   else
     if( strcmp( link[ ii ].io_type, "Slink" ) == 0 )
       mod[ ii ].io_type = __feSLINK__ ;
     else
       if( strcmp( link[ ii ].io_type, "VME" ) == 0 )
	 mod[ ii ].io_type = __feVME__ ;
       else
	 mod[ ii ].io_type = __Undef__ ;

   /* Pointers SetUp */
   /* Remote CSR */
   sprintf( key.machine, "%s", mod[ ii ].ip_addr );
   sprintf( key.dev_id, "%s", "CSR_Space" ); 
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
       != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     return( FAIL );
   }
   /* Fe-RODmodule Map. The problem Arrives:
    Q1: Which Offset is sampled by the ROD_Module to interface with its FE?
    A: 0
    Q2: Which offset is locally sampled to close above control loop?
    A: CTRL space + the module ii* the space of each module CSR
   */
   mod[ ii ].r_csr = ( word32* )addr; /* base address of Mod_i Slave CSR */
   mod[ ii ].l_csr = &rod_hard->l_csr[ CSR_CTRL_SIZE + ( ii*CSR_ROD_SIZE ) ];
   
   /* IBuffer */
   sprintf( key.dev_id, "%s", "Input_Buffer" ); 
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
       != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     return( FAIL );
   }
   mod[ ii ].r_ibuf = ( word32* )addr; /* raw pointer */
   mod[ ii ].i_index = obj_index; /* DMA */

   /* OBuffer */
   sprintf( key.dev_id, "%s", "Output_Buffer" ); 
   if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
       != SUCCESS ){
     printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	    key.machine, key.dev_id, "Object Not Found" );
     return( FAIL );
   }
   mod[ ii ].r_obuf = ( word32* )addr;/* raw pointer */
   mod[ ii ].o_index = obj_index;/* DMA */

 } 
 return( SUCCESS );
 
}


/*+ 

  03/November/99: Function Redesign. 
  Boot Protocol completely isolated. 

+*/   
int
Handshake_Controller( ROD_Hardware *rod_hard ){
word32 *r_csr, *l_csr, time_out;

 r_csr = rod_hard->ctrl.r_csr;
 l_csr = rod_hard->ctrl.l_csr;
 l_csr[ 0 ] = CONTROLLER_DEAD;

 time_out = BOOT_TIME_OUT;
 r_csr[ 0 ] = FE_ALIVEs;
 while( l_csr[ 0 ] == CONTROLLER_DEAD ){
   printf("\n[ %-15s ]", "Waiting Controller ACK" );
   --time_out;
   r_csr[ 0 ] = FE_ALIVEs;
   if( !time_out ){
     printf("\n\n[ ROD Controller is Dead ]\n");
     return( FAIL );
   }
   sleep( 1 );
 } 
 return( SUCCESS );

}

/* 
   Beware !!!!! Remember that Fe-Module Map is retrieved on the fourth 
   configuration Packet word 
   
   "info" pointer Format:
   Local Info
   [0]Virtual Address of Local VME Slave Buffer. Used also for Slink. Offset 0
   [1]Buffer size
   [2]Physical Address

   Module Info
   [3]VPtr to Local CSR
   [4]VPtr to Remote CSR
   [5]Index over vme structure to Perform DMAs on Input Buffer

   [6]Slink stuff
*/
int
Retrieve_Virtual_Map( ROD_Hardware* rod_hard, word32* info, word32 *cfg_word ){
VMESlave* dev;
word32 obj_index;
ROD_Module  *mod;

 if( rod_hard == NULL ){
   printf("\n\n[ %-30s ]","ROD Hardware: Oooooops Null Pointer" );
   return( FAIL );
 }
 mod = rod_hard->module_set; /* Shortcut */
 
 /* Target Mode ----------------------------------------------------------- */
 obj_index = rod_hard->i_index;
 info[ 0 ] = ( word32 )rod_hard->l_ibuf;
 dev = ( VMESlave* )( rod_hard->vme->rod_dev[ obj_index ].vme_obj );

 info[ 1 ] = dev->size;
#if defined( MVME167_Lynx2_5_1 ) || defined( PowerPC_Lynx3_0 )
 info[ 2 ] = dev->phy_addr;
#else
 info[ 2 ] = dev->ram.pci_addr;
#endif 
 
 /* Driver Mode And CSRs -------------------------------------------------- */
#ifdef VERBOSE
 printf("\n\n[ ROD: hardware: Retrieving IO-info for Module %s ]", 
	mod[ cfg_word[ 3 ] ].id );
 fflush( stdout );
#endif

 info[ 3 ] =  ( word32 )
   ( &( rod_hard->vme->rod_dev[ mod[ cfg_word[ 3 ] ].i_index ] ) );
 info[ 4 ] = ( word32 )( mod[ cfg_word[ 3 ] ].r_ibuf  );
 info[ 5 ] = ( word32 )( mod[ cfg_word[ 3 ] ].l_csr );
 info[ 6 ] = ( word32 )( mod[ cfg_word[ 3 ] ].r_csr );

 /* Slink Device */
#if defined( LinuxPCBit3 )
 info[ 7 ] = ( word32 )rod_hard->slink;
#endif
  return( SUCCESS );
}
