/**
    $Header: TileCal ROD Front End Application --> ROD.c
    Created & Maintained by JB. Juanba.Romance@ific.uv.es $

    10/December/99:
    Threads SetUp for Data-Injection starts.

    14/December/99:
    Data Injection Protocol Definition starts. Lot of Troubles with 
    the previous design between Controller(MVME167)-Module(PowerPC) already
    going on ???
    
    15/December/99:
    Added Offset +6 on the Local CSR updated by the FrontEnd 
    Protocol Troubble: Suspiciously related CPU cache
    Solution: r_csr, l_csr defined with "volatile" modifiers 
    Target Mode Already Going On

    18/December/99
    Data Injection Protocol Redefinition. Target And Driver Mode Already 
    going on. 

    21/December/99
    Slink Injection Protocol starts.
**/

#include "RODsoftware.h"
#define __ContinueDelay__ 20000
#define __InjectionDelay__ 300

/* -------------------------- External Interface --------------------------- */
int
Open_Software( ROD_Software **rod_soft, word32 stat_addr ){
int status = SUCCESS;
word32 ii;

 *rod_soft = ( ROD_Software* )malloc( sizeof( ROD_Software ) );
 if( *rod_soft == NULL ){
   printf( "\n[ %-30s ]", "Software: Malloc: Fail");
   fflush( stdout );
   return( FAIL );
 }
 ( *rod_soft )->stat = ( word32* )stat_addr;
 ( *rod_soft )->in.evt_buf  = ( word32* )malloc( sizeof( word32 )*4096 );
 sprintf( ( *rod_soft )->in.io_id, "%s", "Input Unit" );
 sprintf( ( *rod_soft )->pro.io_id, "%s", "Productor Unit" );
 
 /* Syncronization Tools Routing */
 pthread_mutex_init ( &( ( *rod_soft )->pro.io_mtx ), NULL );
 pthread_cond_init  ( &( ( *rod_soft )->pro.io_cv ), NULL  );
 ( *rod_soft )->pro.io_ready = &( ( *rod_soft )->in.io_ready );
 ( *rod_soft )->in.ready     = &( ( *rod_soft )->pro.ready );
 ( *rod_soft )->in.io_mtx    = &( ( *rod_soft )->pro.io_mtx );
 ( *rod_soft )->in.io_cv     = &( ( *rod_soft )->pro.io_cv );

 /* Extra Pointers Routing */
 ( *rod_soft )->pro.evt_buf = ( *rod_soft )->in.evt_buf; 
 ( *rod_soft )->in.stat     = ( *rod_soft )->stat;
 ( *rod_soft )->pro.stat    = ( *rod_soft )->stat;

 /* Some Default Values */
 ( *rod_soft )->in.io_ready  = CLEAR;
 ( *rod_soft )->pro.ready    = CLEAR;
 for( ii = 0; ii < 1024 ; ii++ ){
   ( *rod_soft )->in.evt_buf[ ii ] = 1 << ( ii % 32 );
 }

 *( ( *rod_soft )->stat ) = __Down__ ;
 return( status );

}

int
Close_Software( ROD_Software *rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   if( *( rod_soft->stat ) != __Down__ ){
     *( rod_soft->stat ) = __Down__ ;
#ifdef VERBOSE
     printf("\n[ %-30s ]", "ROD: Software: Killing Data Injection System");
#endif
     pthread_join( rod_soft->in.phy_th, NULL );
     pthread_cancel( rod_soft->pro.phy_th );

     pthread_mutex_destroy ( &rod_soft->pro.io_mtx );
     pthread_cond_destroy  ( &rod_soft->pro.io_cv  );
   }
#ifdef VERBOSE
   printf("\n[ %-30s ]", "ROD: Software: Releasing Resources");
   fflush( stdout );
#endif
   free( rod_soft->in.evt_buf );
   free( rod_soft );
 }
 return( status );

}


int
Config_Software( ROD_Software* rod_soft, word32* io_data, word32* cfg_word ){
int status = SUCCESS;
Config_Word cfg_w;
VME_Dev  *vme_dev;
#if defined( LinuxPCBit3  )
ROD_slink *my_slink;
#endif

 if( rod_soft == NULL ){
   printf("[ %-20s ]", "Oooooooops: ROD: Software: NULL ptr" );
   return( FAIL );
 }

 cfg_w.all = cfg_word[ 0 ];
#if defined( LinuxPCBit3  )
 if( cfg_w.bit.io_type == __Slink__ ){
   rod_soft->in.slink_addr = io_data[ 7 ];   
   my_slink = ( ROD_slink* )io_data[ 7 ];

   rod_soft->in.buffer  = my_slink->dma_dev.buffer;
   rod_soft->in.io_type = __Slink__ ;
   my_slink->dma_ctrl.fld.mode.ctrl = CTRL_PACK;   /* With Control Words */
   my_slink->dma_ctrl.fld.mode.nopack = 1; 
   my_slink->dma_ctrl.fld.pack_addr[ 0 ] = my_slink->dma_dev.pci_addr;
   my_slink->dma_ctrl.fld.pack_size[ 0 ] = 
     ( cfg_w.bit.no_samples*cfg_w.bit.no_channels );
 } 
#endif
 if( cfg_w.bit.io_type == __VME__ ){
   rod_soft->in.io_type = __VME__ ;
   if( cfg_w.bit.io_mode == __Target__ ){
     printf("\n[ ROD: Software: Injection Mode: TARGET ]" );
     rod_soft->in.dma_obj = io_data[ 3 ];
     rod_soft->in.buffer  = ( word32* )io_data[ 4 ];
     vme_dev = ( VME_Dev* )( rod_soft->in.dma_obj );
#ifdef VERBOSE
     printf("\n[ ROD: Software: DMAObject Info ]" );
     Dump_Object( vme_dev );
#endif
   }
   else
     if( cfg_w.bit.io_mode == __Driver__ ){
       printf("\n[ ROD: Software: Injection Mode: DRIVER ]" );
       rod_soft->in.dma_obj = 0xDeadBeef;
       rod_soft->in.buffer  = ( word32* )io_data[ 0 ];
     }
 }

 rod_soft->in.l_csr = ( word32* )io_data[ 5 ];
 rod_soft->in.r_csr = ( word32* )io_data[ 6 ];
 rod_soft->in.cfg   = cfg_word[ 0 ];

 /* Counters Locally Driven by Controller */
 rod_soft->in.l_csr[ __IUHandshake__ +6 ] = CLEAR;
 rod_soft->in.l_csr[ __IUHandshake__ +5 ] = CLEAR;
 rod_soft->in.l_csr[ __IUHandshake__ +4 ] = CLEAR; 
 return( status );
 
}


int 
Run_Software( ROD_Software* rod_soft ){
int status = SUCCESS;
pthread_attr_t  pth_attr;
pthread_t *pth_list[ 2 ];
void      *fun_list[ 2 ];
void      *par_list[ 2 ];
word32 ii;

 if( rod_soft == NULL ){
   printf("[ %-20s ]", "Oooooooops: ROD: Software: NULL ptr" );
   return( FAIL );
 }

 pth_list[ 0 ] = &rod_soft->pro.phy_th;
 pth_list[ 1 ] = &rod_soft->in.phy_th;
 fun_list[ 0 ] = ( void *)Production_Engine;
 fun_list[ 1 ] = ( void *)Input_Engine;
 par_list[ 0 ] = ( void* )( &rod_soft->pro );
 par_list[ 1 ] = ( void* )( &rod_soft->in );
#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "Thread Attr" );
   return( FAIL );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "Thread Attr" );
   return( FAIL );
 }
#endif

 for( ii = 0; ii < 2; ii++ ){
#if defined( LinuxPCBit3 )
   if( pthread_create( pth_list[ ii ],
		       &pth_attr, 
		       fun_list[ ii ], 
		       par_list[ ii ] ) != 0 ){
     perror( "Thread Create" );
     *( rod_soft->stat ) = __Down__ ;
     break;
   }
#else
   if( pthread_create( pth_list[ ii ],
		       pth_attr, 
		       fun_list[ ii ], 
		       par_list[ ii ] ) != 0 ){
     perror( "Thread Create" );
     *( rod_soft->stat ) = __Down__ ;
     break;
   }
#endif
   else
     *( rod_soft->stat ) = __Up__ ;
 } 
 if( *rod_soft->stat == __Down__ )
   status = FAIL;

 printf("\n[ %-30s ]","FrontEnd Data Injection: Running" );
 return( status );

}

int
Pause_Software( ROD_Software* rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   *( rod_soft->stat ) = __Pause__ ;
 }
 return( status );

}

int
Continue_Software( ROD_Software* rod_soft ){
int status = SUCCESS;

 if( rod_soft != NULL ){
   *( rod_soft->stat ) = __Continue__ ;
 }
 return( status );

} 

/* ------------------------------------------------------------------------- */


/* ----------------------= When The Action Begins =------------------------- */
int
Production_Engine ( void *par ){
Productor *pro;
word32 ii, jj, count = CLEAR;

 printf("\n\n[ FrontEnd: Software: %-20s ][%d]", 
	"Production_Engine Starts", getpid( ) );
 fflush( stdout );
 pro = ( Productor* )par;
 *( pro->stat )= __Run__ ;
 jj = 0;

 while( *( pro->stat ) != __Down__ ){
   if( *( pro->stat ) == __Pause__ ){
     ++jj;
     if( !( jj%100 )){
       printf("\n[ %-20s ][ %-8s ]", "Production_Engine", "PAUSED" );
       fflush( stdout );
     }
     usleep( __ContinueDelay__ );
     continue;
   }
   if(( *( pro->stat ) == __Run__ )||( *( pro->stat ) == __Continue__ )){
     jj = 0;
     /* IO Thread Synchronism ---------------------------------------- */
     pthread_mutex_lock( &pro->io_mtx );
     if( *( pro->io_ready ) == CLEAR ){
       pro->ready = SET;
       pthread_cond_wait( &pro->io_cv, &pro->io_mtx );
       pro->ready = CLEAR;
     }
     for( ii = 0 ; ii < 1024 ; ii++ ){
       pro->evt_buf[ ii ] = ( 1 << ( ii % 32 ) );
     }
     pthread_cond_signal( &pro->io_cv );
     pthread_mutex_unlock( &pro->io_mtx );
     /* --------------------------------------------------------------- */
     ++count;

   }
 }
 printf("\n[ FrontEnd: Software: %20s ]", "Production_Engine Ends" );
 fflush( stdout );
 return( SUCCESS );

}

int 
Input_Engine( void *par ){
IO_Type   *in;
word32 count = CLEAR, evt_size, jj;
Config_Word cfg_w;

 printf("\n\n[ FrontEnd: Software: %-20s ][%d]", 
	"Input_Engine Starts", getpid( ) );
 fflush( stdout );

 in = ( IO_Type* )par; 
 *( in->stat ) = __Run__ ;

 cfg_w.all = in->cfg;
 printf("\n[ Config Word: $%08x ][ %-20s ][%s]", 
	in->cfg, "Input Mode", cfg_w.bit.io_mode ? "Driver" : "Target" );
 printf("\n[ %-20s ][%d][%d]",
	"Channels/Samples", 
	( unsigned )cfg_w.bit.no_channels, 
	( unsigned )cfg_w.bit.no_samples );
 fflush( stdout );
 evt_size = cfg_w.bit.no_channels * cfg_w.bit.no_samples;
 jj = 0;

 while( *( in->stat ) != __Down__ ){

   if( *( in->stat ) == __Pause__ ){
     ++jj;
     if( !( jj%100 )){
       printf("\n[ %-20s ][ %-8s ]", "Input_Engine","PAUSED" );
       fflush( stdout );
     }
     usleep( __ContinueDelay__ );
   }

   if( (*( in->stat ) == __Run__ )||( *( in->stat ) == __Continue__ )){
     jj = 0;
     
     /* Production Thread Synchronism --------------------------------- */
     pthread_mutex_lock( in->io_mtx );
     if( *( in->ready ) == CLEAR ){
       in->io_ready = SET;
       pthread_cond_wait( in->io_cv, in->io_mtx );
       in->io_ready = CLEAR;
     }
     memcpy( in->buffer, in->evt_buf, evt_size*sizeof( word32 ) );
     pthread_cond_signal( in->io_cv );
     pthread_mutex_unlock( in->io_mtx );
     /* --------------------------------------------------------------- */

     if( in->io_type == __VME__ ){
       if( In_Handshake_VME( in, &count ) != SUCCESS ){
	 if( in->l_csr[ __IUHandshake__ +1 ] > __DataSRCFail__ ){
	   *( in->stat ) = __Pause__ ; 
	 }
       }
     }
     else{
       if( in->io_type == __Slink__ ){
	 if( In_Handshake_Slink( in, &count ) == FAIL ){
	     if( in->l_csr[ __IUHandshake__ +1 ] ==  __DataSRCFail__ ){
	     }
	 }
       }
     }
     ++count;
   }

 }
#ifdef VERBOSE
 printf("\n[ FrontEnd: Software: %20s ]", "Input_Engine Ends");
 fflush( stdout );
#endif
 return( SUCCESS );

}



int
In_Handshake_VME( IO_Type *in, word32 *count ){
#define __InCTRLTimeOut__ 100000000
int status = SUCCESS;
volatile word32 *l_csr, *r_csr;
word32 time_out, size ;
Config_Word cfg_w;
DMA_Dev   *dma_dev;


 l_csr = in->l_csr;
 r_csr = in->r_csr;
 cfg_w.all = in->cfg;

 if( !( ( *count )%10000 ) ){
   printf("\n[ Trigger State ]\n[ Hits %04d ]", ( unsigned )( *count ) );
   printf("\n[ Local  Trials   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +6 ]);
   printf("\n[ Local  Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +5 ]);
   printf("\n[ Local  Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +4 ]);
   printf("\n[ Module Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +2 ]);
   printf("\n[ Module Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +1 ]);
   printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]", 
	  ( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
   printf("\n[ %-20s ][%s]", 
	  "VME Input DMA Mode",  cfg_w.bit.io_mode ? "Driver" : "Target" );
   fflush( stdout );
 }


 l_csr[ __IUHandshake__ +6 ] = *count;
 time_out = __InCTRLTimeOut__ ;
 while( l_csr[ __IUHandshake__ +3 ] != *count ){
   --time_out;
   if( time_out == 0 ){
     ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 10 ) ){
	printf("\n[ ROD: Software: TriggerON lost ]");
	printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]",
	                 ( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
	fflush( stdout );
     }
     *count = l_csr[ __IUHandshake__ +3 ];
    /* usleep( 1 ); */
     return( FAIL );
   }
 }


 /* IO VME ----------------------------------------------------------------- */
 if( cfg_w.bit.io_mode == __Target__ ){
   size = cfg_w.bit.no_samples * cfg_w.bit.no_channels ;
   dma_dev = ( DMA_Dev* )((( VME_Dev* )( in->dma_obj ))->vme_obj );
   if( DMA_Go( dma_dev, LOCALtoVME, size * sizeof( word32 ) ) != SUCCESS ){
     printf("\n[ IUnit: DMA to Data Source: Fails ]" );
     fflush( stdout );
     return( FAIL );
   }
 }
 /* ------------------------------------------------------------------------ */


 time_out = __InCTRLTimeOut__ ;
 while( l_csr[ __IUHandshake__ ] != __TriggerOff__ ){
   --time_out;
   if( time_out == 0 ){
     ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 100 ) ){
       printf("\n[ ROD: Software: TriggerOff Lost ][%d]", *count );
       fflush( stdout );
     }
     l_csr[ __IUHandshake__ ] = MASK_COMMAND;
     return( FAIL );
   }
 }
 l_csr[ __IUHandshake__ ] = MASK_COMMAND;

 r_csr[ __IUHandshake__  ] = __TriggerOff__ ;
 ++l_csr[ __IUHandshake__ +5 ];
 return( status );
 
}


int
In_Handshake_Slink( IO_Type *in, word32 *count ){
int status = SUCCESS;

#if defined( LinuxPCBit3 )
#define __SlinkTimeOut__ 10000000

ROD_slink *my_slink;
volatile word32 *l_csr, *r_csr;
   word32 time_out;

   l_csr = in->l_csr;
   r_csr = in->r_csr;
   if( !( ( *count )%10000 ) ){
      printf("\n[ Trigger State ]\n[ Hits %04d ]", ( unsigned )( *count ) );
   printf("\n[ Local  Trials   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +6 ]);
   printf("\n[ Local  Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +5 ]);
   printf("\n[ Local  Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +4 ]);
   printf("\n[ Module Sucess   ][%6u]", (unsigned)l_csr[ __IUHandshake__ +2 ]);
   printf("\n[ Module Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +1 ]);
   printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]", 
	  ( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
    fflush( stdout );
 }

   l_csr[ __IUHandshake__ +6 ] = *count;
   time_out = __SlinkTimeOut__ ;
   while( l_csr[ __IUHandshake__ +3 ] != *count ){
    --time_out;
    if( time_out == 0 ){
      ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 10 ) ){
	printf("\n[ ROD: Software: TriggerON lost ]");
	printf("\n[ Trigger Counters ][Module: %6u][Local: %6u]",
	                 ( unsigned )l_csr[ __IUHandshake__ +3 ], *count );
     }
     *count = l_csr[ __IUHandshake__ +3 ];
     /* usleep( 1 ); */
     return( FAIL );
   }
 }


 /* Slink I/O -------------------------------------------------------------- */

 my_slink = ( ROD_slink* )( in->slink_addr );
 status = SlinkDMA_Complex_Write( &( my_slink->dma_dev ), 
				  &( my_slink->slink_dev ), 
				  &( my_slink->dma_ctrl ) );
 if( status != SUCCESS ){
   printf("\n[ ROD: Software: DMASlink: Fails ][%d]", (unsigned)( *count ) );
   fflush( stdout );
   return( FAIL );
 }

 /* ------------------------------------------------------------------------ */


 time_out = __SlinkTimeOut__ ;
 while( l_csr[ __IUHandshake__ ] != __TriggerOff__ ){
   --time_out;
   if( time_out == 0 ){
     ++( l_csr[ __IUHandshake__ +4 ] );
     if( !( l_csr[ __IUHandshake__ +4 ] % 100 ) ){
       printf("\n[ ROD: Software: Slink-TriggerOff Lost ][%d]", 
	      ( unsigned )count);
       fflush( stdout );
     }
     l_csr[ __IUHandshake__ ] = MASK_COMMAND;
     return( FAIL );
   }
 }
 l_csr[ __IUHandshake__ ] = MASK_COMMAND;

 r_csr[ __IUHandshake__  ] = __TriggerOff__ ;
 ++l_csr[ __IUHandshake__ +5 ]; 

#endif

 return( status ); 

}


