/** 
   $Header: RODvme Object for the ROD Front-End Application
   Created & Maintained by JB. Juanba.Romance@ific.uv.es $

   04/April/99 Try to write a Common RODvme.c, VMElib.c on MVME167 & PowerPC 
   has been uptaded to be compliance with:

   - Data Types on Linux( VMESlave, VMEMaster, DMA_Dev )

   - Methods to alloc/free VME Objects

   - Perform Automatic VME Objects production from LocalVME ROD Data Base..
**/

#include "RODvme.h"
extern char bd[ ]; /* Higlighting */
extern char us[ ];
extern char atoff[ ];

int
Open_RODVME( ROD_vme **vme ){

 *vme = ( ROD_vme* )malloc( sizeof( ROD_vme ) );
 if( vme == NULL )
   return( FAIL );

 /*+ Host Chipset  +*/
#ifdef LinuxPCBit3
 ( *vme )->b3csr_dev = ( Bit3CSR_Dev* )malloc( sizeof( Bit3CSR_Dev ) );
 if( ( *vme )->b3csr_dev == NULL ){
   printf("\n[ %-20s ]","ROD_VME Bit3CSR: Device Alloc : FAIL");
   return( FAIL );
 }
 if( B3CSR_Open( ( *vme )->b3csr_dev ) != SUCCESS ){
   return( FAIL );
 }
 B3CSR_SetUp( ( *vme )->b3csr_dev );

 ( *vme )->b3dpt_dev = ( Bit3PT_Dev* )malloc( sizeof( Bit3PT_Dev ) );
 if( ( *vme )->b3dpt_dev == NULL ){
   printf("\n[ %-20s ]","ROD_VME Bit3DPT: Device Alloc : FAIL");
   return( FAIL );
 } 
 if( B3PT_Open( ( *vme )->b3dpt_dev ) != SUCCESS ){
   return( FAIL );
 }
 B3PT_SetUp( ( *vme )->b3dpt_dev );
 B3PT_Clean( ( *vme )->b3dpt_dev, MASTER_type );
 B3PT_Clean( ( *vme )->b3dpt_dev, SLAVE_type  );
 B3PT_Clean( ( *vme )->b3dpt_dev, DMA_type    );
#endif
#ifdef MVME167_Lynx2_5_1
 ( *vme )->chip2_dev = ( Chip2_Dev* )malloc( sizeof( Chip2_Dev ) );
 if( ( *vme )->chip2_dev == NULL ){
   printf("\n[ %-20s ]","ROD_VME VMEChip2: Device Alloc : FAIL");
   return( FAIL );
 }
 if( VME_Chip2_Open( ( *vme )->chip2_dev ) != SUCCESS ){
   return( FAIL );
 }
 if( VME_Chip2_SetUp( ( *vme )->chip2_dev ) != SUCCESS ){
   return( FAIL );
 }
#endif  
#ifdef PowerPC_Lynx3_0
 ( *vme )->universe_dev = ( Universe_Dev* )malloc( sizeof( Universe_Dev ) );
 if( ( *vme )->universe_dev == NULL ){
   printf("\n[ %-20s ]","ROD_VME Universe: Device Alloc : FAIL");
   return( FAIL );
 }
 if( Universe_Open( ( *vme )->universe_dev ) != SUCCESS ){
   return( FAIL );
 }
 if( Universe_SetUp( ( *vme )->universe_dev ) != SUCCESS ){
   return( FAIL );
 }
#endif

 /* ------------------------------------------------------------------------ */
 if( Read_Data_Base( *vme ) != SUCCESS ){
   printf("\n[ %-20s ]", "ROD VME DataBase: Setup : FAIL");
   return( FAIL );
 }
 return( SUCCESS );

}


int
Close_RODVME( ROD_vme* vme ){
int status, ii;
VME_Dev *rod_dev;

 status = SUCCESS;
 /* ----------------- Below Must Release Propertly Memory ------------------ */
 if( vme == NULL )
   return( FAIL );

 rod_dev = vme->rod_dev;
 if( rod_dev == NULL ){
   printf("\n[ %-20s ]","Internal VME ROD Structure not Initialised" );
   return( FAIL );
 }
 for( ii = 0; ii < vme->no_dev ; ii++ ){
   /* We can't Assume nothing about next down Layer pointers */
   if( rod_dev[ ii ].vme_obj == NULL ){
     continue;
   }
   if( rod_dev[ ii ].stat != FAIL ){
#ifdef VERBOSE
     printf("\n[ Releasing ][ %-15s ]", rod_dev[ ii ].id );
     fflush( stdout );
#endif
     switch( rod_dev[ ii ].obj_type ){
     case Master_VME :{
       VMEMaster *vme_master_dev;
       vme_master_dev = ( VMEMaster* )rod_dev[ ii ].vme_obj;
       status = Free_VMEMaster( vme_master_dev );
     }
     break;
     case Slave_VME  :{
       VMESlave *vme_slave_dev;
       vme_slave_dev = ( VMESlave* )rod_dev[ ii ].vme_obj;
       /* 
	  Remember that in Linux doesn't exist Static Slave Objects .... 
	  28/March/99 J.Roldan&JuanBa discussion about Linux Driver, Memory 
	  Device object, Current design, management and constrains  
       */
#ifdef LinuxPCBit3
       status = Free_VMESlave( vme_slave_dev );
#endif
#if defined( PowerPC_Lynx3_0 ) || defined( MVME167_Lynx2_5_1 )
       status = Close_VMESlave( vme_slave_dev );
#endif
     }
     break;
     case dma_VME    :{
       DMA_Dev *dma_dev;
       dma_dev = ( DMA_Dev* )rod_dev[ ii ].vme_obj;
       status = DMA_Free( dma_dev );
     }
     break;
     }
     if( status != SUCCESS )
       printf("\n[ %s Yet OPEN ]", rod_dev[ ii ].id );
   }
   free( rod_dev[ ii ].vme_obj );
   rod_dev[ ii ].stat = SUCCESS;
 }
 free( vme->rod_dev );

 /*+ Chipset Release +*/
#ifdef LinuxPCBit3
 if( B3PT_Close( vme->b3dpt_dev ) != SUCCESS ){
   printf("\n[ %s Yet OPEN ]\n", vme->b3dpt_dev->id );
   status = FAIL ;
 }
 if( B3CSR_Close( vme->b3csr_dev ) != SUCCESS ){
   printf("\n[ %s Yet OPEN ]\n", vme->b3csr_dev->id );
   status = FAIL;
 }
 if( vme->b3csr_dev != NULL )
   free( vme->b3csr_dev );
 if( vme->b3dpt_dev != NULL )
   free( vme->b3dpt_dev );

#endif
#ifdef PowerPC_Lynx3_0
 if( vme->universe_dev != NULL ){
   if( Universe_Close( vme->universe_dev  )!= SUCCESS ){
     printf("\n[ %s yet OPEN ]", vme->universe_dev->id );
     status = FAIL;
   } 
   free( vme->universe_dev );
 }
#endif
#ifdef MVME167_Lynx2_5_1
 if( vme->chip2_dev != NULL ){
   if( VME_Chip2_Close( vme->chip2_dev )!= SUCCESS ){
     printf("\n[ %s yet OPEN ]", vme->chip2_dev->id );
     status = FAIL;
   }
   free( vme->chip2_dev );
 }
#endif
 free( vme );
 return( status );
 
}


int
Read_Data_Base( ROD_vme* vme ){
FILE *file;
char line[ 80 ];
int status, init;
FilePar f_par;

 if( ( file = fopen( FileID, "r" ) ) == NULL ){
   printf("\n[ %-20s ][ %-20s ]\n", "Can Not Open", FileID );
   return( FAIL );
 }

 status = SUCCESS;
 init   = CLEAR;

#ifdef VERBOSE
 printf("\n\n[ %-30s ]", "Processing ROD Hardware VME Data Base" );
 fflush( stdout );
#endif

 while( SET ){
   if( fscanf( file, "%s", line ) != EOF ){
     if( strcmp( line, InitKey ) == 0 ){
       init = SET;
#ifdef VERBOSE
       printf("\n[ Init Key Found ]" );
       fflush( stdout );
#endif
       if( ( status = Scan_File( file, &f_par )) == SUCCESS ){
	 if( ( status = Make_RODVME_Objs( file, vme, &f_par )) == SUCCESS ){
	 }
       }
       break;
     } 
   }
   else{
     if( init == CLEAR ){
       printf("\n[ %-20s ][ %-20s ]", "Key: $Init$", "NOT Found" );
       status = FAIL;
     }
     break;
   }
 }

#ifdef VERBOSE
 printf("\n[ Done ]\n");
 fflush( stdout );
#endif

 fclose( file );
 return( status );

}


int
Scan_File( FILE* file, FilePar* f_par ){
int    status, detected;
char   dummy0[ 30 ], dummy1[ 30 ], dummy2[ 15 ], dummy6[ 15 ];
word32 ii;
#if defined( PowerPC_Lynx3_0 )
unsigned int dummy3, dummy4, dummy5;
#else
word32  dummy3, dummy4, dummy5;
#endif


char  *types[ VME_OBJECTs ]={
  "mst_VME",
  "slv_VME",
  "dma_VME"
};

 f_par->offset = ftell( file ); 
 f_par->count  = CLEAR;
 status = SUCCESS;
 while( SET ){
   status =fscanf( file, "%s\t%s\t%s\t$%08x\t$%02x\t%d\t%s", 
		   dummy1, dummy0, dummy2, &dummy3, &dummy4, &dummy5, dummy6 );
   if( status != 7 ){
     if( status == EOF ){
       return( SUCCESS );
     }
     printf("\n[ File Format Not Compliance @ raw %d ]", f_par->count);
     printf("\n[ Status ][ %d ]", status );
     return( FAIL );
   }
   /* Object Type Scan for each Item File */
   detected = FAIL;
   for( ii = 0; ii < VME_OBJECTs ; ii++ ){
     if( strcmp( dummy2, types[ ii ] ) == 0 ){
       detected = SUCCESS;
       f_par->count++;
       break;
     }
   }
   if( detected == FAIL ){
     printf("\n[ %-20s ][ raw %02d ]", "Unknown VME Object Type", 
	    f_par->count);
     return( FAIL );
   }
 }
}


/* --------- From Database a VME Objects List is made & Initialised -------- */
int
Make_RODVME_Objs(  FILE* file, ROD_vme* rod_vme, FilePar* f_par ){
char *types[ VME_OBJECTs ]={
  "mst_VME",
  "slv_VME",
  "dma_VME"
}; 
char type[ 30 ], dev_id[ 30 ], machine[ 30 ], dummy[ 30 ];
VME_Dev *rod_dev;
#if defined( PowerPC_Lynx3_0 )
unsigned int vme_base, am, size;
#else
word32  vme_base, am, size;
#endif
int ii, jj, status;

 rod_vme->no_dev  = f_par->count;
 rod_vme->rod_dev = NULL;
 rod_vme->rod_dev = ( VME_Dev* )malloc( sizeof( VME_Dev )*f_par->count );
 if( rod_vme->rod_dev == NULL ){
   printf( "\n[ %-20s ]", "Not Enought Memory to make VME Objects" );
   return( FAIL );
 }
 rod_dev = rod_vme->rod_dev;
 for( ii = 0; ii < f_par->count ; ii++ ){
   rod_dev[ ii ].vme_obj = NULL;
 }
 fseek( file, f_par->offset, SEEK_SET );
 
 /* Init Object Production */
 for( ii = 0; ii < f_par->count ; ii++ ){
   status = fscanf( file, "%s\t%s\t%s\t$%08x\t$%02x\t%d\t%s", 
		   machine, dev_id, type, &vme_base, &am, &size, dummy );

   sprintf( rod_dev[ ii ].id, "%s", dev_id );
   sprintf( rod_dev[ ii ].machine, "%s", machine );
   for( jj = 0; jj < VME_OBJECTs ; jj++ ){
     if( strcmp( type, types[ jj ] ) == 0 ){
       rod_dev[ ii ].obj_type = jj + 1;
       break;
     }
   }
   /* Obviously, each Object needs a different SetUp */
   switch( rod_dev[ ii ].obj_type ){
   case Master_VME :{
     VMEMaster *vme_master;
     /* Now Alloc Space to VME Object & Fill Parametters from Data base */
     rod_dev[ ii ].vme_obj = ( char* )malloc( sizeof( VMEMaster ));
     if( rod_dev[ ii ].vme_obj == NULL ){
       printf("\n[ %-20s ]","Not Enought Memory to make Master Object");
       return( FAIL );
     }
     rod_dev[ ii ].stat = SUCCESS;
     vme_master = ( VMEMaster* )rod_dev[ ii ].vme_obj;
     /* 
	04/April/99: 
	Obviously each VME Interface have different parametter input set 
	if we force this feature libVMEBit3, libVMEppc & libVMEchip2 libraries
	should program some parametters which are up to programmer, in summary 
	we will suffer a lack of freedom in other VME applications 
     */
#ifdef LinuxPCBit3
     vme_master->w_swap   = SET;
     vme_master->b_swap   = SET;
     vme_master->size     = size/4;
     vme_master->vme_base = vme_base;
     vme_master->am       = am;
     if( Alloc_VMEMaster( rod_vme->b3dpt_dev, vme_master ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
#ifdef PowerPC_Lynx3_0
     vme_master->vme_addr = vme_base;
     vme_master->size     = size/4;
     vme_master->space    = A32;
     if( Alloc_VMEMaster( vme_master, vme_master->size ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
#ifdef MVME167_Lynx2_5_1
     vme_master->vme_addr = vme_base;
     vme_master->size     = size/4;
     vme_master->space    = A32;
     if( Alloc_VMEMaster( vme_master, vme_master->size ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
     break;
   }

   case Slave_VME  : {
     VMESlave *vme_slave;
     rod_dev[ ii ].vme_obj = ( char* )malloc( sizeof( VMESlave ));
     if( rod_dev[ ii ].vme_obj == NULL ){
       printf("\n[ %-20s ]","Not Enought Memory to make Slave Object");
       return( FAIL );
     }
     rod_dev[ ii ].stat = SUCCESS;
     vme_slave = ( VMESlave* )rod_dev[ ii ].vme_obj;
#ifdef LinuxPCBit3
     vme_slave->w_swap = SET;
     vme_slave->b_swap = SET;
     vme_slave->size   = size/4; 
     if( Alloc_VMESlave( rod_vme->b3dpt_dev, vme_slave ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
     /* 
	04/April/99 Policy: On LynxOS We Only make Static VME Slave Objects. 
	Since in Linux we can not make Static VME Slave Objects. Some day an 
	update of Linux driver will allow this feature and it's better to 
	proccess this kind of object in a separate case inside this switch 
     */
#ifdef PowerPC_Lynx3_0
     vme_slave->vme_addr = vme_base;
     vme_slave->size     = size/4;
     sprintf( vme_slave->id, "%s", dummy );
     if( Open_VMESlave( vme_slave, vme_slave->id, size/4 ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
#ifdef MVME167_Lynx2_5_1
     vme_slave->vme_addr = vme_base;
     vme_slave->size     = size/4;
     sprintf( vme_slave->id, "%s", dummy );
     if( Open_VMESlave( vme_slave, vme_slave->id, size/4 ) != SUCCESS ){
       rod_dev[ ii ].stat = FAIL;
     }
#endif
     break;
   }

   case dma_VME    : {
     DMA_Dev *dma_dev;
     rod_dev[ ii ].vme_obj = ( char* )malloc( sizeof( DMA_Dev ) );
     if( rod_dev[ ii ].vme_obj == NULL ){
       printf("\n[ %-20s ]","Not Enought Memory to make DMA Object");
       return( FAIL );
     }
     rod_dev[ ii ].stat = SUCCESS;
     dma_dev = ( DMA_Dev* )rod_dev[ ii ].vme_obj;

#ifdef LinuxPCBit3
     dma_dev->w_swap   = SET;
     dma_dev->nb_swap  = CLEAR;
     dma_dev->am       = 0x0f;
     dma_dev->dir      = DMA_VME2PCI;
     dma_dev->size     = size/4;
     dma_dev->vme_base = vme_base;
     if( DMA_Alloc( rod_vme->b3dpt_dev, 
		    rod_vme->b3csr_dev, dma_dev ) != SUCCESS )
       rod_dev[ ii ].stat = FAIL;
#endif
#ifdef PowerPC_Lynx3_0
     dma_dev->vme_addr = vme_base;
     dma_dev->size     = size/4;
     if( DMA_Alloc( rod_vme->universe_dev, dma_dev ) != SUCCESS )
       rod_dev[ ii ].stat = FAIL;
#endif
#ifdef MVME167_Lynx2_5_1
     dma_dev->vme_addr = vme_base;
     dma_dev->size     = size/4;
     if( DMA_Alloc( rod_vme->chip2_dev, dma_dev ) != SUCCESS )
       rod_dev[ ii ].stat = FAIL;
#endif
     break;
   }
   }
 }
 return( SUCCESS );

}
/* ------------------------------------------------------------------------- */

/* --- Return VME Object virtual Pointers from internal List(Data Base) ---- */
int
Get_VMEObject( ROD_vme *vme, char *arg, word32 *addr, word32 *idx ){
int ii;
VME_Dev *rod_dev;
Search_Key *key;

  rod_dev = vme->rod_dev;
  key = ( Search_Key* )arg;
  for( ii = 0 ; ii < vme->no_dev ; ii++ ){
    if( ( strcmp( rod_dev[ ii ].machine, key->machine ) == 0 )&&
	( strcmp( rod_dev[ ii ].id, key->dev_id ) == 0 ) ){
      *idx = ii;
      switch( rod_dev[ ii ].obj_type ){
      case Master_VME:{ 
	VMEMaster *vme_master_dev;	
	if( rod_dev[ ii ].stat == SUCCESS ){
	  vme_master_dev = ( VMEMaster* )rod_dev[ ii ].vme_obj;
	  *addr = ( word32 )vme_master_dev->v_ptr;
	  return( SUCCESS );
	}
	else
	  return( FAIL );
      }
      case Slave_VME: {
	VMESlave *vme_slave_dev;
	if( rod_dev[ ii ].stat == SUCCESS ){
	  vme_slave_dev = ( VMESlave* )rod_dev[ ii ].vme_obj;
	  *addr = ( word32 )vme_slave_dev->v_ptr;
	  return( SUCCESS );
	}
	else
	  return( FAIL );
      }
      case dma_VME:{
	DMA_Dev *dma_dev;
	if( rod_dev[ ii ].stat == SUCCESS ){
	  dma_dev = ( DMA_Dev* )rod_dev[ ii ].vme_obj;
	  *addr = ( word32 )dma_dev->v_ptr;
	  return( SUCCESS );
	}
	else
	  return( FAIL );
      }
      }
    }
  }
  return( FAIL );
}


int
Dump_Object( VME_Dev *vme_dev ){
char str[ 25 ];

 printf("\n[ %-20s ]", "ROD: Hardware: VME: Object Dump" );
 printf("\n[ %-15s: %-20s ]", "Object_ID", vme_dev->id );
 printf("\n[ %-15s: %-20s ]", "Machine IP", vme_dev->machine );
 if( vme_dev->stat == SUCCESS )
   sprintf( str, "%s", "Allocated" );
 else
   sprintf( str, "%s", "Refused" );
 printf("\n[ %-15s: %-20s ]", "Object Status", str ); 
 switch( vme_dev->obj_type ){
 case Master_VME :{
   printf("\n[ %-15s: %-20s ]", "Type", "VME Master" );
   Dump_MasterInfo( ( VMEMaster* )vme_dev->vme_obj );
 }
 break;
 case Slave_VME  :{
   printf("\n[ %-15s: %-20s ]", "Type", "VME Slave" );
#ifdef LinuxPCBit3
   Dump_SlaveInfo( ( VMESlave* )vme_dev->vme_obj );
#endif
#ifdef PowerPC_Lynx3_0
   Dump_SlaveInfo( ( char* )vme_dev->vme_obj, STATIC_SLAVE );
#endif
#ifdef MVME167_Lynx2_5_1
   Dump_SlaveInfo( ( char* )vme_dev->vme_obj, STATIC_SLAVE );
#endif
 }
 break;
 case dma_VME    :{
   printf("\n[ %-15s: %-20s ]", "Type", "DMA VME" );
   Dump_DMAInfo( ( DMA_Dev* )vme_dev->vme_obj );
 }
 }
 return( SUCCESS );

}
