#include "DUnit.h"


int
DU_Open( Distribution_Unit *du ){
int status = FAIL;

 sprintf( du->id, "%s", "DUEngine" );
 du->hit = CLEAR;
 pthread_mutex_init ( &du->ib_mtx, NULL );
 pthread_cond_init  ( &du->ib_cv, NULL  );
 status = SUCCESS;
 return( status );

}


int
DU_Close( Distribution_Unit *du ){
int status = FAIL;

 if( *du->stat == UNIT_RUNNING ){
   *du->stat = UNIT_LOST;
   pthread_cancel( du->phy_th );
 }
 printf("\n[ %-18s ][ Hits Received ][%d]", du->id, du->hit );
 fflush( stdout );
 status = SUCCESS;
 return( status );
  
}


int
DU_Run( Distribution_Unit *du ){
int status = FAIL;
pthread_attr_t  pth_attr;  
 
 du->du_ready = CLEAR; 
/* Thread Management */
#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "DU Thread Attr" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "DU Thread Attr" );
   return( status );
 }
#endif

#if defined( LinuxPCBit3 ) 
 if( pthread_create( &du->phy_th, &pth_attr, ( void* )DU_Engine, du ) != 0 ){
   perror( "DU Thread Create" );
   return( status );
 }
#else
 if( pthread_create( &du->phy_th, pth_attr, ( void* )DU_Engine, du ) != 0 ){
   perror( "DU Thread Create" );
   return( status );
 }
#endif
 printf( "\n[ %-15s ][ Running ]", du->id );
 fflush( stdout );
 status = SUCCESS;
 return( status );

}


int
DU_Engine( void *par ){
w32 ii;
Distribution_Unit *du;

#ifdef DEBUG
struct timeval tv;
float offset = 4.0;
#endif

 du = ( Distribution_Unit* )par;
 printf("\n[ %-15s ][ %ld ]", du->id ,( w32 )getpid( ) );
 fflush( stdout );
 *( du->stat ) = UNIT_RUNNING;
 du->hit = CLEAR;

 while( SET ){
   if( *( du->quit ) == SET )
     break;
   /* DU Synchronism --------------------------------------------------- */
   pthread_mutex_lock( du->iu_mtx );
   if( *( du->iu_ready ) == CLEAR ){ 
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 0.+offset);
     fflush( stdout );
#endif
     du->du_ready = SET;
     pthread_cond_wait( du->iu_cv, du->iu_mtx );
     du->du_ready = CLEAR;
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 1.+offset);
     fflush( stdout );
#endif     
   }
   else{
#ifdef DEBUG
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 2.+offset);
     fflush( stdout );
#endif
     pthread_cond_signal( du->iu_cv );
   }
   pthread_mutex_unlock( du->iu_mtx );

   /* PU Barrier -------------------------------------------------------- */
   pthread_mutex_lock( &du->ib_mtx );
   if( --( du->i_barrier  ) == 0 )
     du->i_barrier = __NoProcessors__ + 1;
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 3.+offset);
     fflush( stdout );
#endif
     pthread_cond_wait( &du->du_cv, &du->ib_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 4.+offset);
     fflush( stdout );
#endif
   }
   /* Push Current Event to PU Buffers Now. PU-Barrier is Ready --------- */

   for( ii = 0; ii < du->nopu ; ii++ ){
	memcpy( du->unit[ ii ].ib, 
	du->i_buf + ( du->unit[ ii ].i_size * ii ),
	du->unit[ ii ].i_size * sizeof( w32 ) );
   }
   /* ------------------------------------------------------------------- */
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nDU:\t%ld\t%0.1f", (w32)tv.tv_usec, 5.+offset);
     fflush( stdout );
#endif
   pthread_cond_broadcast( &du->ib_cv );
   pthread_mutex_unlock( &du->ib_mtx );
  
   /* ------------------------------------------------------------------- */
   du->hit++ ;
 }
 *( du->stat ) = UNIT_LOST;
 printf("\n[ DU  ][ %-20s ]","Quit Active" );
 fflush( stdout );
 return( SUCCESS );

}


