#include "EBUnit.h"

int
EBU_Open( Event_Builder_Unit *ebu ){
int status = FAIL;

 sprintf( ebu->id, "%s", "EBEngine" );
 ebu->hit = CLEAR;
 pthread_mutex_init ( &ebu->ebu_mtx, NULL );
 pthread_cond_init  ( &ebu->ebu_cv, NULL );
 pthread_cond_init  ( &ebu->pu_cv, NULL  );

 pthread_mutex_init ( &ebu->ob_mtx, NULL );
 pthread_cond_init  ( &ebu->ob_cv, NULL );
 status = SUCCESS;
 return( status );

}

int
EBU_Close( Event_Builder_Unit *ebu ){
int status = FAIL;

 if( *ebu->stat == UNIT_RUNNING ){
   *ebu->stat = UNIT_LOST;
   pthread_cancel( ebu->phy_th );
 }
 printf("\n[ %-18s ][ Hits Received ][%d]", ebu->id, ebu->hit ); 
 fflush( stdout );
 status = SUCCESS;
 return( status );
  
}

int
EBU_Run( Event_Builder_Unit *ebu ){
int status = FAIL;
pthread_attr_t  pth_attr;

#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "EBU Thread Attr" );
   return( status );
 }
 if(pthread_create( &ebu->phy_th, &pth_attr,( void* )EBU_Engine, ebu )!= 0){
   perror( "EBU Thread Create" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "EBU Thread Attr" );
   return( status );
 }
 if(pthread_create( &ebu->phy_th, pth_attr,( void* )EBU_Engine, ebu ) != 0){
   perror( "EBU Thread Create" );
   return( status );
 }
#endif
 
 printf( "\n[ %-15s ][ Running ]", ebu->id );
 fflush( stdout );
 status = SUCCESS;
 return( status );

}

int
EBU_Engine( void *par ){
TTC_Info  ttc_info;
w32 ii, *next;
Event_Builder_Unit *ebu;

#ifdef DEBUG
struct timeval tv;
float offset=4.5;
#endif

 ebu = ( Event_Builder_Unit* )par;
 printf("\n[ %-15s ][ %ld ]",ebu->id ,( w32 )getpid( ) );
 fflush( stdout );
 *( ebu->stat )  = UNIT_RUNNING;
 ebu->hit  = CLEAR;

 while( SET ){
   if( *( ebu->quit ) == SET )
     break;   
   /* PU Barrier -------------------------------------------------------- */
   pthread_mutex_lock( &ebu->ob_mtx );
   if( --( ebu->o_barrier ) == 0 )
     ebu->o_barrier = __NoProcessors__ + 1;
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 0.+offset);
     fflush( stdout );
#endif    
     pthread_cond_wait( &ebu->pu_cv, &ebu->ob_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 1.+offset);
     fflush( stdout );
#endif
   }   

   /*------= Pull PU Buffers and pack event following ATLAS Format =----- */   
   ttc_info.bcid_no  =  ebu->ttc_info->bcid_no;
   ttc_info.event_no =  ebu->ttc_info->event_no;
   ttc_info.trg_type =  ebu->ttc_info->trg_type;
   next = ebu->o_buf + ATLAS_HEADER + 1;
   for( ii = 0; ii < ebu->nopu ; ii++ ){
#ifdef DEBUG
     if( next == NULL )
       printf("\n[ WARNING ][ DUnit: ROD Output Buffer %ld: NULL ]", ii );
     if( ebu->unit[ ii ].ob == NULL )
       printf("\n[ WARNING ][ DUnit: PU %ld Output Buffer: NULL ]", ii );
#endif 
     memcpy( next, ebu->unit[ ii ].ob,
	     ( ebu->unit[ ii ].o_size ) * sizeof( w32 ) );
     next += ebu->unit[ ii ].o_size;
   }
   Make_ATLAS_Event( ebu->o_buf, &ttc_info );
   ebu->hit++;
   /* ------------------------------------------------------------------- */

#ifdef DEBUG
   gettimeofday( &tv, NULL );
   printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 2.+offset);
   fflush( stdout );
#endif
   pthread_cond_broadcast( &ebu->ob_cv );
   pthread_mutex_unlock( &ebu->ob_mtx );
   /* ------------------------------------------------------------------- */

   /* OUnit FeedBack ---------------------------------------------------- */
   pthread_mutex_lock( &ebu->ebu_mtx );
   if( *( ebu->ou_ready ) == CLEAR ){
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 3.+offset);
     fflush( stdout );
#endif    
     ebu->ebu_ready = SET;
     pthread_cond_wait( &ebu->ebu_cv, &ebu->ebu_mtx );
     ebu->ebu_ready = CLEAR;
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 4.+offset);
     fflush( stdout );
#endif
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nEBU:\t%ld\t%0.1f", (w32)tv.tv_usec, 5.+offset);
     fflush( stdout );
#endif
     pthread_cond_signal( &ebu->ebu_cv );
   }
   pthread_mutex_unlock( &ebu->ebu_mtx );
   /* ------------------------------------------------------------------ */
 } 

 *ebu->stat = UNIT_LOST;
 printf("\n[ EBU ][ %-20s ]","Quit Active" );
 fflush( stdout );
 return( SUCCESS );

}


 

int
Make_ATLAS_Event( w32* ptr, TTC_Info *ttc_info ){
int status = SUCCESS;

 ptr[ 0 ] = H_MARKER;
 ptr[ 1 ] = ATLAS_HEADER;
 ptr[ 2 ] = ATLAS_FORMAT_VERSION;
 ptr[ 3 ] = ATLAS_HEADER + 1;
 ptr[ 4 ] = ROD_MODULE_ID;

 ptr[ 5 ] = ttc_info->event_no;
 ptr[ 6 ] = ttc_info->bcid_no;
 ptr[ 7 ] = ttc_info->trg_type;
 ptr[ 8 ] = CLEAR;
 return( status );

}
