/*******
	$Header: RODhardware Object for the ROD-Module Application
	Created & Maintaned by JB. Juanba.Romance@ific.uv.es $

   26,27/May/99: 
   IUnit IO parametters redefined to work "ONFLY" with different
   I/O Configurations incoming from Controller:

    Target - Driver --------> IO Mode

    Auto - Slink - VME -----> IO Type

   03/June/99:
   VME Data Exchange Protocol implementation starts.

   10/June/99:
   VMElib DMA Interface changes to support parametter size, evt_size 
   parametter provides dynamic Event size using CONFIG command from the ROD 
   Controller. 

   15-28/June/99:
   Driver And Target Mode Already going on

   08/July/99:
   Slink Integration Starts, Routine Changes: IU_Config, IHandshake_Slink

   08-09/November/99:
   Barrier-Machine version integration Starts

   20-28/November/99:
   New Machine version integration

   14-20/December/99:
   VME Data Injection Protocol Redefinition.
   
   21/December/99
   Slink Data Injection Starts.

*******/

#include "IUnit.h"

/* -------------------- Info Provided by ROD Hardware Layer ---------------- */
int
IU_Config( Input_Unit *iu, IOParams *io_cfg ){
int status = FAIL;
VME_Dev  *vme_dev;
DMA_Dev  *dma_dev;
#if defined( PowerPC_Lynx3_0  )
RODSlink *my_slink;
#endif

 if( io_cfg != NULL ){
   iu->io_type = io_cfg->io_type;
   iu->io_mode = io_cfg->io_mode;

#if defined( PowerPC_Lynx3_0  )
   if( iu->io_type == __Slink__ ){
     iu->buffer  = ( w32*  )( io_cfg->buffer );
     /* We Will use VME Slave Buffer as Slink Target Buffer */ 
     iu->rod_slink = io_cfg->slink_addr;
     my_slink = ( RODSlink* )( iu->rod_slink );
     my_slink->dma_dev.buffer  = ( word32*  )( io_cfg->buffer );
     my_slink->dma_dev.bs_addr = io_cfg->paddr;
     my_slink->evt_size = ( ( io_cfg->evt_size ) +2  );
   }
#endif

   if( iu->io_type == __VME__ ){
     if( iu->io_mode == __Target__ ){
       iu->buffer  = ( w32*  )( io_cfg->buffer );
     }
     else{
       if( iu->io_mode == __Driver__ ){
	 iu->dma_dev = io_cfg->dma_dev;
	 printf("\n\n[ Object Address $%08x ]", ( unsigned )iu->dma_dev );
	 fflush( stdout );
	 vme_dev = ( VME_Dev* )( iu->dma_dev );
	 Dump_Object( vme_dev );
	 dma_dev = ( DMA_Dev* )( vme_dev->vme_obj );
	 iu->buffer = ( w32* )( dma_dev->v_ptr );
       }
     }
   }

   if( iu->io_type == __Auto__ )
     iu->buffer  = ( w32*  )( io_cfg->buffer );

   
   iu->l_csr     = ( w32*  )( io_cfg->lcsr_vaddr );
   iu->r_csr     = ( w32*  )( io_cfg->rcsr_vaddr );
   iu->evt_size  = io_cfg->evt_size;
   iu->size      = io_cfg->size;
   
   iu->l_csr[ __IUHandshake__ ] = MASK_COMMAND;
   iu->l_csr[ __IUHandshake__ +1 ] = CLEAR;
   iu->l_csr[ __IUHandshake__ +2 ] = CLEAR;

   iu->r_csr[ __IUHandshake__ +1 ] = CLEAR;
   iu->r_csr[ __IUHandshake__ +2 ] = CLEAR;
   iu->r_csr[ __IUHandshake__ +3 ] = CLEAR;

   *( iu->stat ) = UNIT_READY;
   fflush( stdout );
   status = SUCCESS;
   
 }
 return( status );

}
/* ------------------------------------------------------------------------- */


int
IU_Open( Input_Unit *iu ){
int status = FAIL;

 sprintf( iu->id, "%s", "IUEngine" );
 pthread_mutex_init ( &iu->iu_mtx, NULL ); 
 pthread_cond_init  ( &iu->iu_cv , NULL );
 pthread_mutex_init ( &iu->b_mtx , NULL ); 
 pthread_cond_init  ( &iu->b_cv  , NULL );
 iu->hit = CLEAR;
 status = SUCCESS;
 return( status );

}

int
IU_Close( Input_Unit *iu ){
int status = FAIL;

 if( *( iu->stat ) != UNIT_LOST ){
   *( iu->stat ) = UNIT_LOST;
   usleep( 100000 );
   pthread_cancel( iu->phy_th );   
   pthread_cond_destroy ( &iu->iu_cv );
   pthread_mutex_destroy( &iu->iu_mtx );
   printf("\n[ %-18s ][ Hits Received ][%d]", iu->id, iu->hit );
   fflush( stdout );
   status = SUCCESS;
 }
 
 return( status );
  
}

int
IU_Run( Input_Unit *iu ){
int status = FAIL;
pthread_attr_t  pth_attr;

 if( iu == NULL )
   return( FAIL );
#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "IU Thread Attr" );
   return( status );
 }
 if( pthread_create( &iu->phy_th,&pth_attr,( void* )IU_Engine, iu ) != 0 ){
   perror( "IU Thread Create" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "IU Thread Attr" );
   return( status );
 }
 if( pthread_create( &iu->phy_th, pth_attr, ( void* )IU_Engine, iu ) != 0 ){
   perror( "IU Thread Create" );
   return( status );
 }
#endif
 printf( "\n[ %-15s ][ Running ]", iu->id );
 fflush( stdout );
 return( status );

}

int
IU_Engine( void *par ){
Input_Unit *iu;
#ifdef DEBUG
struct timeval tv;
float offset =3.5;
#endif
 
 iu = ( Input_Unit* )par;
 printf("\n[ %-15s ][ %ld ]", iu->id ,( w32 )getpid( ) );
 fflush( stdout );

 iu->hit = CLEAR;
 *( iu->stat ) = UNIT_RUNNING; 
 
 while( SET ){
   /* Trigger-Barrier Analysis ------------------------------------------ */
   pthread_mutex_lock( &iu->b_mtx );
   if( --( iu->barrier ) == 0 ){ /* Full barrier */
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 0.+offset);
     fflush( stdout );
#endif
     iu->barrier = __BarrierSize__ ;
     pthread_cond_broadcast( &iu->b_cv );
   }
   else{
     /* Wait Barrier Trigger */
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 1.+offset);
     fflush( stdout );
#endif
     pthread_cond_wait( &iu->b_cv, &iu->b_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 2.+offset);
     fflush( stdout );
#endif    
   }
   iu->hit++ ;
   pthread_mutex_unlock( &iu->b_mtx );
   /* Get Data from IO interface ---------------------------------------- */
   if( iu->l_csr[ __IUHandshake__ + 1 ] < __DataSRCFail__ ){
     if( iu->io_type == __Auto__ ){
       IHandshake_Dummy( iu );
     }
     else{
       if( iu->io_type == __VME__ ){
	 if( IHandshake_VME( iu ) == FAIL ){
	   if( iu->l_csr[ __IUHandshake__ + 1 ] == __DataSRCFail__ ){
	   }
	 }
       }
       else{
	 if( iu->io_type == __Slink__ ){
	   if( IHandshake_Slink( iu ) == FAIL ){
	     if( iu->l_csr[ __IUHandshake__ + 1 ] ==  __DataSRCFail__ ){
	     }
	   }
	 }
       }
     }
   }
   /* ------------------------------------------------------------------- */
   /* DU State to be safely signaled */
   pthread_mutex_lock( &iu->iu_mtx );
   if( *( iu->du_ready ) == CLEAR ){
#ifdef DEBUG     
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 3.+offset);
     fflush( stdout );
#endif
     iu->iu_ready = SET;
     pthread_cond_wait( &iu->iu_cv, &iu->iu_mtx );
     iu->iu_ready = CLEAR;
#ifdef DEBUG     
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 4.+offset);
     fflush( stdout );
#endif    
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nIU:\t%ld\t%0.1f", (w32)tv.tv_usec, 5.+offset);
     fflush( stdout );
#endif
     pthread_cond_signal( &iu->iu_cv );
   }
   pthread_mutex_unlock( &iu->iu_mtx );
 }
 *( iu->stat ) = UNIT_LOST;
 printf("\n[ IU  ][ %-20s ]","Quit Active" );
 fflush( stdout );
 return( SUCCESS );

}



int
IHandshake_Dummy( Input_Unit *iu ){
w32 ii;
 

 for( ii = 0 ; ii < iu->evt_size; ii++ ){
   iu->buffer[ ii ] = 1 << ( ii % 32 );
 }

 iu->r_csr[ __IUHandshake__  ] = __TriggerOn__ ;
 ++( iu->r_csr[ __IUHandshake__  +2 ] );
 return( SUCCESS );


}



int
IHandshake_VME( Input_Unit *iu ){
#define __InCTRLTimeOut__ 10000000
int status = SUCCESS;
volatile w32 *l_csr, *r_csr;
w32 time_out;
DMA_Dev *dma_dev;

 l_csr = iu->l_csr;
 r_csr = iu->r_csr;
 if( !( iu->hit % 10000 ) ){
   printf("\n[ Trigger State ]\n[ Hits %04d ]", ( unsigned )iu->hit );
   printf("\n[ Local Trials   ][%6u]", iu->hit );
   printf("\n[ Local Success  ][%6u]", (unsigned)l_csr[ __IUHandshake__ +2 ]); 
   printf("\n[ Local Fails    ][%6u]", (unsigned)l_csr[ __IUHandshake__ +1 ]);
   printf("\n[ %-20s ][%s]", 
	  "VME Input DMA Mode", iu->io_mode ? "Driver" : "Target" );
   fflush( stdout );
 }

 r_csr[ __IUHandshake__  +3 ] = iu->hit;

 
 if( iu->io_mode ==  __Driver__ ){
   dma_dev = ( DMA_Dev* )( (( VME_Dev* )( iu->dma_dev ))->vme_obj );
   if( DMA_Go( dma_dev, VMEtoLOCAL, iu->evt_size*sizeof(word32) ) != SUCCESS ){
     printf("\n[ IUnit: DMA to Data Source: Fails ]" );
     fflush( stdout );
     return( FAIL );
   }
 }
 

 r_csr[ __IUHandshake__  ] = __TriggerOff__ ;
 time_out = __InCTRLTimeOut__ ;
 while( l_csr[ __IUHandshake__ ] != __TriggerOff__  ){
   --time_out;
   if( time_out == 0 ){
     Update_Err_Count( ( w32* )l_csr, ( w32* )r_csr, iu->hit );
     if( !( iu->hit % 10 ) ){
       printf("\n[ IUnit: TriggerOff Lost ][%d]", iu->hit );
       fflush( stdout );
     }
     l_csr[ __IUHandshake__ ] = MASK_COMMAND;
     return( FAIL );
   }
 }
 l_csr[ __IUHandshake__ ] = MASK_COMMAND;

 ++r_csr[ __IUHandshake__  +2 ] ;
 ++l_csr[ __IUHandshake__  +2 ] ;
 return( status );
 
}


int
IHandshake_Slink( Input_Unit *iu ){
int status = SUCCESS;

#if defined( PowerPC_Lynx3_0  )
#define __SlinkTimeOut__ 10000000
RODSlink *my_slink;
volatile w32 *l_csr, *r_csr;
w32 time_out;

 l_csr = iu->l_csr;
 r_csr = iu->r_csr;
 if( !( iu->hit % 10000 ) ){
   printf("\n[ Trigger State ]\n[ Hits %04d ]", ( unsigned )iu->hit );
   printf("\n[ Local Trials   ][%6u]", iu->hit );
   printf("\n[ Local Success  ][%6u]", 
	  ( unsigned )l_csr[ __IUHandshake__ +2 ] ); 
   printf("\n[ Local Fails    ][%6u]", 
	  ( unsigned )l_csr[ __IUHandshake__ +1 ] );
   fflush( stdout );
 }

 r_csr[ __IUHandshake__  +3 ] = iu->hit;


 /* Slink I/O -------------------------------------------------------------- */

 my_slink = ( RODSlink* )( iu->rod_slink );
 status = Slink_DMA_Read( &( my_slink->slink_dev ), 
			  ( char* )( &( my_slink->dma_dev ) ), 
			  my_slink->evt_size, PCItoPHY_M );
 if( status != SUCCESS ){
   printf("\n[ ROD: Software: DMASlink: Fails ][%d]", iu->hit );
   fflush( stdout );
   return( FAIL );
 }

 /* ------------------------------------------------------------------------ */


 r_csr[ __IUHandshake__  ] = __TriggerOff__ ;
 time_out = __SlinkTimeOut__ ;
 while( l_csr[ __IUHandshake__ ] != __TriggerOff__  ){
   --time_out;
   if( time_out == 0 ){
     Update_Err_Count( ( w32* )l_csr, ( w32* )r_csr, iu->hit );
     if( !( iu->hit % 100 ) ){
       printf("\n[ Slink-IUnit: TriggerOff Lost ][%d]", iu->hit );
       fflush( stdout );
     }
     l_csr[ __IUHandshake__ ] = MASK_COMMAND;
     return( FAIL );
   }
 }
 l_csr[ __IUHandshake__ ] = MASK_COMMAND;

 ++r_csr[ __IUHandshake__  +2 ] ;
 ++l_csr[ __IUHandshake__  +2 ] ;
 
#endif

 return( status ); 

}


void
Update_Err_Count( w32* l_csr, w32* r_csr, w32 hit ){
  
  ++( l_csr[ __IUHandshake__ +1 ] );
  ++( r_csr[ __IUHandshake__ +1 ] );

}
