/**
    $Header: ROD.c object for the ROD Module Application
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $
    
    10/December/99
     Added ROD_Stop interface to dynamic Threads StartUp

     Added sys_stat flag "alias" at Machine object level as a pointer to 
     the ROD object field
**/

#include "Machine.h"

void Decode_CFGword( w32 cfg_word );

void
Decode_CFGword( w32 cfg_word ){
Config_Word cfg;
char str[ 20 ];

 cfg.all = cfg_word;
 printf("\n[ %-30s ][ %-20s ]", "I/O Mode", 
	cfg.bit.io_mode ? "Driver" : "Target" );
 printf("\n[ %-30s ][ %-20s ]", "Data Source", 
	cfg.bit.src_type ? "FrontEnd" : "Controller" );
 switch( cfg.bit.io_type ){
 case __Slink__ : sprintf( str, "%s", "Slink" );break;
 case __VME__   : sprintf( str, "%s", "VME" )  ;break;
 case __Auto__  : sprintf( str, "%s", "Auto" ) ;break;
 default: sprintf( str, "Unknown Mode" ) ;break;
 }
 printf("\n[ %-30s ][ %-20s ]", "I/O Type", str );
 
 switch( cfg.bit.format_id ){
 case __Calibration__ : sprintf( str, "%s", "Calibration" );break;
 case __ATLAS__       : sprintf( str, "%s", "ATLAS" )  ;break;
 case __Monitor__     : sprintf( str, "%s", "Monitor" ) ;break;
 case __ByPass__      : sprintf( str, "%s", "ByPass" ) ;break;
 default: sprintf( str, "Unknown Mode" ) ;break; 
 }
 printf("\n[ %-30s ][ %-20s ]", "I/O Format",  str );
 sprintf( str, "%04d",  cfg.bit.no_channels ) ;
 printf("\n[ %-30s ][ %-20s ]", "NoSamples",  str );
 sprintf( str, "%04d",  cfg.bit.no_samples ) ;
 printf("\n[ %-30s ][ %-20s ]", "NoChannels",  str );

}



int
Machine_Config( Machine *mch, w32 *cfg_data ){
char *str[ ]={ 
  "Input Unit Configuration Word",
  "Output Unit Configuration Word"
};
int status = FAIL;
w32 ii, jj;

 printf("\n\n[ Machine: CONFIG ]");
 for( ii = 0 ; ii < 2 ; ii++ ){
   mch->cfg[ ii ] =  cfg_data[ ii ];
   printf("\n[ %-30s ][ $%08x ]", str[ ii ], ( unsigned )mch->cfg[ ii ] );
   Decode_CFGword( cfg_data[ ii ] );
 }
 printf("\nROD: Software: Machine: Trying to config IO-Units" );
 fflush( stdout );
 if( IU_Config( &mch->iu, ( IOParams* )( &cfg_data[ 2 ] ) ) == SUCCESS )
   if( OU_Config( &mch->ou, ( IOParams* )( &cfg_data[ 12 ] ) ) == SUCCESS ){
     status = SUCCESS;
     ( mch->sys_stat->bit.machine ) = MACHINE_READY;
   }
   else
     printf("\n[ OU SetUp: FAILs ]");
 else
   printf("\n[ IU SetUp: FAILs ]");
 printf("\nROD: Software: Machine: IO-Units: Config: SUCESS" );
 fflush( stdout );

 for( ii = 0 ; ii < mch->pu.nopu ; ii++ ){
   for( jj = 0; jj < 2 ; jj++ ){
     mch->pu.unit[ ii ].cfg[ jj ] = cfg_data[ jj ];
   }
 }
 mch->ttc.trg_mode = cfg_data[ 19 ];
 return( status );

}



int
Machine_Open( Machine **mch, w32 csr_addr, w32 sys_addr ){
int status;
w32 ii;
 
 status = SUCCESS; 
 *mch = ( Machine* )malloc( sizeof( Machine ) );
 if( *mch == NULL ){
   printf("\n[ CAN Not Alloc Memory for Machine ROD" );
   status = FAIL;
 }
 ( *mch )->sys_stat = ( System_State* )sys_addr;
 ( *mch )->sys_stat->bit.machine = MACHINE_DOWN;
 ( *mch )->l_csr = csr_addr;

 if( IU_Open( &( ( *mch )->iu )  ) != SUCCESS ){
   printf("\n[ Machine Fails Open IU ]" );
   fflush( stdout );
   status = FAIL;
 }
 if( DU_Open( &( ( *mch )->du ) ) != SUCCESS ){
   printf("\n[ Machine Fails Open PU ]" );
   fflush( stdout );
   status = FAIL;
 }
 if( TTC_Open( &( ( *mch )->ttc ) ) != SUCCESS ){
   printf("\n[ Machine Fails Open TTC ]" );
   fflush( stdout );
   status = FAIL;
 }
 if( EBU_Open( &( ( *mch )->ebu ) ) != SUCCESS ){
   printf("\n[ Machine: Open EBU: FAIL ]" );
   fflush( stdout );
   status = FAIL;
 }
 if( OU_Open( &( ( *mch )->ou ) ) != SUCCESS ){
   printf("\n[ Machine: Open OU: FAIL ]" );
   fflush( stdout ); 
   status = FAIL;
 }
 if( PU_Open( &( ( *mch )->pu ) ) != SUCCESS ){
   printf("\n[ Machine: Open PU: FAIL ]" );
   fflush( stdout );
   status = FAIL;
 }

 /* Machine Routing */
 /* TTCUnit */
 ( *mch )->ttc.quit     = &( ( *mch )->mch_stat.quit   );
 ( *mch )->ttc.stat     = &( ( *mch )->mch_stat.ttc    );
 ( *mch )->ttc.b_mtx    = &( ( *mch )->iu.b_mtx        );
 ( *mch )->ttc.b_cv     = &( ( *mch )->iu.b_cv         );
 ( *mch )->ttc.barrier  = &( ( *mch )->iu.barrier      );
 ( *mch )->ttc.trg_veto = &( ( *mch )->trg_veto        );
 /* IUnit */
 ( *mch )->iu.ttc_info  = &( ( *mch )->ttc.info        );
 ( *mch )->iu.quit      = &( ( *mch )->mch_stat.quit   );
 ( *mch )->iu.stat      = &( ( *mch )->mch_stat.iu     );
 ( *mch )->iu.du_ready  = &( ( *mch )->du.du_ready     );
 
 /* DUnit */
 ( *mch )->du.ttc_info  = &( ( *mch )->ttc.info        );
 ( *mch )->du.iu_mtx    = &( ( *mch )->iu.iu_mtx       );
 ( *mch )->du.iu_cv     = &( ( *mch )->iu.iu_cv        ); 
 ( *mch )->du.quit      = &( ( *mch )->mch_stat.quit   );
 ( *mch )->du.stat      = &( ( *mch )->mch_stat.du     );
 ( *mch )->du.iu_ready = &( ( *mch )->iu.iu_ready  );

 /* EBUnit */
 ( *mch )->ebu.quit     = &( ( *mch )->mch_stat.quit   );
 ( *mch )->ebu.stat     = &( ( *mch )->mch_stat.ebu    );
 ( *mch )->ebu.ttc_info = &( ( *mch )->ttc.info        );
 ( *mch )->ebu.ou_ready = &( ( *mch )->ou.ou_ready     );

 /* OUnit */
 ( *mch )->ou.quit      = &( ( *mch )->mch_stat.quit   );
 ( *mch )->ou.stat      = &( ( *mch )->mch_stat.ou     );
 ( *mch )->ou.b_mtx     = &( ( *mch )->iu.b_mtx        );
 ( *mch )->ou.b_cv      = &( ( *mch )->iu.b_cv         );
 ( *mch )->ou.barrier   = &( ( *mch )->iu.barrier      );
 ( *mch )->ou.ttc_info  = &( ( *mch )->ttc.info        );
 ( *mch )->ou.ttc_mtx   = &( ( *mch )->ttc.ttc_mtx     );
 ( *mch )->ou.ttc_cv    = &( ( *mch )->ttc.ttc_cv      );
 ( *mch )->ou.ebu_mtx   = &( ( *mch )->ebu.ebu_mtx     );
 ( *mch )->ou.ebu_cv    = &( ( *mch )->ebu.ebu_cv      );
 ( *mch )->ou.ebu_ready = &( ( *mch )->ebu.ebu_ready   );
 /* PUnit Matrix */
 for( ii = 0; ii < ( *mch )->pu.nopu ; ii++ ){
   ( *mch )->pu.unit[ ii ].ob_mtx    = &( ( *mch )->ebu.ob_mtx        );
   ( *mch )->pu.unit[ ii ].ob_cv     = &( ( *mch )->ebu.ob_cv         );
   ( *mch )->pu.unit[ ii ].o_barrier = &( ( *mch )->ebu.o_barrier     );
   ( *mch )->pu.unit[ ii ].ebu_cv    = &( ( *mch )->ebu.pu_cv         );

   ( *mch )->pu.unit[ ii ].ib_mtx    = &( ( *mch )->du.ib_mtx         );
   ( *mch )->pu.unit[ ii ].ib_cv     = &( ( *mch )->du.ib_cv          );
   ( *mch )->pu.unit[ ii ].du_cv     = &( ( *mch )->du.du_cv          );
   ( *mch )->pu.unit[ ii ].i_barrier = &( ( *mch )->du.i_barrier      );

   ( *mch )->pu.unit[ ii ].quit      = &( ( *mch )->mch_stat.quit     );
   ( *mch )->pu.unit[ ii ].stat      = &( ( *mch )->mch_stat.pu[ ii ] );
   *(( *mch )->pu.unit[ ii ].stat )  = UNIT_OPEN;
 }

 *(( *mch )->iu.stat  ) = UNIT_OPEN;
 *(( *mch )->du.stat  ) = UNIT_OPEN;
 *(( *mch )->ttc.stat ) = UNIT_OPEN;
 *(( *mch )->ebu.stat ) = UNIT_OPEN;
 *(( *mch )->ou.stat  ) = UNIT_OPEN;
 
 ( *mch )->mch_stat.quit = CLEAR;
 ( *mch )->sys_stat->bit.machine = MACHINE_UP;
#ifdef VERBOSE
 printf("\n[ Virtual ROD Machine is UP  ]\n" );
 fflush( stdout );
#endif
 return( status );

}


int 
Machine_Close( Machine *mch ){
int status = SUCCESS;

 if( mch == NULL )
   return( FAIL );

 if( mch->sys_stat->bit.machine == MACHINE_DOWN ){
   printf("[ %-20s ]","Machine Was DOWN. Nothing To Do" );
   return( SUCCESS );
 } 

 status = PU_Close( &mch->pu );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]", "PU Close" );
 
 status = TTC_Close( &mch->ttc );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]", "TTC Close" );

 status = DU_Close( &mch->du );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]","DU Close");
 fflush( stdout );

 status = IU_Close( &mch->iu );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]","IU Close");
 fflush( stdout );

 status = EBU_Close( &mch->ebu );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]","EBU Close");
 fflush( stdout );

 status = OU_Close( &mch->ou );
 if( status == SUCCESS ) 
   printf("\n[ %-9s ]:[ SUCCESS ]","OU Close");

 /* Machine is Stopped. Release threads resources */
 pthread_mutex_destroy( &mch->ttc.ttc_mtx );
 pthread_cond_destroy ( &mch->ttc.ttc_cv );

 pthread_cond_destroy ( &mch->iu.iu_cv  );
 pthread_mutex_destroy( &mch->iu.iu_mtx );

 pthread_cond_destroy ( &mch->iu.b_cv   );
 pthread_mutex_destroy( &mch->iu.b_mtx  );

 pthread_cond_destroy ( &mch->ebu.ebu_cv );
 pthread_mutex_destroy( &mch->ebu.ebu_mtx );

 pthread_cond_destroy ( &mch->ebu.ob_cv );
 pthread_cond_destroy ( &mch->ebu.pu_cv );
 pthread_mutex_destroy( &mch->ebu.ob_mtx );

 pthread_cond_destroy ( &mch->du.ib_cv );
 pthread_cond_destroy ( &mch->du.du_cv );
 pthread_mutex_destroy( &mch->du.ib_mtx );
 
 free( mch );
#ifdef VERBOSE
 printf("\n[ %-20s : SUCCESS ]\n","Machine Close");
 fflush( stdout );
#endif
 return( status );

}



int
Machine_RunBaby( Machine *mch ){
int status = SUCCESS;
  
 if( mch->sys_stat->bit.machine != MACHINE_READY ){
   printf("\n[ Call Machine_SetUp Before ]" );
   fflush( stdout );
   return( FAIL );
 }

 /* Last Routing Cause we know Here which are the Input/Output 
    virtual pointers */
 mch->du.unit       = mch->pu.unit;
 mch->ebu.unit      = mch->pu.unit;
 mch->du.i_buf      = ( w32* )( mch->iu.buffer );
 mch->ebu.o_buf     = ( w32* )( mch->ou.buffer );
 mch->du.nopu       = mch->pu.nopu;
 mch->ebu.nopu      = mch->pu.nopu;

 mch->iu.barrier    = __BarrierSize__ ;
 mch->ebu.o_barrier = __NoProcessors__ + 1;
 mch->du.i_barrier  = __NoProcessors__ + 1;
 
 mch->ou.ou_ready   = CLEAR;
 mch->ebu.ebu_ready = CLEAR;
 mch->iu.iu_ready   = CLEAR;
 mch->du.du_ready   = CLEAR;

 mch->ttc.l_csr     = ( w32* )( mch->iu.l_csr );
 mch->ttc.r_csr     = ( w32* )( mch->iu.r_csr );
 mch->trg_veto      = CLEAR;
 
 status = DU_Run  ( &mch->du    );
 status = IU_Run  ( &mch->iu    );
 status = EBU_Run ( &mch->ebu   );
 status = OU_Run  ( &mch->ou    );
 status = PU_Run  ( &mch->pu    );
 usleep( 100000 );
 status = TTC_Run ( &mch->ttc   );
 mch->sys_stat->bit.machine = MACHINE_RUNNING;
 
 /* Maybe below must be driven by Controller CRATE instead Machine itself */
 return( status );

}


int
Machine_Dump( Machine *mch ){
int status = FAIL;
w32 ii, jj, *l_csr;

 l_csr = ( w32* )mch->l_csr;

 ii = 1;
 l_csr[ ii++ ] = mch->pu.nopu;
 l_csr[ ii++ ] = mch->iu.hit;
 l_csr[ ii++ ] = mch->ttc.hit;
 l_csr[ ii++ ] = mch->du.hit;
 l_csr[ ii++ ] = mch->ebu.hit;
 l_csr[ ii++ ] = mch->ttc.trg_mode;
 /* l_csr[ ii++ ] = mch->mch_stat.all; */
 l_csr[ ii++ ] = mch->sys_stat->all;
 l_csr[ ii++ ] = mch->iu.io_mode;
 l_csr[ ii++ ] = mch->iu.io_type;
 l_csr[ ii++ ] = mch->iu.size*sizeof( w32 );
 l_csr[ ii++ ] = mch->ou.io_mode;
 l_csr[ ii++ ] = mch->ou.io_type;
 l_csr[ ii++ ] = mch->ou.size*sizeof( w32 );
 l_csr[ ii++ ] = mch->cfg[ 0 ];
 l_csr[ ii++ ] = mch->cfg[ 1 ];
 jj = ii;
 for( ii = 0 ; ii < mch->pu.nopu; ii++, jj += 2 ){
   l_csr[ jj   ] = mch->pu.unit[ ii ].i_size * sizeof( w32 );
   l_csr[ jj+1 ] = mch->pu.unit[ ii ].o_size * sizeof( w32 );
 }
 status = SUCCESS;

 return( status );

}

