#include "OUnit.h"


int
OU_Config( OutPut_Unit *ou, IOParams *io_cfg ){
int status = FAIL;

 if( io_cfg != NULL ){ 
   status = SUCCESS;
   /* Below Could be substituted by bcopy  ? */
   ou->io_type = io_cfg->io_type;
   ou->io_mode = io_cfg->io_mode;
   ou->buffer  = ( w32*  )( io_cfg->buffer );
   ou->l_csr   = ( w32*  )( io_cfg->lcsr_vaddr );
   ou->r_csr   = ( w32*  )( io_cfg->rcsr_vaddr );
   ou->dma_dev = io_cfg->dma_dev;
   ou->size    = io_cfg->size;
   *( ou->stat ) = UNIT_READY;
 }
 return( status );

}


int
OU_Open( OutPut_Unit *ou ){
int status = FAIL;

 sprintf( ou->id, "%s", "OUEngine" );
 ou->hit = CLEAR;
 status = SUCCESS;
 return( status );

}


int
OU_Close( OutPut_Unit *ou ){
int status = FAIL;

 if(*( ou->stat ) == UNIT_RUNNING ){
   *( ou->stat ) = UNIT_LOST;
   pthread_cancel( ou->phy_th );
 }
 printf("\n[ %-18s ][ Hits Received ][%d]", ou->id, ou->hit );
 fflush( stdout );
 status = SUCCESS;
 return( status );
  
}


int
OU_Run( OutPut_Unit *ou ){
int status = FAIL;
pthread_attr_t  pth_attr;  

 ou->ou_ready = CLEAR;
 if( *( ou->stat ) == UNIT_READY ){
#if defined( LinuxPCBit3 )
   if( pthread_attr_init( &pth_attr ) != 0 ){
     perror( "OU Thread Attr" );
     return( status );
   }
   if( pthread_create( &ou->phy_th, &pth_attr, ( void* )OU_Engine, ou )!=0 ){
     perror( "OU Thread Create" );
     return( status );
   }
#else
   if( pthread_attr_create( &pth_attr ) != 0 ){
     perror( "OU Thread Attr" );
     return( status );
   }
   if( pthread_create( &ou->phy_th, pth_attr, ( void* )OU_Engine, ou ) != 0 ){
     perror( "OU Thread Create" );
     return( status );
   }
#endif
   status = SUCCESS;
    printf( "\n[ %-15s ][ Running ]", ou->id );
 }
 else{
   printf("\n[ Call OU_SetUp before RUN ]"); 
 }
 fflush( stdout );
 return( status );

}

int
OU_Engine( void *par ){
OutPut_Unit *ou;

#ifdef DEBUG
struct timeval tv;
float offset = 5;
#endif

 ou = ( OutPut_Unit* )par;
 printf("\n[ %-15s ][ %ld ]", ou->id ,( w32 )getpid( ) );
 fflush( stdout ); 
 *( ou->stat ) = UNIT_RUNNING;
 ou->hit  = CLEAR;
 
 while( SET ){
   if( *( ou->quit ) == SET )
     break;
   /* TRG Barrier Analysis ---------------------------------------- */
   pthread_mutex_lock( ou->b_mtx );
   if( --( *ou->barrier ) == 0 ){
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 0.+offset );
     fflush( stdout );
#endif
     *( ou->barrier ) = __BarrierSize__ ;
     pthread_cond_broadcast( ou->b_cv );
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 1.+offset );
     fflush( stdout );
#endif
     pthread_cond_wait( ou->b_cv, ou->b_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 2.+offset );
     fflush( stdout );
#endif
   }
   pthread_mutex_unlock( ou->b_mtx );
   /* ------------------------------------------------------------------- */

   /* EBU FeedBack ------------------------------------------------------ */
   pthread_mutex_lock( ou->ebu_mtx );
   if( *( ou->ebu_ready ) == CLEAR ){
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 3.+offset );
     fflush( stdout );
#endif
     ou->ou_ready = SET;
     pthread_cond_wait( ou->ebu_cv, ou->ebu_mtx );
     ou->ou_ready = CLEAR;
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 4.+offset );
     fflush( stdout );
#endif
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 5.+offset );
     fflush( stdout );
#endif
     pthread_cond_signal( ou->ebu_cv );
   }
   /* ---------------------= OU Processing =------------------- */
   if( ou->io_type == __Auto__ ){
     OHandshake_Dummy( );
   }
   else
     if( ou->io_type == __VME__ )
       OHandshake_VME( ou );
     else
       if( ou->io_type == __Slink__ )
	 OHandshake_Slink( ou );

   pthread_mutex_unlock( ou->ebu_mtx );
   /* ------------------------------------------------------------------- */

   /* TTC FeedBack ------------------------------------------------------ */   
#ifdef DEBUG
   gettimeofday( &tv, NULL );
   printf("\nOU:\t%ld\t%0.1f", (w32)tv.tv_usec, 6.+offset);
   fflush( stdout );
#endif
   pthread_mutex_lock( ou->ttc_mtx );
   pthread_cond_signal( ou->ttc_cv );
   pthread_mutex_unlock( ou->ttc_mtx );
   ++ou->hit;
   /* ------------------------------------------------------------------- */
 }
 
 *( ou->stat ) = UNIT_LOST;
 printf("\n[ OU  ][ %-20s ]","Quit Active" );
 fflush( stdout );
 return( SUCCESS );

}


int
OHandshake_Dummy( ){
int status = SUCCESS;

  return( status );

}


int
OHandshake_VME( OutPut_Unit *ou ){
int status = SUCCESS;

 return( status );
 
}


int
OHandshake_Slink( OutPut_Unit *ou ){
int status = SUCCESS;

 return( status ); 

}
