#include "PUnit.h"


int
PU_Open( Processor_Unit *pu ){
int clean, status = FAIL;
w32 ii;
 
 /* --------------------------- Memory Management -------------------------- */
 pu->nopu   = __NoProcessors__ ; 
 pu->unit = ( PUnit* )malloc( sizeof( PUnit ) * pu->nopu );
 if( pu->unit == NULL ){
   printf("\n[ Not Enought memory to Build Processors Farm ]" );
   return( status );
 }
 
 for( ii = 0; ii < pu->nopu ; ii++ ){
   pu->unit[ ii ].i_size = PU_IB_SIZE;
   pu->unit[ ii ].o_size = PU_OB_SIZE;
   sprintf( pu->unit[ ii ].id, "%s%d", "PUnit-", ( int )ii );
   pu->unit[ ii ].ib  = ( w32* )
     malloc( sizeof( w32 )*pu->unit[ ii ].i_size );
   pu->unit[ ii ].ob  = ( w32* )
     malloc( sizeof( w32 )*pu->unit[ ii ].o_size );
   pu->unit[ ii ].pat = ( w32* )
     malloc( sizeof( w32 )*pu->unit[ ii ].i_size );
   pu->unit[ ii ].hit = CLEAR;
 }
 /* Check Buffers Integrity
    ( Obviously to be discussed ) Policy: if someone Fails --> Clean All */
 clean = CLEAR;
 for( ii = 0; ii < pu->nopu ; ii++ ){
   if( ( pu->unit[ ii ].ib == NULL )||( pu->unit[ ii ].ob == NULL )||
       ( pu->unit[ ii ].pat == NULL ) ){
     printf("\n[ Unit %d ][ I/O Buffering FAILs ]", ( int )ii );
     fflush( stdout );
     clean = SET;
     break;
   }
 }
 if( clean == SET ){
   for( ii = 0; ii < pu->nopu ; ii++ ){
     free( pu->unit[ ii ].ib );
     free( pu->unit[ ii ].ob );
   }
   return( status );
 }
 status = SUCCESS;
 return( status );

}

int
PU_Close( Processor_Unit *pu ){
int status = FAIL;
w32 ii;

 if( pu->unit != NULL ){ 
   for( ii = 0; ii < pu->nopu ; ii++ ){
     if( *( pu->unit[ ii ].stat ) == UNIT_RUNNING ){
       *( pu->unit[ ii ].stat ) = UNIT_LOST;
       pthread_cancel( pu->unit[ ii ].phy_th );
     }
     free( pu->unit[ ii ].ib );
     free( pu->unit[ ii ].ob );
     free( pu->unit[ ii ].pat );
     printf("\n[ %-18s ][ Hits Received ][%d]", 
	      pu->unit[ ii ].id, pu->unit[ ii ].hit );
     fflush( stdout );
   }
   free( pu->unit );
   pu->unit = NULL;
   status = SUCCESS;
 } 
 return( status );
  
}

int
PU_Run( Processor_Unit *pu ){
int status = FAIL;
w32 ii;
pthread_attr_t  pth_attr;

#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "PU Thread Attr" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "PU Thread Attr" );
   return( status );
 }
#endif

 for( ii = 0; ii < pu->nopu ; ii++ ){
#if defined( LinuxPCBit3 )
   if( pthread_create( &pu->unit[ ii ].phy_th, 
		       &pth_attr,		       
		       ( void* )PU_Engine, 
		       &pu->unit[ ii ] ) != 0 ){
     printf( "PU%d Thread Create", ( int )ii );
     return( status );
   }
#else
   if( pthread_create( &pu->unit[ ii ].phy_th, 
		       pth_attr,		       
		       ( void* )PU_Engine, 
		       &pu->unit[ ii ] ) != 0 ){ 
     printf( "PU%d Thread Create", ( int )ii );
     return( status );
   }
#endif
   printf( "\n[ PU-%d ][ Running ]", ( unsigned )ii );
 }

 status = SUCCESS;
 return( status );

}

int
PU_Engine( void *par ){
PUnit *pu;

#ifdef DEBUG
struct timeval tv;
float offset;
#endif

 pu = ( PUnit* )par; 
 printf("\n[ %-15s ][ %ld ]", pu->id ,( w32 )getpid( ) );
 fflush( stdout ); 
 *( pu->stat ) = UNIT_RUNNING; 
 pu->hit = CLEAR;
#ifdef DEBUG
 offset = 0.5*data;
#endif

 while( SET ){
   /* I-Barrier -------------------------------------------------------- */
   pthread_mutex_lock( pu->ib_mtx );
   if( --( *( pu->i_barrier ) ) == 0 ){
     *( pu->i_barrier )= __NoProcessors__ + 1;
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 0.+offset );
     fflush( stdout );
#endif    
     pthread_cond_signal( pu->du_cv );
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 1.+offset );
     fflush( stdout );
#endif
     pthread_cond_wait( pu->ib_cv, pu->ib_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 2.+offset );
     fflush( stdout );
#endif
   }
   pthread_mutex_unlock( pu->ib_mtx );
   /* ------------------------------------------------------------------- */
   /* --------------------= Processor Unit Algorithm =------------------- */

   pu->hit++;
   Do_Something( pu );

   /* ------------------------------------------------------------------- */
   /* O-Barrier --------------------------------------------------------- */
   pthread_mutex_lock( pu->ob_mtx );
   if( --( *( pu->o_barrier ) ) == 0 ){
     *( pu->o_barrier ) = __NoProcessors__ + 1;
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 3.+offset );
     fflush( stdout );
#endif
     pthread_cond_signal( pu->ebu_cv );
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 4.+offset );
     fflush( stdout );
#endif
     pthread_cond_wait( pu->ob_cv, pu->ob_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nPU-%d:\t%ld\t%0.1f", data, (w32)tv.tv_usec, 5.+offset );
     fflush( stdout );
#endif
   }
   pthread_mutex_unlock( pu->ob_mtx );
   /* ------------------------------------------------------------------ */
 }

 *( pu->stat ) = UNIT_LOST;
 printf("\n[ PU  ][ %-20s ]","Quit Active" );
 fflush( stdout );
 return( SUCCESS );

}



int
Do_Something( PUnit *pu ){
w32 ii, jj, kk, *ob, *ib, *patt ;
Config_Word cfg;

 ob     = pu->ob;
 ib     = pu->ib;
 patt   = pu->pat;

 cfg.all = pu->cfg[ 1 ];
 switch( cfg.bit.format_id ){

 case __ATLAS__ :
   kk = 0; 
   for( jj = 0; jj < cfg.bit.no_channels*3 ; jj++, kk++ ){
     ob[ jj ] = CLEAR;
     ob[ jj + 1 ] = CLEAR;
     ob[ jj + 2 ] = CLEAR;
     for( ii = 0 ; ii < cfg.bit.no_samples ; ii++ ){
       ob[ jj ] += ib[ ii*( kk + 1 ) ] * 3;
       ob[ jj + 1 ] += ib[ ii*( kk + 1 ) ] * 4;
     }
     for( ii = 0 ; ii < cfg.bit.no_samples ; ii++ ){
       ob[ jj + 2 ] += abs( ib[ ii*( kk + 1 ) ]  - 
			    ( ob[ jj + 1 ]*patt[ ii*( kk + 1 ) ] ) ) ;
     }
     jj += 2;
   }
   break;

 case __Monitor__ :
   break;

 case __ByPass__ :
   memcpy( ob, ib, pu->o_size*sizeof( w32 ) );
   break;

 case __Calibration__ :
   break;

 }
 return( SUCCESS );

}
