/***************
    $Header: ROD.c object for the ROD Module Application
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $
    
    02/November/99:
    Front-End Implementation Starts 
    
    04/November/99:
    RODmap.cfg decoding routines built

    10/December/99
    - Added ROD_Stop interface to dynamic Threads StartUp
    
    - Added sys_stat flag field at ROD object level
    
    - Signal Handler OutPut Redesigned
***************/

#include "ROD.h"
ROD    *sh_rod;
Handler_Filter h_filter;

/*++++++++++++++++++++++++++++++++++++++
  Rod Handler process OS signals from BusError, Ctrl-C And Segmentation Faults

  int sig_num Processed Signal
 
  ++++++++++++++++++++++++++++++++++++++*/
void
ROD_Handler( int sig_num ){

  printf("\n\n[ Received Signal %d ]", sig_num );   
  switch( sig_num ){
    case SIGBUS  :printf("[ Bus Error ]" );break;
    case SIGINT  :printf("[ Ctrl-C    ]" );break;
    case SIGSEGV :printf("[ Segmentation Fault ]");break;
  }
  fflush( stdout );
 
  sleep( 2 );

  if( sh_rod != NULL ){
    if( h_filter.bit.f_3 == CLEAR ){
      h_filter.bit.f_3 = SET;
      if( sh_rod->software != NULL ){
	if( sh_rod->software->machine != NULL )
	  sh_rod->software->machine->trg_veto = SET;
	Close_Software( sh_rod->software );
      }
    }    
    if( h_filter.bit.f_2 == CLEAR ){
      h_filter.bit.f_2 = SET;
      if( sh_rod->hardware != NULL )
	Close_Hardware( sh_rod->hardware );
    }
  }
  
  if( h_filter.bit.f_4 == CLEAR ){
    h_filter.bit.f_4 = SET;
    if( sh_rod != NULL )
      free( sh_rod );
  }
  printf("\nBye ;-)\n");
  exit( FAIL );

}

/*++++++++++++++++++++++++++++++++++++++
  Handler Install

  Warning: If Install Fails Application Ends

  ++++++++++++++++++++++++++++++++++++++*/
void
Install_Handler( ){
struct sigaction rod_sa;
sigset_t rod_set;

  sigemptyset( &rod_set );
  sigaddset( &rod_set, SIGINT );
  sigprocmask( SIG_BLOCK, &rod_set, NULL ); /* Block CTRLc only on SetUp */ 
  sigfillset( &rod_sa.sa_mask );
  rod_sa.sa_flags = CLEAR; 
  rod_sa.sa_handler = ROD_Handler;
  if( sigaction( SIGINT, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGSEGV, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  if( sigaction( SIGBUS, &rod_sa, NULL )< 0){
    perror("sigaction SIGINT");
    exit( 1 );
  }
  sigprocmask( SIG_UNBLOCK, &rod_set, NULL ); /* Now Ready to Proccess */
  sigemptyset( &rod_set );
  sigaddset( &rod_set, SIGPIPE );
  sigprocmask( SIG_BLOCK, &rod_set, NULL );
}


/*++++++++++++++++++++++++++++++++++++++
  Create memory for the MacroStructure ROD 

  **rod: Address of ROD pointer

  ++++++++++++++++++++++++++++++++++++++*/
int
ROD_Open( ROD **rod ){
word32 addr, status = SUCCESS, sys_addr;

 Install_Handler( );
 h_filter.all = CLEAR;

 *rod = ( ROD* )malloc( sizeof( ROD ) );
 if( *rod == NULL ){
   printf("\n[ ROD: Alloc: FAIL" );
   fflush( stdout );
   return( FAIL );
 }
 sh_rod = *rod;

 /* RODMap.cfg processing -------------------------------------------------- */
 if( Make_Module( &(( *rod )->info.mod), &(( *rod )->info.fe) ) == FAIL )
   return( FAIL );
 if( Make_FE( &(( *rod )->info.mod), &(( *rod )->info.fe) ) == FAIL )
   return( FAIL );
 if( Make_Ctrl( &(( *rod )->info.ctrl) ) == FAIL )
   return( FAIL );
 
 /* Still Alive? Open Hardware And Software Resources ---------------------- */
 ( *rod )->hardware = NULL ;
 ( *rod )->software = NULL ;

 if( Open_Hardware( &( ( *rod )->hardware ), 
		    ( word32 )( &( ( *rod )->info ) )) != SUCCESS ){
   printf("\n[ ROD: Hardware: Open: FAIL ]" );
   fflush( stdout );
   return( FAIL );
 }

 addr = ( w32 )( ( *rod )->hardware->ctrl.l_csr );
 sys_addr = ( w32 )( & ( ( *rod )->sys_stat ) );
 if( Open_Software( &( ( *rod )->software ), addr, sys_addr ) != SUCCESS ){
   printf("\n[ ROD: Software: Open: FAIL ]" );
   fflush( stdout );
   status = FAIL;
 }
 return( status );

}

/*+ Destroys ROD object and Software Resources +*/
int
ROD_Close( ROD* rod ){
int status = FAIL;

 if( rod == NULL ){ 
   printf("\n[ ROD: Oooooops Null Pointer ]" );
   return( status );
 }

 if(  rod->software != NULL ){
   if( Close_Software( rod->software ) != SUCCESS ){
     printf("\n[ ROD: Software: Close: FAIL ]" );
     fflush( stdout );
   }
   rod->software = NULL;
 }

 if( rod->hardware != NULL ){
   if( Close_Hardware( rod->hardware ) != SUCCESS ){
     printf("\n[ ROD: Hardware: Close: FAIL ]" );
     fflush( stdout );
   }
   rod->hardware = NULL;
 }

 free( rod );
 rod = NULL;

 status = SUCCESS;
 return( status );
 
}



int
ROD_Config( ROD *rod, word32 *cfg_word ){
int status = FAIL;
word32 info[ 30 ], addr, sys_addr;
Config_Word cfg;

 cfg.all = cfg_word[ 0 ];
 if( rod == NULL ){
   printf("\n[ ROD: Oooooops Null Pointer ]" );
   fflush( stdout );
   return( status );
 }
 if( Retrieve_VirtualMap( rod->hardware, info, cfg.bit.src_type ) != SUCCESS){
   printf("\n[ ROD: Hardware: Map Retrieve: FAIL ]" );
   fflush( stdout );
   return( status );
 }
 if( rod->software == NULL ){
   addr = ( w32 )( rod->hardware->ctrl.l_csr );
   sys_addr = ( w32 )( & ( rod->sys_stat ) );
   if( Open_Software( &( rod->software ), addr, sys_addr ) != SUCCESS ){
     printf("\n[ ROD: Software: Open: FAIL ]" );
     fflush( stdout );
     return( status );
   }
 }
 if( Config_Software( rod->software, (w32*)cfg_word, (w32*)info ) != SUCCESS){
   printf("\n[ ROD: Software: Config: FAIL ]" );
   fflush( stdout );
   return( status );
 }   
 status = SUCCESS;
 return( status );
  
}

int
ROD_Stop( ROD *rod ){
int status = FAIL;

 if(( rod == NULL )||( rod->software == NULL )){
   printf("\n[ ROD: Oooooops Null Pointer ]" );
   fflush( stdout );
   return( status );
 }
 if( Close_Software( rod->software ) != SUCCESS ){
   printf("\n[ ROD: Software: Close: FAIL ]" );
   fflush( stdout );
   return( status );
 }
 rod->software = NULL;
 status  = SUCCESS;
 return( status );

}


int
ROD_Run( ROD *rod ){
int status = FAIL;


 if( rod == NULL ){ 
   printf("\n[ ROD: Oooooops Null Pointer ]" );
   return( status );
 }
 
 if( Run_Software( rod->software ) != SUCCESS ){
   printf("\n[ ROD: Software: RUN: FAIL ]" );
   fflush( stdout );
   return( status );
 }

 status = SUCCESS;
 return( status );

}


int
ROD_Dump( ROD* rod ){
int status = FAIL;

 if( rod == NULL ){ 
   printf("\n[ ROD: Oooooops Null Pointer ]" );
   return( status );
 }
 if( rod->software == NULL ){
   printf("\n[ ROD: Software: Oooooops Null Pointer ]" );
 }
 if( Dump_Software( rod->software ) != SUCCESS ){
   printf("\n[ ROD: Software: DUMP: FAIL ]" );
   return( status );
 }
 status = SUCCESS;
 return( status );

}


void
RODinfo_Dump( ROD *rod ){
ROD_Ctrl   *ctrl;
Front_End  *fe;

 printf("\n\n#--------------= RODhardware Info =--------------#");
 /* Controller Info */
 ctrl = &( rod->hardware->ctrl ); /* shorcut */
 printf("\n# RODController #");
 printf("\n[%s]", ctrl->id );
 printf("\n[%s]", ctrl->ip_addr );
 printf("\n[Output Buffer][idx][%d]", ctrl->o_index );
 printf("\n[Input  Buffer][idx][%d]", ctrl->i_index );
 printf("\n[ACK base address][ local ][%08x]", ( unsigned )ctrl->l_csr );
 printf("\n[ACK base address][ VME   ][%08x]", ( unsigned )ctrl->r_csr );

 /* Front-End Info */
 fe = &( rod->hardware->fe ); /* shorcut */ 
 printf("\n# FrontEnd #");
 printf("\n[%s]", fe->id );
 printf("\n[%s]", fe->ip_addr );
 printf("\n[Output Buffer][idx][%d]", fe->o_index );
 printf("\n[Input  Buffer][idx][%d]", fe->i_index );
 printf("\n[ACK base address][ local ][%08x]", ( unsigned )fe->l_csr );
 printf("\n[ACK base address][ VME   ][%08x]", ( unsigned )fe->r_csr );
 printf("\n#--------------= RODhardware Info =--------------#\n");
 
}


int
Make_Ctrl( Ctrl_entry* c_entry ){
FILE* file;
char entry[ 80 ];
word32 dummy;
int done = CLEAR;

 /* Get out Controller Parametters */ 
 file = fopen( RODMap, "r" );
#ifdef VERBOSE
 printf("\n[ Searching Controller_Section... ]");
 fflush( stdout );
#endif
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Controller_Section:%d", &dummy ) == 1 ){
     if( fgets( entry, 80, file ) == NULL )
       break;
     else{
       strcpy( c_entry->id, strtok( entry, "\t" ) );
       strcpy( c_entry->ip_addr, strtok( NULL, "\n" ) );
       done = SET;
       break;
     }
   }
 }
 fclose( file );

 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );      
  
}

int
Make_Module( M_entry*  mod, FE_entry* fe ){
FILE* file;
char entry[ 80 ];
word32 dummy, ii;
int done = CLEAR;

 /* Get out Controller Parametters */ 
 file = fopen( RODMap, "r" );
#ifdef VERBOSE
 printf("\n[ Searching Module_Section... ]");
 fflush( stdout );
#endif
 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Module_Section:%d", &dummy ) == 1 ){
     for( ii = 0; ii < dummy; ii++ ){
       if( fgets( entry, 80, file ) == NULL )
	 break;
       strcpy( mod->id, strtok( entry, "\t" ) );
       strcpy( mod->ip_addr, strtok( NULL, "\t" ));
       strcpy( fe->ip_addr, strtok( NULL, "\n" ));
       if( !strcmp( mod->ip_addr, __MySelf__ ) ){
	 mod->ctrl_idx = ii;
	 done = SET;
	 break;
       }
     }
   }
 }
 fclose( file );

 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );      
}

int
Make_FE( M_entry* mod, FE_entry* fe ){
FILE* file;
char entry[ 80 ], test_1[ 20 ], test_2[ 20 ];
word32 dummy, ii, loop;
int done = CLEAR;

 file = fopen( RODMap, "r" );

#ifdef VERBOSE
 printf("\n[ Searching Module_Section... ]");
 fflush( stdout );
#endif

 while( SET ){
   if( fgets( entry, 80, file ) == NULL )
     break;
   if( sscanf( entry, "Front_End_Section:%d", &dummy ) == 1 ){
     for( ii = 0; ii < dummy; ii++ ){
       if( fgets( entry, 80, file ) == NULL )
	 break;
       strcpy( fe->id, strtok( entry, "\t" ) );
       strcpy( test_1, strtok( NULL, "\t" ));
       strcpy( test_2, strtok( NULL, "\t" ));
       if( !strcmp( test_1, fe->ip_addr ) ){
	 done = SET;
	 break;
       }
     }
     if( done != SET )
       break;

     loop = atoi( test_2 );
     for( ii = 0; ii < loop ; ii++ ){
       strcpy( test_1, strtok( NULL, "\t" ));
       if( !strcmp( test_1, __MySelf__ ) ){
	 if( ii == (loop-1) )
	   strcpy( test_2, strtok( NULL, "\n" ));
	 else
	   strcpy( test_2, strtok( NULL, "\t" ));
	 mod->fe_idx = ii;
	 done = SET;
	 break;
       }
     }
     break;
   }
 }
 fclose( file );

 if( done == SET )
   return( SUCCESS );
 else
   return( FAIL );      
  return( SUCCESS );
}


