/******
    $Header: TileCal ROD Module. Main Thread.
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $
    
    02/November/99: 
    Front-End Integration

    08/November/99:
    Barrier-Machine Integration

    10/December/99
    Controller State Machine Recoded
*******/

#include "ROD.h"

char bd[ ] = { 27, 91, 49, 109, 0 };
char us[ ] = { 27, 91, 49, 27, 91, 52, 109,0 };
char atoff[ ] = { 27, 91, 48, 109, 0 };


/*+ 
  Main Thread will be the ROD Module Controller. At least two VME pointers are necessary: The local Module CSR and The Remote ROD Controller CSR.
  Bellow must be something like a switch case which decodes the VME 
  incomming ROD controller commands.   
+*/

int
main( int argc, char *argv[ ] ){
ROD *rod;
Config_Word cfg[ __ConfigPackSize__ ];
Machine *mch;
System_State *sys_stat;
volatile word32 *r_csr, *l_csr;
word32  jj = CLEAR, set =SET, TRGStat =CLEAR, ii =CLEAR;


 if( ROD_Open( &rod ) != SUCCESS ){
   printf("\n[ ROD Open: FAIL" );
   if( ROD_Close( rod ) != SUCCESS ){
     printf("\n[ ROD Close: FAIL" );
     fflush( stdout );
     return( FAIL );
   }
   printf("\nBye, ;-)\n");
   fflush( stdout );
   return( FAIL );
 }
#ifdef VERBOSE 
 printf("\n[ ROD Open: SUCESS");
 fflush( stdout );
#endif

 /* Shortcuts */
 r_csr = rod->hardware->ctrl.r_csr;
 l_csr = rod->hardware->ctrl.l_csr;
 l_csr[ 0 ] = MASK_COMMAND;
 mch = rod->software->machine;
 mch->trg_veto = CLEAR;
 sys_stat = &( rod->sys_stat );
 sys_stat->all = CLEAR;
 sys_stat->bit.stop = SET;

 fflush( stdout );
 set = SET;

 /* ROD-Module Loop ------------------------------------------------------- */
 while( set ){
   usleep( 200000 ); /* Command Response Timing 200msec */

   ++jj;
   if( !( jj%5 ) ){
     if( sys_stat->bit.stop == CLEAR )
       ROD_Dump( rod ); /* Local Status Downloading */
     printf("\n[ Module: CTRL-Loop: Waiting Command ]" );
     fflush( stdout );
   }

   if( l_csr[ 0 ] == MASK_COMMAND )
     continue;
  
   switch( l_csr[ 0 ] ){

   case SAMPLE:
     printf("\n[SAMPLE][ Received ]");
     TRGStat = mch->trg_veto; 
     mch->trg_veto = CLEAR;
     printf("\n[ Trigger Veto %u ][ Event Buffer Ready ]", (unsigned)TRGStat );
     mch->trg_veto = TRGStat;
     r_csr[ 1 ] = SAMPLE;
     break;

   case RUN:
     printf("\n[RUN][ Received ]");
     if( sys_stat->bit.config == CLEAR ){
       printf("\n[WARNING][ROD-Machine: NOT Configured]");
       r_csr[ 1 ] = ACK_1;
       break;
     }
     if( sys_stat->bit.run == SET ){
       printf("\n[WARNING][ROD-Machine: Already Running]");
       r_csr[ 1 ] = ACK_2;
       break;
     }
     if( sys_stat->bit.stop == SET ){
       printf("\n[WARNING][ROD-Machine: Down]");
       r_csr[ 1 ] = ACK_3;
       break;
     }
     /* Real Action Begins ---------------- */

     if( ROD_Run( rod )  == SUCCESS ){
       sys_stat->bit.run = SET;
       sys_stat->bit.stop = CLEAR;
       sys_stat->bit.pause = CLEAR;
       printf("\n[ ROD-Machine RUNNING ]");
       r_csr[ 1 ] = RUN;
     }

     /* ----------------------------------- */
     break;
       
   case PAUSE:
     printf("\n[PAUSE][ Received ]");
     if(sys_stat->bit.stop == SET ){
       printf("\n[WARNING][ROD-Machine: Down]");
       r_csr[ 1 ] = ACK_3;
       break;
     }
     if(sys_stat->bit.run == CLEAR ){
       printf("\n[WARNING][ROD-Machine: NOT Running]");
       r_csr[ 1 ] = ACK_4;
       break;
     }
     /* RODMachine Pause ------------------- */

     sys_stat->bit.pause = SET;
     mch->trg_veto = SET;
     printf("\n[ Module-Machine PAUSED ]");
     r_csr[ 1 ] = PAUSE;
     /* ------------------------------------ */
     break;

   case CONTINUE:
     printf("\n[CONTINUE][ Received ]");
     if( sys_stat->bit.stop == SET ){
       printf("\n[WARNING][ROD-Machine: Down]");
       r_csr[ 1 ] = ACK_3;
       break;
     }
     if( ( sys_stat->bit.pause == CLEAR )&&( sys_stat->bit.run == CLEAR )){
       printf("\n[WARNING][ROD-Machine: NOT Running]");
       r_csr[ 1 ] = ACK_4;
       break;
     }

     /* RODMachine Continue ---------------- */
     
     mch->trg_veto = CLEAR;
     sys_stat->bit.pause = CLEAR;
     printf("\n[ ROD-Machine: TRG-Veto Released ]");
     r_csr[ 1 ] = CONTINUE;
     /* ------------------------------------ */
     break;

   case CONFIG:
     printf("\n[CONFIG][ Received ]");
     if( sys_stat->bit.run == SET ){
       printf("\n[WARNING][ROD-Machine: Running: STOP/PAUSE Before]");
       r_csr[ 1 ] = ACK_7;
       break;
     }
     cfg[ 0 ].all = l_csr[ __cfgOffset__ ];
     cfg[ 1 ].all = l_csr[ __cfgOffset__ +1 ];
     cfg[ 2 ].all = l_csr[ __cfgOffset__ +2 ];
     for( ii = 0; ii <  __ConfigPackSize__ ; ii++ ){
       printf("\n[ %-30s ][ $%08x ]", "Config Word", 
	      ( unsigned )( cfg[ ii ].all ) );
     }
     /* Configure RODModule Software ------ */
     if( cfg[ 0 ].all == 0 ){
       printf("\n[WARNING][ROD-Machine: Invalid Configuration Value ]");
       r_csr[ 1 ] = ACK_6;
       break;
     }
     if( ROD_Config( rod, ( word32* )( &( cfg[ 0 ] ) ) ) != SUCCESS ){
       printf("\n[WARNING][ROD-Machine: Internal FAIL]");
       r_csr[ 1 ] = ACK_5;
       break;
     }
     sys_stat->bit.config = SET;
     sys_stat->bit.stop = CLEAR;
     r_csr[ 1 ] = CONFIG;
     printf("\n[ ROD-Module CONFIGURED ]");

     /* ----------------------------------- */
     break;

   case STOP:
     printf("\n[STOP][ Received ]");
     if( sys_stat->bit.stop == SET ){
       printf("\n[WARNING][ROD-Machine: NOT Running]");
       r_csr[ 1 ] = ACK_4;
       break;
     }
     /* Kill RODMachine ------------------- */

     if( ROD_Stop( rod ) != SUCCESS ){
       printf("\n[WARNING][ROD-Machine: Internal FAIL]");
       fflush( stdout );
       r_csr[ 1 ] = ACK_5;
       break;
     }
     sys_stat->bit.run = CLEAR;
     sys_stat->bit.stop = SET;
     sys_stat->bit.pause = CLEAR;
     r_csr[ 1 ] = STOP;
     printf("\n[ ROD-Machine Stopped ]");
     fflush( stdout );

     /* ----------------------------------- */
     break;

   case SHUTDOWN:
     r_csr[ 1 ] = SHUTDOWN;
     printf( "\n\n[SHUTDOWN][ Received ]\n[ROD-Module: Going Down]\n");
     set = CLEAR;
     break;

   default:
     printf( "\n[ %-30s ]", "Unknown Command" );

   }
   l_csr[ 0 ] = MASK_COMMAND;
   r_csr[ 0 ] = CMD_RECEIVED;
 }
 /* ---------------------------------------------------------------------- */
  
 if( ROD_Close( rod ) != SUCCESS ){
   printf("\n[ ROD Close: FAIL" );
   fflush( stdout );
   return( FAIL );
 }
#ifdef VERBOSE
 else{
   printf("\n[ ROD Close: SUCESS" );
   fflush( stdout );
 }
#endif
 printf("\nBye, ;-)\n");
 return( SUCCESS );

}

