/*************
    $Header: RODhardware Object for the ROD-Module Application
    Created & Maintaned by JB. Juanba.Romance@ific.uv.es $

    26/March/99      ROD-Hardware Layer Integration

    29/March/99      ROD-VME interface done

    03/May/99        Cross Develop. starts. LinuxPCBit3/MVME167/MVME2604PowerPC

    08/May/99        RODSlink squeletor ready

    13/May/99        ROD-Controller Interface starts

    26/May/99        Retrieve_Virtual_Map is Updated to work "ONFLY"

    02/November/99   Front-End Integration Starts

    07/December/99   Patch LynxOS SMEM default value 10. Module Front-End Map: CSR-Input-Buffer. Data types are maintaned if modified LynxOS kernels arrive. It is forced to the RODModule Application :-(
**************/

#include "RODhardware.h"

int 
Open_Hardware( ROD_Hardware **rod_hard, word32 info_ptr ){
int status = SUCCESS;
  
 *rod_hard = ( ROD_Hardware* )malloc( sizeof( ROD_Hardware ) );
 if( *rod_hard == NULL ){
   printf("\n[ ROD: Hardware: Alloc: FAIL ]");
   status = FAIL;
 }
 else{
#if defined( PowerPC_Lynx3_0 ) 
   if( Open_RODSlink( &( ( *rod_hard )->slink ) ) != SUCCESS ){
     printf("\n[ ROD SlinkHardware: Open: FAIL" );
     fflush( stdout );
     status = FAIL;
   }
#endif
   if( Open_RODVME( ( &( *rod_hard )->vme ), info_ptr ) != SUCCESS ){
     printf("\n[ VMEHardware: Open: FAIL" );
     fflush( stdout  );
     return( FAIL );
   }

   if( SetUp_Hardware( *rod_hard, info_ptr ) != SUCCESS ){
     printf("\n[ ROD-Hardware: SetUp: FAILs" );
     return( FAIL );
   }

   if( Handshake_Controller( *rod_hard ) != SUCCESS ){
     printf("\n[ ROD-Hardware: Controller: Detection: FAILs" );
     status = FAIL;
   }
#ifdef VERBOSE
   else
     printf("\n[ ROD-Hardware: Controller: Detection: SUCCESS" );
#endif

   fflush( stdout );

 }
 return( status );

}


int
Close_Hardware( ROD_Hardware *rod_hard ){
int status = FAIL;

 if( rod_hard == NULL ){ 
   printf("\n[ ROD Hardware: Oooooops Null Pointer" );
   return( status );
 }

#if defined( PowerPC_Lynx3_0 )
 if( rod_hard->slink != NULL ){
   if( Close_RODSlink( rod_hard->slink ) != SUCCESS ){
     printf("\n[ROD SlinkHardware: Close: FAIL" );
     status = FAIL;
   }
#ifdef VERBOSE
   else
     printf("\n[ ROD SlinkHardware: Close: SUCESS" );
#endif
   fflush( stdout );
 }
#endif
 if( rod_hard->vme != NULL ){
   if( Close_RODVME( rod_hard->vme ) != SUCCESS ){
     printf("\n[ ROD VMEHardware: Close: FAIL" );
     status = FAIL;
   }
#ifdef VERBOSE
   else
     printf("\n[ ROD VMEHardware: Close: SUCESS" );
#endif
   fflush( stdout  );
 }
 free( rod_hard );
 status = SUCCESS;
 return( status );

}


int
SetUp_Hardware( ROD_Hardware* rod_hard, word32 info_ptr ){
int status =SUCCESS;

 if( SetUp_Local( rod_hard ) != SUCCESS ){
   printf("\n[ ROD-Hardware: ROD-Controller SetUp: FAILs" );
   status = FAIL;
 }
 if( SetUp_Controller( rod_hard, info_ptr ) != SUCCESS ){
   printf("\n[ ROD-Hardware: ROD-Controller SetUp: FAILs" );
   status = FAIL;
 }
 if( SetUp_FrontEnd( rod_hard, info_ptr ) != SUCCESS ){
   printf("\n[ROD-Hardware: FE-SetUp: FAILs" );
 }
 return( status );

}

/*+ 

  03/November/99: Retrieves from RODvme Local hardware resources.
  CSR pointers SetUp and DMA objects

+*/
int
SetUp_Local( ROD_Hardware* rod_hard ){
Search_Key  key;
word32      obj_index, addr;

 /* CSR */ 
 sprintf( key.machine, "%s", "localhost" );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_csr = ( word32* )addr;
 
 /* IBuffer */
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_ibuf = ( word32* )addr;
 rod_hard->i_index = obj_index;

 /* OBuffer */
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->l_obuf = ( word32* )addr;
 rod_hard->o_index = obj_index;
 fflush( stdout ); 
 return( SUCCESS );

}

/*+ 

  03/November/99: Retrieves from RODvme ROD-Controller hardware resources.
  CSR pointers and DMA objects SetUp
  !!!!! WARNING !!!!! SetUp_Local MUST be called first.

+*/
int
SetUp_Controller( ROD_Hardware* rod_hard, word32 info_ptr ){
ROD_info   *info;
Search_Key  key;
word32      obj_index, addr;

 info = ( ROD_info* )info_ptr;
 sprintf( rod_hard->ctrl.id, "%s", info->ctrl.id );
 sprintf( rod_hard->ctrl.ip_addr, "%s", info->ctrl.ip_addr );

 /* CSR */ 
 sprintf( key.machine, "%s", rod_hard->ctrl.ip_addr );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.r_csr = ( word32* )addr;
 rod_hard->ctrl.r_csr = 
   &rod_hard->ctrl.r_csr[ ( info->mod.ctrl_idx * CSR_ROD_SIZE ) ];
 rod_hard->ctrl.l_csr = &rod_hard->l_csr[ CSR_FE_SIZE ];

 /* IBuffer */
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.i_index = obj_index;

 /* OBuffer */
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->ctrl.o_index = obj_index;
 rod_hard->ctrl.lives   = LIVE_CHANCE;
 return( SUCCESS );

}

/*+ 

  03/November/99: Retrieves from RODvme Front-End hardware resources.
  CSR pointers SetUp and DMA objects
  !!!!! WARNING !!!!! SetUp_Local MUST be called first.

+*/
int
SetUp_FrontEnd( ROD_Hardware *rod_hard, word32 info_ptr ){
ROD_info   *info;
Search_Key  key;
word32      obj_index, addr, offset;

 info = ( ROD_info* )info_ptr;
 sprintf( rod_hard->fe.id, "%s", info->fe.id );
 sprintf( rod_hard->fe.ip_addr, "%s", info->fe.ip_addr );
 offset = (( info->mod.fe_idx * CSR_ROD_SIZE ) + CSR_CTRL_SIZE );
 /* CSR */ 
 sprintf( key.machine, "%s", rod_hard->fe.ip_addr );
 sprintf( key.dev_id, "%s", "CSR_Space" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->fe.r_csr = &( (( word32* )addr )[ offset ]);
 rod_hard->fe.l_csr = rod_hard->l_csr;

 /* IBuffer */
 sprintf( key.dev_id, "%s", "Input_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->fe.i_index = obj_index;

 /* OBuffer relies on the LynxOS Kernel SMEM value, see top header */
#if defined( SMEM_10 )
 rod_hard->fe.o_index = 0;
#else
 sprintf( key.dev_id, "%s", "Output_Buffer" ); 
 if( Get_VMEObject( rod_hard->vme, ( char* )( &key ), &addr, &obj_index ) 
     != SUCCESS ){
   printf("\n[ %-12s ][ %-12s ][ %-12s ]",
	  key.machine, key.dev_id, "Object Not Found" );
   return( FAIL );
 }
 rod_hard->fe.o_index = obj_index;
#endif

 rod_hard->fe.lives   = LIVE_CHANCE;
 return( SUCCESS );
}


/*+ 

  03/November/99: Function Redesign. 
  Boot Protocol completely isolated. 

+*/   
int
Handshake_Controller( ROD_Hardware *rod_hard ){
word32 time_out;
word32 *l_csr, *r_csr;

 r_csr = rod_hard->ctrl.r_csr;
 l_csr = rod_hard->ctrl.l_csr;
 l_csr[ 0 ] = CONTROLLER_DEAD; /* We don't know Controller Status */
 time_out = BOOT_TIME_OUT;
 while( l_csr[ 0 ] == CONTROLLER_DEAD ){
   --time_out;
   r_csr[ 0 ] = ROD_ALIVEs;  /* Try to connect with a ROD controller */
   printf("\n[ Waiting Controller ACK .. ]" );
   if( !time_out ){
     printf("\n[ ROD Controller is Dead ..... ]\n");
     return( FAIL );
   }
   sleep( 1 );
 }
 l_csr[ 0 ] = BUSY; /* Still Alive ---> ROD Controller has been Detected */
 l_csr[ __IUHandshake__ ] = MASK_COMMAND; 
 printf("\n[ ROD Controller is Up ]\n");
 return( SUCCESS );
  
}



/* 
 * 06/May/99: Interface With RODMachine to retrieve Physical Object features
 * 26/May/99: New implementation to be able to work "ONFLY" and perform I/O
 * 03/Nov/99: Recoded to extract IO parametters from internal RODhardware 
 
  The Program flow follows below:
     ROD.c        RODhardware.c           RODvme.c
   RODConfig_______Retrieve_Virtual_______Get_VME_Object..
     \ 
      \            RODsoftware.c           Machine.c
       \___________Config_Software_________Machine_Config___IUnit_Config 
                                                     \___OUnit_Config
					     
  --= info pointer format =--
     Local info
       Input Buffer  [0][1][2] vaddr, dev_size, paddr
       OutPut Buffer [3][4][5] idem than above
       LocalCSR      [6] vaddr
     Remote info
       DMAObjects    [7][8] vaddr of Input/OutPut Buffers
       RemoteCSR     [9] vaddr

  key_word --> selects external data injection device

 * 06/July/99: Add to info pointer SlinkObject pointer info[ 10 ]

*/
int
Retrieve_VirtualMap( ROD_Hardware *rod_hard, word32 *info, word32 src_type ){
VMESlave* dev;
word32 obj_index;

 if( rod_hard == NULL ){
   printf("\n[ ROD Hardware: Oooooops Null Pointer" );
   return( FAIL );
 }
 
 obj_index = rod_hard->i_index;
 info[ 0 ] = ( word32 )rod_hard->l_ibuf;
 dev = ( VMESlave* )( rod_hard->vme->rod_dev[ obj_index ].vme_obj );

 info[ 1 ] = dev->size;
#if defined( MVME167_Lynx2_5_1 ) || defined( PowerPC_Lynx3_0 )
 info[ 2 ] = dev->phy_addr;
#else
 info[ 2 ] = dev->ram.pci_addr;
#endif 

 obj_index = rod_hard->o_index;
 info[ 3 ] = ( word32 )rod_hard->l_obuf;
 dev = ( VMESlave* )( rod_hard->vme->rod_dev[ obj_index ].vme_obj );

 info[ 4 ] = dev->size;
#if defined( MVME167_Lynx2_5_1 ) || defined( PowerPC_Lynx3_0 )
 info[ 5 ] = dev->phy_addr;
#else
 info[ 5 ] = dev->ram.pci_addr;
#endif

 if( src_type == __Controller__ ){
   info[ 6 ] = ( word32 )( rod_hard->ctrl.l_csr );
   info[ 7 ] = ( word32 )( &rod_hard->vme->rod_dev[ rod_hard->ctrl.i_index ] );
   info[ 8 ] = ( word32 )( &rod_hard->vme->rod_dev[ rod_hard->ctrl.o_index ] );
   info[ 9 ] = ( word32 )( rod_hard->ctrl.r_csr );
 }
 else{
   if( src_type == __FrontEnd__ ){
     printf("\n\n[ ROD: hardware: Retrieving IO-info for front_end %s ]",
	    rod_hard->fe.id );
     info[ 6 ] = ( word32 )( rod_hard->fe.l_csr );
     info[ 9 ] = ( word32 )( rod_hard->fe.r_csr );
     info[ 7 ] = ( word32 )( &rod_hard->vme->rod_dev[ rod_hard->fe.i_index ] );
     /* !!!!! Beware Below pointer could be NULL if SMEM_10 is defined !!!!! */
     info[ 8 ] = ( word32 )( &rod_hard->vme->rod_dev[ rod_hard->fe.o_index ] );

     printf("\n[ ROD: hardware: Checking front_end CSR pointers ]" );
     printf("\n[ Slave  Pointer: $%08x ]", rod_hard->fe.l_csr[ 0 ] );
     fflush( stdout );
     printf("\n[ Master Pointer: $%08x ]", rod_hard->fe.r_csr[ 0 ] );
     fflush( stdout );

   }
   else{
     printf("\n[ ROD Hardware: Unknown Data Source");
   }
 }

 /* Slink Device */
#if defined( PowerPC_Lynx3_0 )
 info[ 10 ] = ( word32 )( rod_hard->slink );
#endif
 return( SUCCESS );
 
}
