#include "TTCUnit.h"

int
TTC_Open( TTC_Unit *ttc ){
int status = FAIL;

 sprintf( ttc->id, "%s", "TTCEngine" );
 ttc->hit = CLEAR;
 ttc->trg_mode = __NoneTRG__ ;
 ttc->info.event_no = CLEAR;
 ttc->info.bcid_no  = CLEAR;

 /* Thread Management */
 pthread_mutex_init( &ttc->ttc_mtx, NULL );
 pthread_cond_init( &ttc->ttc_cv, NULL );

 status = SUCCESS;
 return( status );

}

int
TTC_Close( TTC_Unit *ttc ){
int status = FAIL;

 if( *( ttc->stat ) == UNIT_RUNNING ){
   *( ttc->stat ) = UNIT_LOST;
   pthread_cancel( ttc->phy_th );
 }
 printf("\n[ %-18s ][ Hits Received ][%ld]", ttc->id, ttc->hit );
 fflush( stdout );
 status = SUCCESS;
 return( status );
  
}


int
TTC_Run( TTC_Unit *ttc ){
int status = FAIL;
pthread_attr_t  pth_attr;

 if( ttc == NULL ){
   printf("\n[ TTC Object not Initialized ]" );
   fflush( stdout );
   return( status );
 }
#if defined( LinuxPCBit3 )
 if( pthread_attr_init( &pth_attr ) != 0 ){
   perror( "TTC Thread Attr" );
   return( status );
 }
#else
 if( pthread_attr_create( &pth_attr ) != 0 ){
   perror( "TTc Thread Attr" );
   return( status );
 }
#endif

#if defined( LinuxPCBit3 )
 if( pthread_create( &ttc->phy_th, 
		     &pth_attr, 
		     ( void* )TTC_Engine, 
		     ttc ) != 0)
#else
 if( pthread_create( &ttc->phy_th, 
		     pth_attr, 
		     ( void* )TTC_Engine, 
		     ttc ) != 0)
#endif
   {
     perror( "EBU Thread Create" );
     return( status );
   }
 pthread_mutex_unlock( &ttc->ttc_mtx );
 printf( "\n[ %-15s ][ Running ]", ttc->id );
 fflush( stdout );
 status = SUCCESS;
 return( status );

}

int
TTC_Engine( void *par ){
w32 jj;
TTC_Unit *ttc;
w32 *r_csr, *l_csr;

#ifdef DEBUG
struct timeval tv;
float offset = 3.0;
#endif

 ttc = ( TTC_Unit* )par;
 printf("\n[ %-15s ][ %ld ]", ttc->id , ( w32 )getpid( ) );
 fflush( stdout );
 *( ttc->stat ) = UNIT_RUNNING;
 jj = CLEAR;
 pthread_mutex_lock( &ttc->ttc_mtx ); /* Thread AutoBlocking */
 ttc->info.trg_type = TYPE_1;
 /* ATLAS will:
    - Reset BCID every 88usec 
    - Reset event_no every 160 sec
    In Automode we can simulate it, using alarms...
 */
 l_csr = ttc->l_csr;
 r_csr = ttc->r_csr;  

 while( SET ){

   if( *( ttc->quit ) == SET )
     break;

   if( *( ttc->trg_veto ) == CLEAR ){
     /* Trigger Barrier Analysis ---------------------------------------- */
     pthread_mutex_lock( ttc->b_mtx );

     /* Critical Section --------------------------------------------- */
     switch( ttc->trg_mode ){
     case __AutoTRG__ :
     
       /*Get_timing( &timing ); */
       /*       for( jj = 0 ; jj < 20000 ; jj++ ); */
       ttc->hit++;
       ttc->info.event_no++;
       ttc->info.bcid_no++;
       if( l_csr[ __IUHandshake__ + 1 ] > __DataSRCFail__ )
	 AutoClean( ttc ); 
       break;
       
     case __NoneTRG__ :
       usleep( 1000 );
       break;
       
     case __ControllerTRG__ :
       usleep( 1000 );
       break;
       
     case __TTC__ :
       usleep( 1000 );
       break;
       
     }
     /* ------------------------------------------------------------- */
   if( --( *( ttc->barrier ) ) == 0 ){ /* Full barrier */
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nTTC:\t%ld\t%0.1f", (w32)tv.tv_usec, 0.+offset);
     fflush( stdout );
#endif
     *( ttc->barrier ) = __BarrierSize__ ;
     pthread_cond_broadcast( ttc->b_cv );
          
   }
   else{
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nTTC:\t%ld\t%0.1f", (w32)tv.tv_usec, 1.+offset);
     fflush( stdout );
#endif
     pthread_cond_wait( ttc->b_cv, ttc->b_mtx );
#ifdef DEBUG
     gettimeofday( &tv, NULL );
     printf("\nTTC:\t%ld\t%0.1f", (w32)tv.tv_usec, 2.+offset);
     fflush( stdout );
#endif
   }
   pthread_mutex_unlock( ttc->b_mtx );
     /* ------------------------------------------------------------------- */
#ifdef DEBUG
   gettimeofday( &tv, NULL );
   printf("\nTTC:\t%ld\t%0.1f", (w32)tv.tv_usec, 3.+offset);
   fflush( stdout );
#endif
   pthread_cond_wait( &ttc->ttc_cv, &ttc->ttc_mtx );/* Wait Cycle on IU-DU */
#ifdef DEBUG
   gettimeofday( &tv, NULL );
   printf("\nTTC:\t%ld\t%0.1f", (w32)tv.tv_usec, 4.+offset);
   fflush( stdout );
#endif
   }
   else
     usleep( 10000 );
 }
 pthread_mutex_unlock( &ttc->ttc_mtx );
 pthread_mutex_unlock( ttc->b_mtx );

 *( ttc->stat ) = UNIT_LOST;
 printf("\n[ TTC ][ %-20s ]","Quit Active" );
 fflush( stdout );

 return( SUCCESS );

}


int
Get_timing( w32 *wait ){
  /* First Approach Normal distribution */
  *wait = 1 + ( int )( 200.0*rand( ) / ( RAND_MAX + 1.0 ) );
  return( SUCCESS );

}

void
AutoClean( TTC_Unit *ttc ){
w32 *r_csr, *l_csr;
 
 *( ttc->trg_veto ) = SET;
 r_csr = ttc->r_csr;
 l_csr = ttc->l_csr;
 printf("\n[ [WARNING]IUnit: Lost Last %u Handshakes [NOmax Faults[%u] ]",
	( unsigned )l_csr[ __IUHandshake__ + 1 ], 
	( unsigned ) __DataSRCFail__ );
 fflush( stdout );

}
